; Copyright 1995-1998 Just For Fun Software, Inc., all rights reserved
; Author:  George Woltman
; Email: woltman@magicnet.net
;
; This routine implements a discrete weighted transform to quickly multiply
; two numbers.  This code derived from the Lucas-Lehmer assembly code, but
; only handles small numbers (one pass FFTs) for now.
;
; This code handles FFT sizes from 32 doubles up to 1024 doubles.
;
; You will not stand a chance of understanding any of this code without
; thoroughly familiarizing yourself with fast fourier transforms.  This
; code was adapted from an algorithm described in Richard Crandall's article
; on Discrete Weighted Transforms and Large-Integer Arithmetic.
;

	TITLE   setup

	.386

_TEXT32 SEGMENT PARA USE32 PUBLIC 'DATA'

EXTRN	_MAXERR:QWORD
EXTRN	_SUMINP:QWORD
EXTRN	_SUMOUT:QWORD
EXTRN	_ERRCHK:DWORD
EXTRN	_PARG:DWORD
EXTRN	_FFTLEN:DWORD
EXTRN	_SRCARG:DWORD
EXTRN	_SRC2ARG:DWORD
EXTRN	_DESTARG:DWORD
EXTRN	_NORM_ARRAY:DWORD

;
; Global variables needed in squaring
;

	align	32
SQRTHALF	DQ	0.0		; Used in all ffts
HALF		DD	0.5		; Used in all ffts
BIGVAL		DD	0.0		; Used to round to an integer
scaled_numbig	DD	0		; numbig * (2^32 / n)
scaled_numlit	DD	0		; numlit * (2^32 / n)
P309		DQ	0.309		; Used in five_reals_fft/unfft
M809		DQ	-0.809		; Used in five_reals_fft/unfft
M262		DQ	-2.618		; Used in five_reals_fft/unfft
M382		DQ	-0.382		; Used in five_reals_fft/unfft
P951		DQ	0.951		; Used in five_reals_fft/unfft
P588		DQ	0.588		; Used in five_reals_fft/unfft
M162		DQ	-1.617		; Used in five_reals_fft/unfft
P618		DQ	0.618		; Used in five_reals_fft/unfft
P623		DQ	0.623		; Used in seven_reals_fft/unfft
M358		DQ	-0.358		; Used in seven_reals_fft/unfft
P404		DQ	4.040		; Used in seven_reals_fft/unfft
P975		DQ	0.975		; Used in seven_reals_fft/unfft
P445		DQ	0.445		; Used in seven_reals_fft/unfft
P180		DQ	1.802		; Used in seven_reals_fft/unfft
M223		DQ	-0.223		; Used in seven_reals_fft/unfft
M901		DQ	-0.901		; Used in seven_reals_fft/unfft
M691		DQ	-0.691		; Used in seven_reals_fft/unfft
P866		DQ	0.866		; Used in six_reals_fft/unfft
P433		DQ	0.433		; Used in six_reals_fft/unfft
P577		DQ	0.577		; Used in six_reals_fft/unfft
P25		DD	0.25		; Used in six_reals_fft/unfft
P75		DD	0.75		; Used in six_reals_fft/unfft
P3		DD	3.0		; Used in six_reals_fft/unfft
count1		DD	0		; Counters used in common fft code
count2		DD	0
count3		DD	0
count4		DD	0
sincos1		DD	0		; Pointers to sine/cosine tables
sincos2		DD	0		; used in common fft code
sincos3		DD	0
sincos4		DD	0
srcptr		DD	0		; Saved pointer in common fft code
extra_bits	DD	0		; Number of unnormalized adds
					; that can be safely performed.
fft_1_addr	DD	0		; Address of fft routine
fft_2_addr	DD	0		; Address of squaring routine
fft_3_addr	DD	0		; Address of multiply routine
fft_4_addr	DD	0		; Address of no fft multiply routine
		; These values only used during setup
p		DD	0		; Mersenne prime being tested
numlit		DD	0		; number of small words in the fft
numbig		DD	0		; number of big words in the fft
P5		DD	5.0		; Used in pfa_5_setup
P7		DD	7.0		; Used in pfa_7_setup
jmptable	DD	755,	32,	OFFSET fft32_1, OFFSET fft32_2
		DD			OFFSET fft32_3, OFFSET fft32_4
		DD			1, 1, 1, 1, 0
		DD	943,	40,	OFFSET fft40_1, OFFSET fft40_2
		DD			OFFSET fft40_3, OFFSET fft40_4
		DD			1, 1, 1, 1, 20, 40, 0
		DD	1115,	48,	OFFSET fft48_1, OFFSET fft48_2
		DD			OFFSET fft48_3, OFFSET fft48_4
		DD			1, 1, 1, 1, 24, 48, 0
		DD	1301,	56,	OFFSET fft56_1, OFFSET fft56_2
		DD			OFFSET fft56_3, OFFSET fft56_4
		DD			1, 1, 1, 1, 28, 56, 0
		DD	1497,	64,	OFFSET fft64_1, OFFSET fft64_2
		DD			OFFSET fft64_3, OFFSET fft64_4
		DD			1, 1, 1, 1, 0
		DD	1859,	80,	OFFSET fft80_1, OFFSET fft80_2
		DD			OFFSET fft80_3, OFFSET fft80_4
		DD			2*256+4, 8, 1, 1, 20, 80, 0
		DD	2209,	96,	OFFSET fft96_1, OFFSET fft96_2
		DD			OFFSET fft96_3, OFFSET fft96_4
		DD			2*256+4, 10, 1, 1, 24, 96, 0
		DD	2585,	112,	OFFSET fft112_1, OFFSET fft112_2
		DD			OFFSET fft112_3, OFFSET fft112_4
		DD			3*256+4, 12, 1, 1, 28, 112, 0
		DD	2951,	128,	OFFSET fft128_1, OFFSET fft128_2
		DD			OFFSET fft128_3, OFFSET fft128_4
		DD			3*256+4, 14, 1, 1, 0
		DD	3669,	160,	OFFSET fft160_1, OFFSET fft160_2
		DD			OFFSET fft160_3, OFFSET fft160_4
		DD			2*256+8, 8*256+2, 16, 1, 20,80,160,0
		DD	4365,	192,	OFFSET fft192_1, OFFSET fft192_2
		DD			OFFSET fft192_3, OFFSET fft192_4
		DD			2*256+8, 10*256+2, 20, 1, 24,96,192,0
		DD	5091,	224,	OFFSET fft224_1, OFFSET fft224_2
		DD			OFFSET fft224_3, OFFSET fft224_4
		DD			3*256+8, 12*256+2, 24, 1, 28,112,224,0
		DD	5823,	256,	OFFSET fft256_1, OFFSET fft256_2
		DD			OFFSET fft256_3, OFFSET fft256_4
		DD			3*256+8, 14*256+2, 28, 1, 0
		DD	7267,	320,	OFFSET fft320_1, OFFSET fft320_2
		DD			OFFSET fft320_3, OFFSET fft320_4
		DD			2*256+16, 9*256+4, 39, 1, 20,80,320,0
		DD	8673,	384,	OFFSET fft384_1, OFFSET fft384_2
		DD			OFFSET fft384_3, OFFSET fft384_4
		DD			2*256+16, 11*256+4, 47, 1, 24,96,384,0
		DD	10121,	448,	OFFSET fft448_1, OFFSET fft448_2
		DD			OFFSET fft448_3, OFFSET fft448_4
		DD			3*256+16, 13*256+4, 55, 1, 28,112,448,0
		DD	11571,	512,	OFFSET fft512_1, OFFSET fft512_2
		DD			OFFSET fft512_3, OFFSET fft512_4
		DD			3*256+16, 15*256+4, 63, 1, 0
		DD	14381,	640,	OFFSET fft640_1, OFFSET fft640_2
		DD			OFFSET fft640_3, OFFSET fft640_4
		DD			2*256+32, 9*256+8, 39*256+2, 79
		DD			20, 80, 320, 640, 0
		DD	17111,	768,	OFFSET fft768_1, OFFSET fft768_2
		DD			OFFSET fft768_3, OFFSET fft768_4
		DD			2*256+32, 11*256+8, 47*256+2, 95
		DD			24, 96, 384, 768, 0
		DD	19931,	896,	OFFSET fft896_1, OFFSET fft896_2
		DD			OFFSET fft896_3, OFFSET fft896_4
		DD			3*256+32, 13*256+8, 55*256+2, 111
		DD			28, 112, 448, 896, 0
		DD	22831,	1024,	OFFSET fft1024_1, OFFSET fft1024_2
		DD			OFFSET fft1024_3, OFFSET fft1024_4
		DD			3*256+32, 15*256+8, 63*256+2, 127, 0
_TEXT32	ENDS

	ASSUME  CS: _TEXT32, DS: _TEXT32, SS: _TEXT32, ES: _TEXT32

assign_addrs	MACRO	sc0_size, sc1_size
sine_cosine0	EQU	20010000h
sine_cosine1	EQU	sine_cosine0 + sc0_size
norm_mults	EQU	sine_cosine1 + sc1_size
norm_aux1	EQU	norm_mults + 1024 * 24
sc1_gap		EQU	32
		ENDM

sc1_distances MACRO
	GAP = sc1_gap
	SCF = 0
	SCD = 48
	ENDM

INCLUDE	unravel.mac
INCLUDE	lucas1.mac
INCLUDE pfa.mac
INCLUDE mult.mac

	assign_addrs 16384, 8192
	dist1 =	8

_TEXT32	SEGMENT

; Setup routine

	PUBLIC	_gw_setup
_gw_setup PROC NEAR
	push	ebp
	push	ebx
	push	edi
	push	esi

; Initialize

	finit				; Init FPU

; Save p, compute fftlen

	mov	eax, _PARG		; p
	mov	p, eax

	mov	esi, OFFSET jmptable	; table to compute FFTLEN
flp:	mov	edx, [esi]		; load maximum p
	cmp	eax, edx		; is our p less than maximum p?
	jle	short fdn		; yes, we've found our fftlen
flp1:	cmp	DWORD PTR [esi], 0	; look for zero terminator
	lea	esi, [esi+4]		; next table entry
	jnz	short flp1
	jmp	short flp		; loop
fdn:	mov	ecx, [esi+4]		; Save FFTLEN
	mov	_FFTLEN, ecx
	mov	ecx, [esi+8]		; Save 4 fft routine offsets
	mov	fft_1_addr, ecx
	mov	ecx, [esi+12]
	mov	fft_2_addr, ecx
	mov	ecx, [esi+16]
	mov	fft_3_addr, ecx
	mov	ecx, [esi+20]
	mov	fft_4_addr, ecx
	mov	ecx, [esi+24]		; Save 4 counters
	mov	count1, ecx
	mov	ecx, [esi+28]
	mov	count2, ecx
	mov	ecx, [esi+32]
	mov	count3, ecx
	mov	ecx, [esi+36]
	mov	count4, ecx
	lea	esi, [esi+40]		; save table addr for later
	push	esi
	; Compute extra bits (the number of adds we can tolerate without
	; a normalization operation).  Studies show that exponents below
	; maxp - fftlen/2 - 4 can withstand one addition without normalization.
	sub	edx, eax
	mov	eax, _FFTLEN
	shr	eax, 1
	add	edx, eax
	sub	edx, 4
	mov	eax, edx
	sub	edx, edx
	div	_FFTLEN
	mov	extra_bits, eax

; Compute square root of 0.5

	fld	HALF
	fsqrt
	fstp	SQRTHALF

; Compute two-to-phi and two-to-minus-phi multipliers

	normalize_setup

; Compute the array of sine and cosine/sine values

	pfa_5_setup
	pfa_6_setup
	pfa_7_setup

	mov	esi, sine_cosine0	;; Address of sine/cosine data
	mov	edi, OFFSET sincos1	;; Address of table pointers
	mov	DWORD PTR [edi], sine_cosine1
	mov	DWORD PTR [edi+4], sine_cosine1
	mov	DWORD PTR [edi+8], sine_cosine1
	mov	DWORD PTR [edi+12], sine_cosine1
slp:	pop	ebp			;; Get next data set to generate
	mov	ebx, [ebp]		
	cmp	ebx, 0
	JZ_X	sdn			;; No more data to generate
	lea	ebp, [ebp+4]
	mov	[edi], esi
	push	ebp
	push	edi
	sc0_setup
	pop	edi
	lea	edi, [edi+4]
	JMP_X	slp			;; Look for more to generate
sdn:	sine_cosine_1_setup

; Return pointer to normalization multipliers

	mov	eax, norm_mults
	mov	_NORM_ARRAY, eax

; Return

	pop	esi
	pop	edi
	pop	ebx
	pop	ebp
	ret
_gw_setup ENDP


;; All the FFT routines for each FFT length

_gw_ffts PROC NEAR
fft32_2:	type2	32
fft32_3:	type3	32
fft32_1:	fft32	1
fft32_4:	fft32	4
fft40_2:	type2	40
fft40_3:	type3	40
fft40_1:	fft40	1
fft40_4:	fft40	4
fft48_2:	type2	48
fft48_3:	type3	48
fft48_1:	fft48	1
fft48_4:	fft48	4
fft56_2:	type2	56
fft56_3:	type3	56
fft56_1:	fft56	1
fft56_4:	fft56	4
fft64_2:	type2	64
fft64_3:	type3	64
fft64_1:	fft64	1
fft64_4:	fft64	4
fft80_2:	type2	80
fft80_3:	type3	80
fft80_1:	fft80	1
fft80_4:	fft80	4
fft96_2:	type2	96
fft96_3:	type3	96
fft96_1:	fft96	1
fft96_4:	fft96	4
fft112_2:	type2	112
fft112_3:	type3	112
fft112_1:	fft112	1
fft112_4:	fft112	4
fft128_2:	type2	128
fft128_3:	type3	128
fft128_1:	fft128	1
fft128_4:	fft128	4
fft160_2:	type2	160
fft160_3:	type3	160
fft160_1:	fft160	1
fft160_4:	fft160	4
fft192_2:	type2	192
fft192_3:	type3	192
fft192_1:	fft192	1
fft192_4:	fft192	4
fft224_2:	type2	224
fft224_3:	type3	224
fft224_1:	fft224	1
fft224_4:	fft224	4
fft256_2:	type2	256
fft256_3:	type3	256
fft256_1:	fft256	1
fft256_4:	fft256	4
fft320_2:	type2	320
fft320_3:	type3	320
fft320_1:	fft320	1
fft320_4:	fft320	4
fft384_2:	type2	384
fft384_3:	type3	384
fft384_1:	fft384	1
fft384_4:	fft384	4
fft448_2:	type2	448
fft448_3:	type3	448
fft448_1:	fft448	1
fft448_4:	fft448	4
fft512_2:	type2	512
fft512_3:	type3	512
fft512_1:	fft512	1
fft512_4:	fft512	4
fft640_2:	type2	640
fft640_3:	type3	640
fft640_1:	fft640	1
fft640_4:	fft640	4
fft768_2:	type2	768
fft768_3:	type3	768
fft768_1:	fft768	1
fft768_4:	fft768	4
fft896_2:	type2	896
fft896_3:	type3	896
fft896_1:	fft896	1
fft896_4:	fft896	4
fft1024_2:	type2	1024
fft1024_3:	type3	1024
fft1024_1:	fft1024	1
fft1024_4:	fft1024	4
_gw_ffts ENDP


;;
;; Perform an FFT in preparation for multiplying two numbers mod 2**p-1
;;

	PUBLIC	_gw_fft
_gw_fft PROC NEAR
	push	ebp
	push	ebx
	push	edi
	push	esi

; Do the FFT

	mov	esi, _SRCARG
	mov	ebp, _DESTARG
	sub	ebp, esi
	mov	eax, fft_1_addr
	jmp	eax
_gw_fft ENDP


;;
;; Square a number mod 2**p-1
;;

	PUBLIC	_gw_square
_gw_square PROC NEAR
	push	ebp
	push	ebx
	push	edi
	push	esi

; Do the FFT, squaring, and inverse FFT

	mov	esi, _DESTARG
	mov	eax, fft_2_addr
	jmp	eax
_gw_square ENDP


;;
;; Multiply two numbers mod 2**p-1.  One of the input numbers (SRCARG) must
;; have already been passed through gw_fft.
;;

	PUBLIC	_gw_mul
_gw_mul PROC NEAR
	push	ebp
	push	ebx
	push	edi
	push	esi

; Do the FFT, multiply, and inverse FFT

	mov	esi, _DESTARG
	mov	ebp, _SRCARG
	sub	ebp, esi
	mov	eax, fft_3_addr
	jmp	eax

; Code to check for how close we are to incurring a round-off error

echk:	fld	_MAXERR			; Load old maximum error
	mov	ebx, norm_mults		; Address of the multipliers
	mov	esi, _DESTARG		; Address of multiplied number
	mov	ecx, _FFTLEN		; Loop counter
elp:	error_check			; Error check eight values
	lea	esi, [esi+dist1]	; Next source
	lea	ebx, [ebx+24]		; Next multipliers
	dec	ecx
	jnz	short elp		; Loop til done
	fstp	_MAXERR			; Store maximum error

; Normalize the results

noechk:	fldz				; Init SUMOUT
	fld	BIGVAL			; Start process with no carry
	mov	ebx, norm_mults		; Address of the multipliers
	mov	esi, _DESTARG		; Address of multiplied number
	mov	edi, scaled_numlit	; Used to compute big vs little words
	mov	ebp, scaled_numbig
	mov	eax, -1			; First word is a big word (-1)
	mov	ecx, _FFTLEN		; Load loop counter
nlp:	normalize			; Normalize 8 values
	lea	esi, [esi+8*dist1]	; Next source
	lea	ebx, [ebx+8*24]		; Next set of 8 multipliers
	sub	ecx, 8			; Test loop counter
	JNZ_X	nlp			; Loop til done
					; Add any carry to first values
	mov	esi, _DESTARG		; Address of multiplied number
	mov	ebx, norm_mults		; Restart the multipliers
	normalize_012
	mov	DWORD PTR [esi-4], 0	; Clear needs-normalize counter

; Normalize SUMOUT value by multiplying by 1 / (n/2).
; This is the value of the first two-to-minus-phi multiplier

	fmul	QWORD PTR ds:(norm_mults)
	fadd	_SUMOUT
	fstp	_SUMOUT

; Return

exit:	pop	esi
	pop	edi
	pop	ebx
	pop	ebp
	ret
_gw_mul ENDP


;;
;; Multiply two numbers mod 2**p-1.  Both of the input numbers must
;; have already been passed through gw_fft.
;;

	PUBLIC	_gw_mulf
_gw_mulf PROC NEAR
	push	ebp
	push	ebx
	push	edi
	push	esi

; Do the multiply and inverse FFT

	mov	esi, _DESTARG
	mov	ebp, _SRCARG
	mov	ebx, _SRC2ARG
	mov	eax, fft_4_addr
	sub	ebp, esi
	sub	ebx, esi
	jmp	eax
_gw_mulf ENDP


;;
;; Copy a number with maximal pipelining
;;

	PUBLIC	_gw_copy
_gw_copy PROC NEAR
	push	esi
	mov	esi, _SRCARG		; Address of first number
	push	edi
	mov	edi, _DESTARG		; Address of second number
	mov	eax, [esi-4]		; Get needs-normalize counter
	mov	ecx, _FFTLEN		; Load loop counter
	mov	[edi-4], eax		; Store needs-normalize counter
copylp:	mov	eax, [esi][ecx*8-32]
	mov	edx, [esi][ecx*8-28]
	mov	[edi][ecx*8-32], eax
	mov	[edi][ecx*8-28], edx
	mov	eax, [esi][ecx*8-24]
	mov	edx, [esi][ecx*8-20]
	mov	[edi][ecx*8-24], eax
	mov	[edi][ecx*8-20], edx
	mov	eax, [esi][ecx*8-16]
	mov	edx, [esi][ecx*8-12]
	mov	[edi][ecx*8-16], eax
	mov	[edi][ecx*8-12], edx
	mov	eax, [esi][ecx*8-8]
	mov	edx, [esi][ecx*8-4]
	mov	[edi][ecx*8-8], eax
	mov	[edi][ecx*8-4], edx
	sub	ecx, 4			; Check loop counter
	jnz	short copylp		; Loop if necessary
	pop	edi
	pop	esi
	ret
_gw_copy ENDP


;;
;; Add two numbers without carry propogation.  Caller can use this for
;; consecutive add or subtract operations.  However, the last operation
;; before a multiply must use the routine that will normalize data.
;;

	PUBLIC	_gw_addq
_gw_addq PROC NEAR
	push	esi			; U
	mov	esi, _SRCARG		; V - Address of first number
	push	edi			; U
	mov	edi, _SRC2ARG		; V - Address of second number
	mov	eax, [esi-4]		; U - Load needs-normalize counter
	push	ebp			; V
	inc	eax			; U - Bump needs-normalize counter
	mov	edx, [edi-4]		; V - Load needs-normalize counter
	mov	ebp, _DESTARG		; U - Address of destination
	add	eax, edx		; V - Add needs-normalize counters
	jmp	short uadd		; Do the unnormalized add
_gw_addq ENDP


;;
;; Add two numbers.  This will be done without carry propogation
;; (for extra speed) if it can be done safely.
;;

	PUBLIC	_gw_add
_gw_add PROC NEAR

; See if add can be done without normalization

	push	esi			; U
	mov	esi, _SRCARG		; V - Address of first number
	push	edi			; U
	mov	edi, _SRC2ARG		; V - Address of second number
	mov	eax, [esi-4]		; U - Load needs-normalize counter
	push	ebp			; V
	inc	eax			; U - Bump needs-normalize counter
	mov	edx, [edi-4]		; V - Load needs-normalize counter
	mov	ebp, _DESTARG		; U - Address of destination
	add	eax, edx		; V - Add needs-normalize counters
	cmp	eax, extra_bits		;UV - Is normalization needed?
	jg	short nadd		; V - yes

; Do an unnormalized add

uadd:	mov	ecx, _FFTLEN		; U - Load loop counter
	mov	[ebp-4], eax		; V - Store needs-normalize counter
					; AGI STALL
uaddlp:	fld	QWORD PTR [esi][ecx*8-32] ; Load first number
	fadd	QWORD PTR [edi][ecx*8-32] ; Add in second number
	fld	QWORD PTR [esi][ecx*8-24] ; Load first number
	fadd	QWORD PTR [edi][ecx*8-24] ; Add in second number
	fld	QWORD PTR [esi][ecx*8-16] ; Load first number
	fadd	QWORD PTR [edi][ecx*8-16] ; Add in second number
	fxch	st(1)
	fld	QWORD PTR [esi][ecx*8-8]  ; Load first number
	fadd	QWORD PTR [edi][ecx*8-8]  ; Add in second number
	fxch	st(3)
	fstp	QWORD PTR [ebp][ecx*8-32] ; Save result
	fstp	QWORD PTR [ebp][ecx*8-24] ; Save result
	fstp	QWORD PTR [ebp][ecx*8-16] ; Save result
	fstp	QWORD PTR [ebp][ecx*8-8]  ; Save result
	sub	ecx, 4			; Check loop counter
	jnz	short uaddlp		; Loop if necessary
	pop	ebp
	pop	edi
	pop	esi
	ret

; Do a normalized add

nadd:	push	ebx			; Save registers for C compiler
	fld	BIGVAL			; Start process with no carry
	mov	ebx, norm_mults		; Address of the multipliers
	mov	edx, scaled_numlit	; Used to compute big vs little words
	mov	eax, -1			; First word is a big word (-1)
	mov	ecx, _FFTLEN		; Loop counter
naddlp:	norm_op	fadd			; Add and normalize 8 values
	sub	ecx, 8			; Decrement loop counter
	JNZ_X	naddlp			; Loop til done
	fsub	BIGVAL
	mov	ecx, _DESTARG		; Address of squared number
	fadd	QWORD PTR [ecx]		; Add in carry
	mov	DWORD PTR [ecx-4], 0	; Clear needs-normalize flag
	pop	ebx
	pop	ebp
	pop	edi
	pop	esi
	fstp	QWORD PTR [ecx]		; Save new value
	ret
_gw_add ENDP


;;
;; Subtract two numbers without carry propogation.  Caller can use this for
;; consecutive add or subtract operations.  However, the last operation
;; before a multiply must use the routine that will normalize data.
;;

	PUBLIC	_gw_subq
_gw_subq PROC NEAR
	push	esi			; U
	mov	esi, _SRCARG		; V - Address of first number
	push	edi			; U
	mov	edi, _SRC2ARG		; V - Address of second number
	mov	eax, [esi-4]		; U - Load needs-normalize counter
	push	ebp			; V
	inc	eax			; U - Bump needs-normalize counter
	mov	edx, [edi-4]		; V - Load needs-normalize counter
	mov	ebp, _DESTARG		; U - Address of destination
	add	eax, edx		; V - Add needs-normalize counters
	jmp	short usub		; Do the unnormalized add
_gw_subq ENDP

;;
;; Subtract two numbers.  This will be done without carry propogation
;; (for extra speed) if it can be done safely.
;;

	PUBLIC	_gw_sub
_gw_sub PROC NEAR

; See if subtract can be done without normalization

	push	esi			; U
	mov	esi, _SRCARG		; V - Address of first number
	push	edi			; U
	mov	edi, _SRC2ARG		; V - Address of second number
	mov	eax, [esi-4]		; U - Load needs-normalize counter
	push	ebp			; V
	inc	eax			; U - Bump needs-normalize counter
	mov	edx, [edi-4]		; V - Load needs-normalize counter
	mov	ebp, _DESTARG		; U - Address of destination
	add	eax, edx		; V - Add needs-normalize counters
	cmp	eax, extra_bits		;UV - Is normalization needed?
	jg	short nsub		; V - yes

; Do an unnormalized subtract

usub:	mov	ecx, _FFTLEN		; Load loop counter
	mov	[ebp-4], eax		; V - Store needs-normalize counter
					; AGI STALL
usublp:	fld	QWORD PTR [edi][ecx*8-32] ; Load second number
	fsub	QWORD PTR [esi][ecx*8-32] ; Subtract first number
	fld	QWORD PTR [edi][ecx*8-24] ; Load second number
	fsub	QWORD PTR [esi][ecx*8-24] ; Subtract first number
	fld	QWORD PTR [edi][ecx*8-16] ; Load second number
	fsub	QWORD PTR [esi][ecx*8-16] ; Subtract first number
	fxch	st(1)
	fld	QWORD PTR [edi][ecx*8-8]  ; Load second number
	fsub	QWORD PTR [esi][ecx*8-8]  ; Subtract first number
	fxch	st(3)
	fstp	QWORD PTR [ebp][ecx*8-32] ; Save result
	fstp	QWORD PTR [ebp][ecx*8-24] ; Save result
	fstp	QWORD PTR [ebp][ecx*8-16] ; Save result
	fstp	QWORD PTR [ebp][ecx*8-8]  ; Save result
	sub	ecx, 4			; Check loop counter
	jnz	short usublp		; Loop if necessary
	pop	ebp
	pop	edi
	pop	esi
	ret

; Do a normalized subtract

nsub:	push	ebx			; Save registers for C compiler
	fld	BIGVAL			; Start process with no carry
	mov	ebx, norm_mults		; Address of the multipliers
	mov	edx, scaled_numlit	; Used to compute big vs little words
	mov	eax, -1			; First word is a big word (-1)
	mov	ecx, _FFTLEN		; Loop counter
nsublp:	norm_op	fsub			; Subtract and normalize 8 values
	sub	ecx, 8			; Decrement loop counter
	JNZ_X	nsublp			; Loop til done
	fsub	BIGVAL
	mov	ecx, _DESTARG		; Address of squared number
	fadd	QWORD PTR [ecx]		; Add in carry
	mov	DWORD PTR [ecx-4], 0	; Clear needs-normalize flag
	pop	ebx
	pop	ebp
	pop	edi
	pop	esi
	fstp	QWORD PTR [ecx]		; Save new value
	ret
_gw_sub ENDP

_TEXT32	ENDS
END
