/*
 * Common routines and variables used by Prime95, Saver95, and NTPrime
 *
 * Comm95a contains information used only during setup
 * Comm95b contains information used only during execution
 * Comm95c contains information used during setup and execution
 */ 

/* Common routines */

LPVOID memreserved = NULL;

int reserveMem ()
{
	freeMem ();

// Allocate memory for the assembly code

	memreserved = VirtualAlloc ((void *) DATA_ADDR,
				    THE_END - DATA_ADDR,
				    MEM_RESERVE | MEM_COMMIT,
				    PAGE_READWRITE);
	if (memreserved == NULL) {
		OutputBoth ("Error allocating memory.\n");
		return (FALSE);
	}
	return (TRUE);
}

// Free memory used in the assembly code

void freeMem ()
{
	if (memreserved == NULL) return;
	if (!VirtualFree (memreserved,0, MEM_RELEASE))
		OutputBoth ("Error freeing memory.\n");
	memreserved = NULL;
}

// Reserve memory for the factoring code

int loadfac ()
{
	THE_END = DATA_ADDR + 0x00100000;	// 1MB should be plenty
	return (reserveMem ());
}

// Start initializing for a Lucas-Lehmer test

int loadRightExe (
	unsigned long p,
	unsigned long *fftlen)
{

// Find the best fft length -- the smallest one that generates a
// maximum round-off error of 0.30.

	if (*fftlen == 0) *fftlen = pick_fft_length (p);

// The fft length determines the addresses.

	if (*fftlen == 1048576) THE_END = THE_END_1024, lucas = lucas1024;
	if (*fftlen == 917504) THE_END = THE_END_896, lucas = lucas896;
	if (*fftlen == 786432) THE_END = THE_END_768, lucas = lucas768;
	if (*fftlen == 655360) THE_END = THE_END_640, lucas = lucas640;
	if (*fftlen == 524288) THE_END = THE_END_512, lucas = lucas512;
	if (*fftlen == 458752) THE_END = THE_END_448, lucas = lucas448;
	if (*fftlen == 393216) THE_END = THE_END_384, lucas = lucas384;
	if (*fftlen == 327680) THE_END = THE_END_320, lucas = lucas320;
	if (*fftlen == 262144) THE_END = THE_END_256, lucas = lucas256;
	if (*fftlen == 229376) THE_END = THE_END_224, lucas = lucas224;
	if (*fftlen == 196608) THE_END = THE_END_192, lucas = lucas192;
	if (*fftlen == 163840) THE_END = THE_END_160, lucas = lucas160;
	if (*fftlen == 131072) THE_END = THE_END_128, lucas = lucas128;
	if (*fftlen == 114688) THE_END = THE_END_112, lucas = lucas112;
	if (*fftlen == 98304) THE_END = THE_END_96, lucas = lucas96;
	if (*fftlen == 81920) THE_END = THE_END_80, lucas = lucas80;

// Now we know how much memory to reserve

	if (! reserveMem ()) return (FALSE);

// Done

	return (TRUE);
}

// Load the FFT data and call setup

void lucasSetup (
	unsigned long p,
	unsigned long fftlen)
{
	PARG = p;
	if (fftlen == 1048576) setup1024 ();
	if (fftlen == 917504) setup896 ();
	if (fftlen == 786432) setup768 ();
	if (fftlen == 655360) setup640 ();
	if (fftlen == 524288) setup512 ();
	if (fftlen == 458752) setup448 ();
	if (fftlen == 393216) setup384 ();
	if (fftlen == 327680) setup320 ();
	if (fftlen == 262144) setup256 ();
	if (fftlen == 229376) setup224 ();
	if (fftlen == 196608) setup192 ();
	if (fftlen == 163840) setup160 ();
	if (fftlen == 131072) setup128 ();
	if (fftlen == 114688) setup112 ();
	if (fftlen == 98304) setup96 ();
	if (fftlen == 81920) setup80 ();
}

// Each FFT word is multiplied by a two-to-phi value.  These
// routines set and get the FFT value without the two-to-phi
// multiplier.

void get_fft_value (
	unsigned long p,
	unsigned long fftlen,
	unsigned long i,
	long	*retval)
{
	double	*dblp, tmp, pow;
	unsigned long b, s;

// Get address of fft value

	dblp = (double *) addr (p, fftlen, i);

// The first FFT word is not multiplied by anything.

	if (i == 0) {
		*retval = (long) *dblp;
		return;
	}

// Compute the number of big FFT words.

	b = p % fftlen;

// Compute the number of small FFT words.

	s = fftlen - b;

// Divide by two-to-phi to generate an integer.
// Do not trust the pow routine.  The assembly code will
// generate the value with greater accuracy.

	tmp = *dblp;
	pow = (double) mulmod (s, i, fftlen) / (double) fftlen;
	__asm fld tmp
	__asm fld pow
	__asm f2xm1
	__asm fld1
	__asm faddp st(1), st
	__asm fdivp st(1), st
	__asm frndint
	__asm fstp tmp
	*retval = (long) tmp;
}

void set_fft_value (
	unsigned long p,
	unsigned long fftlen,
	unsigned long i,
	long	val)
{
	double	*dblp, tmp, pow;
	unsigned long b, s;

// Get address of fft value

	dblp = (double *) addr (p, fftlen, i);

// The first FFT word is not multiplied by anything.

	if (i == 0) {
		*dblp = val;
		return;
	}

// Compute the number of big FFT words.

	b = p % fftlen;

// Compute the number of small FFT words.

	s = fftlen - b;

// Multiply value by two-to-phi before setting fft value.
// Do not trust the pow routine.  The assembly code will
// generate the value with greater accuracy.

	pow = (double) mulmod (s, i, fftlen) / (double) fftlen;
	__asm fld pow
	__asm f2xm1
	__asm fld1
	__asm faddp st(1), st
	__asm fimul val
	__asm fstp tmp
	*dblp = tmp;
}
