/*----------------------------------------------------------------------
| This file contains routines and global variables that are common for
| all operating systems the program has been ported to.  It is included
| in one of the source code files of each port.  See common.h for the
| common #defines and common routine definitions.
|
| Commona contains information used only during setup
| Commonb contains information used only during execution
| Commonc contains information used during setup and execution
+---------------------------------------------------------------------*/

/* Global variables */

int STARTUP_IN_PROGRESS = 0;

/* Routine to eliminate odd puctuation characters from user ID */
/* and computer ID */

void sanitizeString (
	char	*p)
{
	int	i;
	for (i = strlen (p); i > 0 && isspace (p[i-1]); i--) p[i-1] = 0;
	while (*p) {
		if (!IsCharAlphaNumeric (*p) &&
		    *p != '.' && *p != '-' && *p != '_')
			*p = '_';
		p++;
	}
}

/* Create a status report message from the work-to-do file */

#define STAT0 "Below is a report on the work you have queued and any expected completion dates.\n"
#define STAT1 "The chance that one of the %d exponents you are testing will yield a Mersenne prime is about 1 in %ld. "
#define STAT1a "The chance that the exponent you are testing will yield a Mersenne prime is about 1 in %ld. "
#define STAT3 "You have no work queued up."

void rangeStatusMessage (
	char	*buf)		/* 2000 character buffer */
{
	unsigned int i, ll_cnt;
	double	prob, est;

	ll_cnt = 0;
	prob = 0.0;
	est = 0.0;
	strcpy (buf, STAT0);
	IniFileOpen (WORKTODO_FILE, 0);
	for (i = 1; ; i++) {
		struct work_unit w;
		unsigned long iteration;
		double	work;

/* Read the line of the work file */

		if (!parseWorkToDoLine (i, &w)) break;

/* If testing then adjust our probabilities */
/* This assumes our error rate is roughly 0.5% */

		if (w.bits < 32) w.bits = 32;
		if (w.work_type == WORK_TEST) {
			ll_cnt++;
			prob += (double) ((w.bits - 1) * 2) / w.p;
		}
		if (w.work_type == WORK_DBLCHK) {
			ll_cnt++;
			prob += (double) ((w.bits - 1) * 2) / w.p * 0.005;
		}

/* Adjust our time estimate */

		work = work_estimate (&w);
		work *= (1.0 - pct_complete (w.work_type, w.p, &iteration));
		est += work;

/* Add the exponent to the output message */

		if (strlen(buf) < 1800) {
			char	timebuf[30];
			char	*work_type;

			if (w.work_type == WORK_ECM)
				work_type = "ECM";
			if (w.work_type == WORK_TEST ||
			    w.work_type == WORK_ADVANCEDTEST)
				work_type = "Lucas-Lehmer test";
			if (w.work_type == WORK_FACTOR)
				work_type = "Factoring";
			if (w.work_type == WORK_DBLCHK)
				work_type = "Double-checking";

			if (w.work_type == WORK_ECM)
				strcpy (timebuf, "completion date unknown\n");
			else {
				time_t	this_time;
				time (&this_time);
				this_time += (long) est;
				strcpy (timebuf, ctime (&this_time)+4);
				strcpy (timebuf+12, timebuf+15);
			}

			sprintf (buf+strlen(buf), "M%ld - %s - %s", w.p, work_type, timebuf);
		}
	}
	IniFileClose (WORKTODO_FILE);

/* Format more of the message */

	if (est == 0.0) {
		strcpy (buf, STAT3);
		return;
	}

/* Make a pretty message */

	if (ll_cnt == 1)
		sprintf (buf+strlen(buf), STAT1a, (long) (1.0 / prob));
	if (ll_cnt > 1)
		sprintf (buf+strlen(buf), STAT1, ll_cnt, (long) (1.0 / prob));
}
