/*----------------------------------------------------------------------
| This file contains routines and global variables that are common for
| all operating systems the program has been ported to.  It is included
| in one of the source code files of each port.  See common.h for the
| common #defines and common routine definitions.
|
| Commona contains information used only during setup
| Commonb contains information used only during execution
| Commonc contains information used during setup and execution
+---------------------------------------------------------------------*/

/* Macros */

#undef RDTSC
#if defined (__linux__) || defined (__EMX__)
#define RDTSC(x) {__asm__ __volatile__ (".byte 0x0F; .byte 0x31" : "=a" (x) : : "%eax", "%edx");}
#elif defined (WIN31)
#define RDTSC(x) { __asm _emit 0x0F __asm _emit 0x31 __asm _emit 0x66 __asm mov WORD PTR x, ax }
#else
#define RDTSC(x) { __asm __emit 0x0F __asm __emit 0x31 __asm mov x, eax }
#endif

/* Global variables */

int	TIMED_RUN = 0;
time_t	END_TIME;

double	*ECM_ADDR = NULL;		/* Hack for ECM code to use addr func*/
long	THE_END;			/* End of the used address space */
EXTERNC unsigned long FACPASS = 0;
EXTERNC unsigned long FACMSW = 0;
EXTERNC unsigned long FACLSW = 0;
EXTERNC unsigned long PARG = 0;
EXTERNC double MAXERR = 0.0;
EXTERNC double SUMINP = 0.0;
EXTERNC double SUMOUT = 0.0;
void (*lucas)();

char ERRMSG0[] = "Iteration: %ld/%ld, %s";
char ERRMSG1A[] = "ERROR: ILLEGAL SUMOUT\n";
char ERRMSG1B[] = "ERROR: SUM(INPUTS) != SUM(OUTPUTS), %.16g != %.16g\n";
char ERRMSG1C[] = "ERROR: ROUND OFF (%.10g) > 0.40\n";
char ERRMSG2[] = "Possible hardware failure, consult the readme file.\n";
char ERRMSG3[] = "Continuing from last save file.\n";
char ERRMSG4[] = "Waiting five minutes before restarting.\n";
char ERROK[] = "Disregard last error.  Result is reproducible and thus not a hardware problem.\n";
char WRITEFILEERR[] = "Error writing intermediate file: %s\n";
char DONE_MSG1[] = "There are no more exponents to test in this range.\n";
char DONE_MSG2[] = "Please send the results.txt file to woltman@magicnet.net\n";
char DONE_MSG3[] = "Select a new range from http://www.mersenne.org/range.htm!\n";
char RENAME_MSG[] = "Renaming intermediate file %s to %s.\n";

/* Return true is exponent yields a known Mersenne prime */

int isKnownMersennePrime (
	unsigned long p)
{
	return (p == 2 || p == 3 || p == 5 || p == 7 || p == 13 || p == 17 ||
		p == 19 || p == 31 || p == 61 || p == 89 || p == 107 ||
		p == 127 || p == 521 || p == 607 || p == 1279 || p == 2203 ||
		p == 2281 || p == 3217 || p == 4253 || p == 4423 ||
		p == 9689 || p == 9941 || p == 11213 || p == 19937 ||
		p == 21701 || p == 23209 || p == 44497 || p == 86243 ||
		p == 110503 || p == 132049 || p == 216091 || p == 756839 ||
		p == 859433 || p == 1257787 || p == 1398269 || p == 2976221 ||
		p == 3021377);
}

/* In the self-test we initialize the FFT data with zero. */

void clear_memory (
	unsigned long p,
	unsigned long fftlen)
{
	unsigned long k;
	for (k = 0; k < fftlen; k++) set_fft_value (p, fftlen, k, 0);
}


/* Make a string out of a 64-bit value (a found factor) */

void makestr (
	unsigned long msw,
	unsigned long lsw,
	char	*buf)			/* An 80 character output buffer */
{
	int	i, j, k, carry;
	unsigned long x[2];
	char	pow[80];

	x[0] = msw; x[1] = lsw;
	for (i = 0; i < 79; i++) pow[i] = '0', buf[i] = '0';
	pow[78] = '1';
	pow[79] = buf[79] = 0;

	for (i = 2; i--; ) {
		for (j = 0; j < 32; j++) {
			if (x[i] & 1) {
				carry = 0;
				for (k = 79; k--; ) {
					buf[k] = buf[k] - '0' +
						pow[k] - '0' + carry;
					carry = buf[k] / 10;
					buf[k] %= 10;
					buf[k] += '0';
				}
			}
			carry = 0;
			for (k = 79; k--; ) {
				pow[k] = (pow[k] - '0') * 2 + carry;
				carry = pow[k] / 10;
				pow[k] %= 10;
				pow[k] += '0';
			}
			x[i] >>= 1;
		}
	}
	while (buf[0] == '0') strcpy (buf, buf+1);
}

/* Generate the 64-bit residue of a Lucas-Lehmer test */

void generateResidue64 (
	unsigned long p,
	unsigned long fftlen,
	unsigned long *reshi,
	unsigned long *reslo)
{
	unsigned long i, high32, low32;
	long	val;
	int	j, bits, bitsout, carry;

/* Check for a carry out of the most significant word */

	for (i = fftlen-1; ; i--) {
		get_fft_value (p, fftlen, i, &val);
		if (val != 0) {
			carry = (val > 0) ? 0 : -1;
			break;
		}
	}

/* Collect bits until we have 64 of them */

	high32 = low32 = 0;
	bitsout = 0;
	for (i = 0; i < fftlen && bitsout < 64; i++) {
		get_fft_value (p, fftlen, i, &val);
		bits = (int) (p / fftlen);
		if (is_big_word (p, fftlen, i)) bits++;
		val += carry;
		for (j = 0; j < bits && bitsout < 64; j++) {
			low32 >>= 1;
			if (high32 & 1) low32 += 0x80000000;
			high32 >>= 1;
			if (val & 1) high32 += 0x80000000;
			val >>= 1;
			bitsout++;
		}
		carry = (int) val;
	}

/* Return the result */

	*reshi = high32;
	*reslo = low32;
}

/* To optimize use of the L1 cache we scramble the FFT data as follows. */
/* Note:  The Intel L1 cache has a 32 byte cache line. */

/* 1) Each cache line contains data that is used by both pass 1 and pass 2. */
/*    For example, when N=16, you have 65536 values.  9 levels of the FFT */
/*    are performed in pass 1 and 7 levels of FFT are performed in pass 2. */
/*    Each cache line contains data that is 2^16 / 2^9 = 2^7 apart.  That */
/*    is, the first cache line contains the 0th, 128th, 256th, and 384th */
/*    FFT data values. */

/* 2) After each 512 data values an extra 64 bytes are wasted.  This allows */
/*    pass 1 to operate efficiently.  If we did not do this every pass 1 */
/*    value would try to occupy the same cache line! */

/* Putting it all together, for N=16 you get this memory layout:	*/
/*	384	256	128	0		(32 bytes)		*/
/*	385	257	129	1		(32 bytes)		*/
/*		   etc.							*/
/*	511	383	255	127		(32 bytes)		*/
/*		(64 wasted bytes)					*/
/*	896	768	640	512		(32 bytes)		*/
/*	897	769	641	513		(32 bytes)		*/
/*		   etc.							*/
/*	1023	895	767	639		(32 bytes)		*/
/*		(64 wasted bytes)					*/

/* Well.... I implemented the above only to discover I had dreadful */
/* performance in pass 1.  How can that be?  The problem is that each  */
/* cache line in pass 1 comes from a different logical 4K page.  Therefore, */
/* pass 1 accesses 128 different pages.  This is a problem because the */
/* Pentium chip has only 64 TLBs (translation lookaside buffers) to map */
/* logical page addresses into physical addresses.  So we need to shuffle */
/* the data even further. */

/*1st 128 cache lines	2nd through 8th 128 lines   9th 128 cache lines */
/* 384 256 128 0		16*512		    392 264 136 8	*/
/* 385 257 129 1						9	*/
/*             2						10	*/
/*             3						11	*/
/*             4						12	*/
/*             5		  ......			13	*/
/*             6						14	*/
/*             7						15	*/
/* 896 768 640 512						520	*/
/*             513							*/
/*             ...							*/
/*             15*512+7			127*512+7		15*512+15 */

/* As before, there are 64 bytes wasted between each set of 128 cache lines. */
/* If you look carefully at the above, you'll see that in pass 1 the */
/* FFT data will come from just 16 4K pages.  Similarly, the pass 2 data */
/* comes from 16 different 4K pages. */

/* The above is *still* slow!  I changed to wasting 7*64 bytes so */
/* that there are fewer 64K distances.  The 64K distance is really */
/* bad for the on-chip TLB cache. */

/* I once had the brilliant idea of interleaving the sin/cos data */
/* with the FFT data.  That is, the data occupies the even cache */
/* lines and the sin/cos data is in the odd cache lines.  At first */
/* this seems counter productive, as only 4K of FFT data will now fit */
/* in the 8K L1 cache.  However, if you look at how an FFT operates */
/* you'll see loading FFT data, multiply by sin/cos data, store FFT data, */
/* load next block of FFT data, multiply by sin/cos data, store FFT data, */
/* etc.  By storing the sin/cos data in the odd cache lines, loading the */
/* next block of FFT data will toss out the previous block of FFT data */
/* rather than the reusable sin/cos data. */
/* For some reason, however, interleaving resulted in slower performance. */
/* Consequently, this #define is set to one to disable interleaving. */
#define INTERLEAVE		1

long addr (unsigned long p, unsigned long fftlen, unsigned long i)
{
	unsigned long addr, pass1, pass2, col, section, section_size;

/* A hack that lets the ECM code share all the functions that */
/* call this routine.  ECM code uses arrays that are not scrambled. */

	if (ECM_ADDR) return ((long) (ECM_ADDR + i));

/* Break the element number into parts for scrambling */

	pass1 = i >> 9;				/* Top bits */
	col = (i >> 7) & 3;			/* Middle 2 bits */
 	pass2 = i & 127;			/* Bottom 7 bits */

/* Compute the lower part of the address, inserting waste bytes. */
/* Waste is 1 cache line every 128 cache lines or 32 bytes per 4K page. */
/* Addr is: pass1 || pass2 % 8 || optional-interleave-bit || col || 000 */

	addr = pass1;
	addr = (addr << 3) + (pass2 & 7);
	addr = (addr >> 7) * (128 + 1) + (addr & 127);  /* Add waste */
	addr *= INTERLEAVE;
	addr = ((addr << 2) + col) << 3;

/* Compute the upper part of the address adding more waste bytes */
/* to avoid 64KB spacing.  Note there is no guarantee that the OS */
/* will place our 4K pages linearly in memory, but there are times */
/* that the OS does, so we'll optimize for that case. */

	section = pass2 >> 3;
	section_size = (fftlen >> 4) * sizeof (double);
	section_size += 4096 + 8*32;
	section_size *= INTERLEAVE;
	addr += section * section_size;

/* Return the full address */

	return (DATA_ADDR + addr);
}

/* Routine to handle a multiplication and modulo operation where */
/* the intermediate multiplication result can be more than 32 bits. */

unsigned long mulmod (
	unsigned long a,
	unsigned long b,
	unsigned long c)
{
	unsigned long tmp;
	tmp = a * (b & 0x3FF);
	tmp += ((a << 10) % c) * (b >> 10);
	return (tmp % c);
}

/* Some words in the FFT data contain floor(p/N), some words contain */
/* floor(p/N)+1 bits.  This function returns TRUE in the latter case. */

int is_big_word (
	unsigned long p,
	unsigned long fftlen,
	unsigned long i)
{
	unsigned long b, s, t;

/* Compute the number of big FFT words. */

	b = p % fftlen;

/* Compute the number of small FFT words. */

	s = fftlen - b;

/* The rest are uniformly distributed. Use special */
/* arithmetic to avoid overflows in the b * i operation. */

	t = mulmod (b, i, fftlen);
	return (t == 0 || t > s);
}

/* Return TRUE if a continuation file exists.  If one does exist, */
/* make sure it is named pXXXXXXX. */

int continuationFileExists (
	char	*filename)
{
	char	backupname[32];
	char	buf[80];

	if (fileExists (filename)) return (TRUE);
	strcpy (backupname, filename);
	backupname[0] = 'q';
	if (fileExists (backupname)) {
		sprintf (buf, RENAME_MSG, backupname, filename);
		OutputBoth (buf);
		rename (backupname, filename);
		return (TRUE);
	}
	backupname[0] = 'r';
	if (fileExists (backupname)) {
		sprintf (buf, RENAME_MSG, backupname, filename);
		OutputBoth (buf);
		rename (backupname, filename);
		return (TRUE);
	}
	return (FALSE);
}

/* Write intermediate Lucas-Lehmer results to a file */
/* Note: we can't output the bits == 16 case in two bytes because the */
/* rounding code in lucas.asm can create a value of 32768 which does*/
/* not fit in a short. */

int writeToFile (
	unsigned long p,
	char	*filename,
	unsigned long fftlen,
	unsigned long counter)
{
	int	fd, bits;
	unsigned long i;
	char	buf[4096];
	char	*bufp;
	short	type;
	long	sum = 0;

/* If we are allowed to create multiple intermediate files, then */
/* write to a file called rXXXXXXX. */

	if (TWO_BACKUP_FILES && strlen (filename) == 8)
		filename[0] = 'r';

/* Now save to the intermediate file */

	fd = _open (filename, _O_BINARY | _O_WRONLY | _O_TRUNC | _O_CREAT, 0666);
	if (fd < 0) return (FALSE);
	type = (short) (fftlen / 1024);
	if (_write (fd, &type, sizeof (short)) != sizeof (short))
		goto writeerr;
	if (_write (fd, &counter, sizeof (long)) != sizeof (long))
		goto writeerr;

	bits = (int) (p / fftlen + 1);
	bufp = buf;
	for (i = 0; ; i++) {
		long	x;
		if (bufp == buf + sizeof (buf)) {
			if (_write (fd, buf, sizeof (buf)) != sizeof (buf))
				goto writeerr;
			bufp = buf;
		}
		if (i == fftlen) break;
		get_fft_value (p, fftlen, i, &x);
		if (bits <= 15) {
			short y;
			y = (short) x;
			memcpy (bufp, &y, sizeof (y));
			bufp += sizeof (y);
  		} else {
			long y;
			y = (long) x;
			memcpy (bufp, &y, sizeof (y));
			bufp += sizeof (y);
		}
		sum += x;
	}
	memcpy (bufp, &sum, sizeof (sum));
	bufp += sizeof (sum);
	if (_write (fd, buf, bufp - buf) != (bufp - buf)) goto writeerr;
	_commit (fd);
	_close (fd);

/* Now rename the intermediate files */

	if (TWO_BACKUP_FILES && strlen (filename) == 8) {
		char	backupname[16];
		strcpy (backupname, filename);
		backupname[0] = 'q'; filename[0] = 'p';
		_unlink (backupname);
		 rename (filename, backupname);
		backupname[0] = 'r';
		rename (backupname, filename);
	}

	return (TRUE);

/* An error occured.  Delete the current file and rename the backup */
/* intermediate file */

writeerr:
	_close (fd);
	_unlink (filename);
	return (FALSE);
}

/* Read the data portion of an intermediate Lucas-Lehmer results file */

int readFileData (
	int	fd,
	unsigned long p,
	unsigned long fftlen)
{
	unsigned long i;
	long	sum, filesum;
	int	bits, zero;		/* Guard against a zeroed out file */

	if (fd < 0) return (FALSE);

	bits = (int) (p / fftlen + 1);
	sum = 0;
	zero = TRUE;
	for (i = 0; i < fftlen; i++) {
		long	x;
		if (bits <= 15) {
			short y;
			if (_read (fd, &y, sizeof (y)) != sizeof (y))
				goto err;
			x = y;
		} else {
			long y;
			if (_read (fd, &y, sizeof (y)) != sizeof (y))
				goto err;
			x = y;
		}
		sum += x;
		if (x) zero = FALSE;
		set_fft_value (p, fftlen, i, x);
	}
	if (_read (fd, &filesum, sizeof (filesum)) != 0) {
		if (filesum != sum) goto err;
	}
	if (zero) goto err;
	_close (fd);
	return (TRUE);
err:	_close (fd);
	return (FALSE);
}

/* Check if any of the Lucas-Lehmer test lines also require factoring. */
/* This will force factoring to be done first - giving us more accurate */
/* estimates of how much work is queued up. */

int getPriorityWork () {
	unsigned int i;
	struct work_unit w;

	for (i = 1; ; i++) {
		if (!parseWorkToDoLine (i, &w)) break;
		if (w.work_type == WORK_ADVANCEDTEST) return (i);
		if ((w.work_type == WORK_TEST ||
		     w.work_type == WORK_DBLCHK) &&
		    w.bits < factorLimit (w.p, 0)) return (i);
	}

/* No priority work to do */

	return (0);
}

/* Clear rolling average start time in case Factoring, Torture Test, */
/* Advanced/Test, etc. interrupted a Lucas-Lehmer test. */

void clearRollingStart ()
{
	IniWriteInt (LOCALINI_FILE, "RollingStartTime", 0);
}
		
/* We're starting a prime, save the current time so that */
/* we can maintain the rolling average computations. */

void startRollingAverage ()
{
	time_t	current_time;
	time (&current_time);
	if (VACATION_END && !ON_DURING_VACATION) current_time = 0;
	IniWriteInt (LOCALINI_FILE, "RollingStartTime", current_time);
}

/* We've completed 65536 Lucas-Lehmer iterations of a factoring. */
/* Examine how long it took compared to our expectaions.  Maintain */
/* a rolling average for more accurate time estimates. */

void updateRollingAverage (
	double	est)			/* estimated time for last work unit */
{
#ifndef SERVER_TESTING
	time_t	start_time, current_time;
	unsigned int old_rolling_average, percent;

/* Get the current time and when the iterations started */

	time (&current_time);
	start_time = IniGetInt (LOCALINI_FILE, "RollingStartTime", 0);
	if (start_time == 0 || current_time <= start_time) return;

/* Compute the rolling average where the existing rolling average
/* accounts for 90% of the new rolling average and the current */
/* data point accounts for 10%.  The rolling average */
/* measures this computer's speed as compared to our estimated */
/* speed for this CPU.  A value of 1200 means the this computer */
/* is 20% faster than expected.  A value of 1000 means the this computer */
/* is as fast as expected.   A value of 800 means this computer */
/* is 20% slower than expected. */

	est = est / (current_time - start_time) * 1000.0;
	if (est > 50000.0) return;	/* A safeguard against bogus data */
	if (est < 0.5 * ROLLING_AVERAGE) est = 0.5 * ROLLING_AVERAGE;
	if (est > 2.0 * ROLLING_AVERAGE) est = 2.0 * ROLLING_AVERAGE;

	ROLLING_AVERAGE = (unsigned int)
		(0.90 * ROLLING_AVERAGE + 0.10 * est + 0.5);
	if (ROLLING_AVERAGE < 20) ROLLING_AVERAGE = 20;
	if (ROLLING_AVERAGE > 4000) ROLLING_AVERAGE = 4000;
	IniWriteInt (LOCALINI_FILE, "RollingAverage", ROLLING_AVERAGE);

/* If rolling average is more than 10% different than it was when */
/* the primenet server was told about completion dates, then we need */
/* to send new completion dates to the server */

	old_rolling_average = (unsigned int) IniGetInt (LOCALINI_FILE, "OldRollingAverage", 1000);
	if (old_rolling_average < 10) old_rolling_average = 10;
	percent = ROLLING_AVERAGE * 100 / old_rolling_average;
	if (percent <= 90 || percent >= 110) UpdateEndDates ();
#endif
}

/* Utility output routines */

void LineFeed ()
{
	OutputStr ("\n");
}

void OutputNum (
	unsigned long num)
{
	char	buf[20];
	sprintf (buf, "%lu", num);
	OutputStr (buf);
}

void OutputClocks (
	double	tickCount)
{
	char	buf[80];

        sprintf (buf, "Clocks: %.0f = %.3f sec.\n",
                 tickCount, tickCount / CPU_SPEED / 1000000);
	OutputStr (buf);
}

/* More utility routines */

int timeCheck ()
{
	if (TIMED_RUN) {
		time_t	this_time;
		time (&this_time);
		if (this_time > END_TIME) return (TRUE);
	}
	return (FALSE);
}

/* Continue factoring/testing Mersenne numbers */

void primeContinue () 
{
	struct work_unit w;
	unsigned long fftlen;
	char	filename[20];
	unsigned int line;

/* Send new completion dates once a month */

	ConditionallyUpdateEndDates ();

/* Does a primeSieve continuation file exist? */

	if (fileExists ("p0000000")) {
		strcpy (filename, "p0000000");
		goto readloop;
	}

/* Every time the user chooses Test/Continue, clear the timer that */
/* prevents communication for a period of time.  This allows the user */
/* to try something and if it doesn't work, ESC and choose Test/Continue */
/* to try some other system settings (withoug waiting an hour). */

	next_comm_time = 0; 

/* Loop until the ESC key is hit or the entire what-to-do INI file */
/* is processed and we are not connected to the server. */

	for ( ; ; ) {

/* Send any queued results and get work to do (if necessary) */

	CHECK_WORK_QUEUE = 1;
	if (! communicateWithServer ()) break;

/* If the ESC key is hit, break out of this loop */

	if (escapeCheck ()) break;
			
/* See if there is any work queued up for us to do. */
/* If not, end the thread if we're not connected to primenet. */
/* Otherwise loop until we get some work from primenet. */

	IniFileOpen (WORKTODO_FILE, 1);
	if (IniGetNumLines (WORKTODO_FILE) == 0) {
		IniFileClose (WORKTODO_FILE);
		if (! USE_PRIMENET) {
			OutputSomewhere (DONE_MSG1);
			OutputSomewhere (DONE_MSG2);
			OutputSomewhere (DONE_MSG3);
			break;
		}
		title ("IDLE");
		Sleep (1000);
		continue;
	}

/* See if there is any high priority work to do. */

	line = getPriorityWork ();
	if (line == 0) line = 1;

/* Read the line from the work file */

	if (!parseWorkToDoLine (line, &w)) {
		IniFileClose (WORKTODO_FILE);
		continue;
	}
	IniFileClose (WORKTODO_FILE);

/* Make sure this line of work from the file makes sense. The exponent */
/* should be a prime number, bounded by values we can handle, and we */
/* should never be asked to factor a number more than we are capable of. */

	if (!isPrime (w.p) && w.work_type != WORK_ECM) {
		char	buf[80];
		sprintf (buf, "Error: Work-to-do file contained composite exponent: %ld\n", w.p);
		LogMsg (buf);
		clearExponents (w.p, w.p, 0);
		continue;
	}
	if ((w.work_type == WORK_TEST || w.work_type == WORK_DBLCHK) &&
	    (w.p < MIN_PRIME || w.p > MAX_PRIME)) {
		char	buf[80];
		sprintf (buf, "Error: Work-to-do file contained bad LL exponent: %ld\n", w.p);
		LogMsg (buf);
		clearExponents (w.p, w.p, 0);
		continue;
	}
	if (w.work_type == WORK_FACTOR &&
	    (w.p < 727 || w.p > MAX_FACTOR || w.bits >= factorLimit (w.p, 1))) {
		char	buf[100];
		sprintf (buf, "Error: Work-to-do file contained bad factoring assignment: %ld,%d\n", w.p, w.bits);
		LogMsg (buf);
		clearExponents (w.p, w.p, 0);
		continue;
	}

/* Does a continuation file exist? */

	tempFileName (filename, w.p);
readloop:
	if (continuationFileExists (filename)) {
		int	fd;
		short	type;
		unsigned long counter;

/* Read the initial data from the file */

		if (! readFileHeader (filename, &fd, &type, &counter))
			goto readerr;

/* Continue factoring */

		if (type == 2) {
			int	res;
			if (! primeFactor (w.p, w.bits, &res, w.work_type, fd)) return;
			if (res == 999) goto readerr;
			continue;
		}

/* If type is three then this a sieve continuation */

		if (type == 3) {
			if (! primeSieve (0, 0, 0, 0, 0, fd)) return;
			continue;
		}

/* Deduce the fftlen from the type field */

		fftlen = (unsigned long) type * 1024;

/* Continue Lucas-Lehmer test where it left off.  Check */
/* for corrupt continuation files. */

		if (fftlen != pick_fft_length (w.p) || counter > w.p) {
			_close (fd);
			goto readerr;
		}

/* Load the appropriate EXE file */

		if (! loadRightExe (w.p, &fftlen)) return;

/* Run the setup */

		lucasSetup (w.p, fftlen);

/* Read the initial data from the file, on failure try the backup */
/* intermediate file. */

		if (! readFileData (fd, w.p, fftlen)) {
			char	buf[80];
readerr:		sprintf (buf, READFILEERR, filename);
			OutputBoth (buf);
			_unlink (filename);
			goto readloop;
		}

/* We read the continuation file successfully, continue LL testing */

		if (! prime (w.p, fftlen, counter)) return;
		continue;
	}

/* No continuation file exists */
/* Make sure the first-time user runs a successful self-test. */

	if (line == 1 &&
	    (w.work_type == WORK_TEST ||
	     w.work_type == WORK_DBLCHK)) {
		if (! selfTest (w.p)) return;
	}

/* See if the prime needs more factoring */
/* See if there is more recent information in a local database. */

	if (w.work_type == WORK_FACTOR ||
	    w.work_type == WORK_TEST ||
	    w.work_type == WORK_DBLCHK) {
		int	res;
		w.bits = checkDatabaseForNewInfo (w.p, w.bits);
		if (! primeFactor (w.p, w.bits, &res, w.work_type, 0)) return;
		if (res || getPriorityWork ()) continue;
	}

/* See if this is an ECM factoring line */

	if (w.work_type == WORK_ECM) {
		if (! ecm (w.p, w.B1, w.B2, w.curves_to_do, w.curves_completed, w.curve)) return;
		continue;
	}

/* Otherwise, the prime must need an LL test */

	if (line == 1 &&
	    (w.work_type == WORK_TEST ||
	     w.work_type == WORK_DBLCHK ||
	     w.work_type == WORK_ADVANCEDTEST)) {

/* Load the appropriate EXE file */

		fftlen = 0;
		if (! loadRightExe (w.p, &fftlen)) return;

/* Setup */

		lucasSetup (w.p, fftlen);

/* Start off with the 1st Lucas number */

		if (! prime (w.p, fftlen, 2)) return;
	}

/* Process more lines from the work-to-do file */

	}
}

/* Do the Lucas-Lehmer test */

int prime (
	unsigned long p,
	unsigned long fftlen,
	unsigned long counter)
{
	unsigned long countdown;
	double	totticks;
	int	iters;
	char	filename[20];
	double	reallyminerr = 1.0;
	double	reallymaxerr = 0.0;
	double	maxdiff = (p <  3900000 ? 128.0 :
			  (p <  7700000 ? 256.0 :
			  (p < 15300000 ? 512.0 : 1024.0)));
	int	priority_work = 0;
	int	timed_out = 0;
	int	escaped = 0;
	unsigned long i, high32, low32;
	int	isPrime, rc;
	char	buf[80];

/* Init filename */

	tempFileName (filename, p);

/* Write out intermediate results every DISK_WRITE_TIME minutes. */

	countdown = (unsigned long)
		(DISK_WRITE_TIME * 60.0 / map_fftlen_to_timing (fftlen));

/* Init the title */

begin:	{
		char buf[80];
		sprintf (buf, "%ld / %ld", counter, p);
		title (buf);
	}

/* Init global vars for Test/Status and CommunicateWithServer */

	EXP_BEING_WORKED_ON = p;
	EXP_BEING_FACTORED = 0;
	EXP_PERCENT_COMPLETE = (double) counter / (double) p;

/* Start off with the 1st Lucas number */

	if (counter == 2) {
		unsigned long i;
		for (i = 0; i < fftlen; i++) {
			set_fft_value (p, fftlen, i, (i == 0) ? 4 : 0);
		}
		startRollingAverage ();
	}

/* Compute numbers in the lucas series, write out every 30 minutes to a file */

	totticks = 0.0; iters = 0;
	while (counter < p) {
		unsigned long temp, ticks;
		int	stopping, echk;

/* Use this oppurtunity to perform other miscellaneous tasks that may */
/* be required by this particular OS port */

		doMiscTasks ();

/* Every so often, communicate with the server. */
/* Even more rarely, set flag to see if we have enough work queued up. */
/* Also, see if we should switch to factoring.  This is sometimes done */
/* when we get a new assignment that hasn't been factored.  To be sure */
/* that we will be able to immediately startup a LL test when this one */
/* completes, we will factor the next exponent and then come back to */
/* finish the LL test of this exponent. */

		if ((counter & 0x3F) == 0) {
			if ((counter & 0xFFFF) == 0) {
				ConditionallyUpdateEndDates ();
				CHECK_WORK_QUEUE = 1;
				EXP_PERCENT_COMPLETE =
					(double) counter / (double) p;
			}
			if (!communicateWithServer ()) escaped = 1;
			if ((counter & 0xFFFF) == 0) {
				IniFileOpen (WORKTODO_FILE, 0);
				if (getPriorityWork ()) priority_work = 1;
				IniFileClose (WORKTODO_FILE);
				updateRollingAverage (
					map_fftlen_to_timing (fftlen) *
					65536.0 * 24.0 / CPU_HOURS);
				startRollingAverage ();
			}

/* If this is a timed run, check if our time has expired */

			timed_out = timeCheck ();
		}

/* Error check the last 50 iterations, before writing an */
/* intermediate file (either user-requested stop or a */
/* 30 minute interval expired), and every 128th iteration. */

		escaped |= escapeCheck ();
		stopping = escaped || priority_work || timed_out;
		echk = (counter >= p - 50) || stopping ||
			(countdown == 1) || (counter & 127) == 0;

/* Do a Lucas-Lehmer iteration */

		if (echk) ERRCHK++;
		if (CPU_TYPE >= 5) RDTSC (temp);
#ifndef SERVER_TESTING
		(*lucas)();
#endif
		if (CPU_TYPE >= 5) {
			RDTSC (ticks);
			totticks += ticks - temp;
			iters++;
		}
		if (echk) ERRCHK--;

/* If the sum of the output values is an error (such as infinity) */
/* then raise an error. */

		{
			short fp_status;
#if defined (__linux__) || defined (__EMX__)
			__asm__ __volatile__ (
				"fldl %1\n fxam\n fnstsw %0\n fstpl %1"
				: "=m" (fp_status)
				: "m " (SUMOUT)
			);
#else
			__asm fld SUMOUT
			__asm fxam
			__asm fnstsw fp_status
			__asm fstp SUMOUT
#endif
			if (fp_status & 0x0100) {
				char buf[80];
				sprintf (buf, ERRMSG0, counter, p, ERRMSG1A);
				OutputBoth (buf);
				spoolMessage (PRIMENET_RESULT_MESSAGE, buf);
				goto restart;
			}
		}

/* Check that the sum of the input numbers squared is approximately */
/* equal to the sum of unfft results.  Since this check may not */
/* be perfect, check for identical results after a restart. */

		if (SUMOUT < SUMINP - maxdiff || SUMOUT > SUMINP + maxdiff) {
			static unsigned long last_counter = 0;
			static double last_suminp = 0.0;
			static double last_sumout = 0.0;
			if (counter == last_counter &&
			    SUMINP == last_suminp &&
			    SUMOUT == last_sumout) {
				writeResults (ERROK);
				countdown = 1;
			} else {
				char buf[120], msg[80];
				sprintf (msg, ERRMSG1B, SUMINP, SUMOUT);
				sprintf (buf, ERRMSG0, counter, p, msg);
				OutputBoth (buf);
				spoolMessage (PRIMENET_RESULT_MESSAGE, buf);
				last_counter = counter;
				last_suminp = SUMINP;
				last_sumout = SUMOUT;
				goto restart;
			}
		}

/* Check for excessive roundoff error  */

		if ((ERRCHK || echk) && MAXERR > 0.40) {
			static unsigned long last_counter = 0;
			static double last_maxerr = 0.0;
			if (counter == last_counter &&
			    MAXERR == last_maxerr) {
				writeResults (ERROK);
				countdown = 1;
			} else {
				char buf[80], msg[80];
				sprintf (msg, ERRMSG1C, MAXERR);
				sprintf (buf, ERRMSG0, counter, p, msg);
				OutputBoth (buf);
				spoolMessage (PRIMENET_RESULT_MESSAGE, buf);
				last_counter = counter;
				last_maxerr = MAXERR;
				goto restart;
			}
		}

/* Some special debugging code to dump out FFT results */

#ifdef XXX
xxx:{
unsigned long i;
char	buf[1024];
int	fd = _open (RESFILE, _O_TEXT | _O_RDWR | _O_CREAT, 0666);
while (_read (fd, buf, sizeof (buf)));
for (i = 0; i < fftlen; i++) {
	double *dblp = (double *) addr (p, fftlen, i);
	if (*dblp < -0.0001 || *dblp > 0.0001) {
		sprintf (buf, "i: %ld, %10.6f\n", i, *dblp);
		_write (fd, buf, strlen (buf));
	}
}
_close (fd);
}
#endif

/* Update counter and maximum round-off error */

		counter++;
		if (ERRCHK) {
			if (MAXERR < reallyminerr && counter > 30)
				reallyminerr = MAXERR;
			if (MAXERR > reallymaxerr)
				reallymaxerr = MAXERR;
		}

/* Print a message every so often */

		if (counter % ITER_OUTPUT == 0) {
			char buf[80];
			sprintf (buf, "%ld / %ld", counter, p);
			title (buf);
			sprintf (buf, "Iteration: %ld / %ld [%ld%%]",
				 counter, p, counter * 100 / p);
			OutputStr (buf);
			if (ERRCHK && counter > 30) {
				OutputStr (".  Round off: ");
				sprintf (buf, "%10.10f", reallyminerr);
				OutputStr (buf);
				sprintf (buf, " to %10.10f", reallymaxerr);
				OutputStr (buf);
			}
			OutputStr (".  ");
			if (CPU_TYPE >= 5) {
				OutputClocks (totticks / iters);
				totticks = 0.0; iters = 0;
			}
			else LineFeed ();
		}

/* Print a results file message every so often */

		if (counter % ITER_OUTPUT_RES == 0 || (NO_GUI && stopping)) {
			char buf[80];
			sprintf (buf, "Iteration %ld / %ld\n", counter, p);
			writeResults (buf);
		}

/* If an escape key was hit, write out the results and return */

		if (stopping) {
			if (! writeToFile (p, filename, fftlen, counter)) {
				char buf[80];
				sprintf (buf, WRITEFILEERR, filename);
				OutputBoth (buf);
				return (FALSE);
			}
			if (escaped || timed_out) return (FALSE);
			return (TRUE);
		}

/* Write results to a file every DISK_WRITE_TIME minutes */
/* On error, retry in 10 minutes (it could be a temporary */
/* disk-full situation) */

		if (--countdown == 0) {
			unsigned long write_time = DISK_WRITE_TIME;
			if (! writeToFile (p, filename, fftlen, counter)) {
				char buf[80];
				sprintf (buf, WRITEFILEERR, filename);
				OutputBoth (buf);
				if (write_time > 10) write_time = 10;
			}
			countdown = (unsigned long)
				(write_time * 60.0 / map_fftlen_to_timing (fftlen));
		}
	}

/* Check for a successful completion */
/* We found a prime if result is zero */

	for (i = 0, isPrime = TRUE; isPrime && i < fftlen; i++) {
		long	val;
		get_fft_value (p, fftlen, i, &val);
		if (val) isPrime = FALSE;
	}

/* Format the output message */

	if (isPrime) {
		sprintf (buf, "M%ld is prime! WS%d: %08lX\n",
			 p, PORT, SEC1 (p));
		high32 = low32 = 0;
	} else {
		generateResidue64 (p, fftlen, &high32, &low32);
		sprintf (buf,
			 "M%ld is not prime. Res64: %08lX%08lX. WS%d: %08lX\n",
			 p, high32, low32, PORT, SEC2 (p, high32, low32));
	}

/* Output results to the screen, results file, and server */

	{
		struct primenetAssignmentResult pkt;
		memset (&pkt, 0, sizeof (pkt));
		pkt.exponent = p;
		pkt.resultType =
			isPrime ? PRIMENET_RESULT_PRIME : PRIMENET_RESULT_TEST;
		sprintf (pkt.resultInfo.residue,
			 "%08lX%08lX", high32, low32);
		spoolMessage (PRIMENET_ASSIGNMENT_RESULT, &pkt);
	}
	OutputStr (buf);
	rc = writeResults (buf);
	spoolMessage (PRIMENET_RESULT_MESSAGE, buf);

/* Delete the continuation files - assuming the results file write */
/* was successful. */

	if (rc) _unlink (filename);
	filename[0] = 'q';
	_unlink (filename);

/* Clear prime from the to-do list */

	clearExponents (p, p, 0);

/* Clear rolling average start time in case Advanced/Test */
/* interrupted a Lucas-Lehmer test. */

	clearRollingStart ();
		
/* Output good news to the screen in an infinite loop */

	if (isPrime && !SILENT_VICTORY && !isKnownMersennePrime (p)) {
		title ("New Prime!!!");
		for ( ; ; ) {
			OutputStr ("New Mersenne Prime!!!!  ");
			OutputStr (buf);
			flashWindowAndBeep ();
			if (escapeCheck ()) return (FALSE);
			if (!communicateWithServer ()) return (FALSE);
		}
	}

/* All done */

	return (TRUE);

/* An error occured, try restarting at last save point */

restart:
	{
		int	fd;
		short	type;

/* Clear rolling average in case we will be backtracking past a */
/* 64K iteration boundary. */

		clearRollingStart ();

/* Output a message saying we are restarting */

		OutputBoth (ERRMSG2);
		OutputBoth (ERRMSG3);

/* Sleep five minutes before restarting */

		OutputStr (ERRMSG4);
		Sleep (300000);

/* Read the initial data from the file */

		if (! readFileHeader (filename, &fd, &type, &counter)) {
			counter = 2;
			goto begin;
		}

/* Check for bad type or counter */

		if (type != (short) (fftlen / 1024) || counter > p) {
			_close (fd);
			counter = 2;
			goto begin;
		}

/* Run the setup (in case the sine/cosine data was corrupted) */

		lucasSetup (p, fftlen);

/* Read the initial data from the file */

		if (! readFileData (fd, p, fftlen)) counter = 2;
		goto begin;
	}
}

#define SELFMSG1A "The program will now perform a self-test to make sure the\n"
#define SELFMSG1B "Lucas-Lehmer code is working properly on your computer.\n"
#define SELFMSG1C "This will take about an hour.\n"
#define SELFMSG5C "This will take about sixteen hours.\n"
#define SELF1 "Test %i of %i, %i Lucas-Lehmer iterations of M%ld using %ldK FFT length.\n"
#define SELFFAIL "FATAL ERROR: Exponent %ld, LSW: %ld.\n"
#define SELFFAILW "FATAL ERROR: Writing to temp file.\n"
#define SELFFAILR "FATAL ERROR: Reading from temp file.\n"
#define SELFPASS "Self-test %i passed!\n"
char SelfTestIniMask[] = "SelfTest%iPassed";

#define MAX_SELF_TEST_ITERS	160

struct self_test_info {
	unsigned long p;
	unsigned long reshi;
};

struct self_test_info SELF_TEST_DATA[MAX_SELF_TEST_ITERS] = {
{2654999, 0x07EE900D},{2621441, 0x2BC1DACD},{2621439, 0xBCBA58F1},
{2653987, 0xB005CACC},{2651879, 0x38DCD06B},{2654003, 0x1ED556E7},
{2620317, 0x09DB64F8},{2539613, 0x4146EECA},{2573917, 0x939DA3B3},
{2359297, 0x73A131F0},{2359295, 0x53A92203},{2646917, 0x71D4E5A2},
{2605473, 0xE11637FC},{2495213, 0x89D80370},{2540831, 0x2CF01FBB},
{2654557, 0x4106F46F},{2388831, 0xA508B5A7},{2654777, 0x9E744AA3},
{2584313, 0x800E9A61},{2408447, 0x8C91E8AA},{2408449, 0x437ECC01},
{2345677, 0x60AEE9C2},{2332451, 0xAB209667},{2330097, 0x3FB88055},
{2333851, 0xFE4ECF19},{2444819, 0x56BF33C5},{2555671, 0x9DC03527},
{2654333, 0xE81BCF40},{2543123, 0x379CA95D},{2432123, 0x5952676A},
{2321123, 0x24DCD25F},{2654227, 0xAC3B7F2B},
{2329999, 0xF5E902A5},{2293761, 0x9E4BBB8A},{2293759, 0x1901F07B},
{2236671, 0x45EB162A},{2193011, 0x382B6E4B},{2329001, 0x4FF052BB},
{2327763, 0x3B315213},{2325483, 0x0DC5165A},{2323869, 0xD220E27F},
{2315679, 0xF650BE33},{2004817, 0xC2FF3440},{2130357, 0xC25804D8},
{2288753, 0xA4DD9AAD},{2266413, 0x675257DB},{2244765, 0xC08FF487},
{2222517, 0x1A128B22},{2200339, 0x0EB0E827},{2328117, 0x0A24673A},
{2329557, 0x2E267692},{2188001, 0xD012AF6A},{2166567, 0x509BA41A},
{2144651, 0x54CFC0E6},{2122923, 0xA47068E6},{2100559, 0xACFAB4E1},
{2088461, 0xEA01E860},{2066543, 0x847DF0D0},{2044767, 0x04225888},
{2022823, 0x6EA34B32},{2328527, 0xC55E3E05},{2327441, 0x207C8CEC},
{2326991, 0x0A4F2ACD},{2009987, 0xE6A59DEF},
{1999999, 0xD645A18F},{1966081, 0xB88828A1},{1966079, 0x5BD87C45},
{1998973, 0xCBDD74F7},{1997651, 0x666B0CB1},{1675001, 0x50A94DB7},
{1977987, 0x30D1CD1F},{1955087, 0x5B9426A4},{1933071, 0x23C1AF0B},
{1911957, 0xF7699248},{1899247, 0x11C76E04},{1877431, 0xA3299B39},
{1855067, 0x35243683},{1833457, 0xCF630DC0},{1811987, 0x7C7022EC},
{1799789, 0xEFEC47B7},{1777773, 0x0F16E2D6},{1755321, 0x1AC5D492},
{1733333, 0x5DA0555E},{1711983, 0xDC19DA8B},{1699779, 0x2B44914E},
{1677323, 0x03D3980B},{1995091, 0x922E555B},{1993041, 0x0CA8451B},
{1991991, 0xDFFB212D},{1679779, 0x51D75E0F},{1684993, 0x048BBCE8},
{1970009, 0x646E0DFA},{1957445, 0xC8D244ED},{1999997, 0x5FC899D0},
{1998983, 0x1CD518AA},{1999007, 0xA9DD8591},
{1674999, 0xDB0169D8},{1638401, 0xD3F8A8C5},{1638399, 0xF270D8DD},
{1674997, 0xC824EF15},{1674551, 0xD844AEAD},{1674001, 0x8F5EFA50},
{1345001, 0x18EE2E2D},{1655083, 0x09B30DEE},{1633941, 0x0B87C8B1},
{1611557, 0x6B57E48D},{1599549, 0x48EA38B2},{1577771, 0xCE84D9DC},
{1555947, 0x6797EEF4},{1533349, 0xD6897409},{1511861, 0x8A8177AC},
{1499625, 0x56BB6FB3},{1477941, 0xF3DD8ED3},{1455931, 0x31A222C7},
{1433069, 0x28F01E1B},{1411747, 0x680C6E39},{1399449, 0xB7F01A54},
{1377247, 0xE656F652},{1355991, 0xB2AA2819},{1350061, 0x31F9A728},
{1673881, 0xA51D38E4},{1672771, 0x5474B6F9},{1671221, 0x2710DDEA},
{1670551, 0x31FC3838},{1660881, 0x4C5B22C5},{1650771, 0x998F747B},
{1655001, 0x164659A6},{1674339, 0xED2D23E2},
{1344999, 0x158AA064},{1310721, 0x5694A427},{1310719, 0x258BDDE3},
{1344997, 0x1D059D4F},{1344551, 0x60606AA3},{1344001, 0x9AC6AB36},
{675001,  0x688DF7EE},{1322851, 0x3A000D0A},{1300993, 0x77CB0184},
{1288771, 0x7431D9E2},{1266711, 0xB4BC4E8D},{1244881, 0x48BC9FF9},
{1222991, 0x3F5FC39E},{1200881, 0xD5DF4944},{1188441, 0xD9D8968B},
{1166661, 0xD4AB97F4},{1144221, 0x9940943B},{1122001, 0x647406B8},
{1100881, 0x3AD40CE0},{1088511, 0xD578BB51},{1066837, 0x2F82BFBB},
{1044811, 0x7C6EDDD1},{1022991, 0x6A1C2DD4},{1000001, 0x2879748F},
{1343881, 0xB59E8006},{1342771, 0x87563FFE},{1341221, 0x29AD6127},
{1340551, 0x17DB4ACB},{1330881, 0x9642F068},{951951,  0xB57517EE},
{891891,  0xC0D97D8D},{773337,  0xA1FF9438}
};

int selfTest (
	unsigned long pArg)
{
	unsigned long fftlen;
	int	i;

/* Clear rolling average start time in case we've */
/* interrupted a Lucas-Lehmer test. */

	clearRollingStart ();

/* Are we to self-test all word lengths? */
/* Check for the goofy torture-test code */

	if (pArg <= 1) {
		int	j;

/* Make sure the user really wants to spend nine hours doing this now */

		if (pArg == 0) {
			OutputStr (SELFMSG1A);
			OutputStr (SELFMSG1B);
			OutputStr (SELFMSG5C);
		}

/* Now test each fft length */

		for (j = 0; ; j++) {
		    for (i = 0; i <= 15; i++) {
			unsigned long p;
			if (i == 15) p = LIMIT80-1;
			if (i == 14) p = LIMIT96-1;
			if (i == 13) p = LIMIT112-1;
			if (i == 12) p = LIMIT128-1;
			if (i == 11) p = LIMIT160-1;
			if (i == 10) p = LIMIT192-1;
			if (i == 9) p = LIMIT224-1;
			if (i == 8) p = LIMIT256-1;
			if (i == 7) p = LIMIT320-1;
			if (i == 6) p = LIMIT384-1;
			if (i == 5) p = LIMIT448-1;
			if (i == 4) p = LIMIT512-1;
			if (i == 3) p = LIMIT640-1;
			if (i == 2) p = LIMIT768-1;
			if (i == 1) p = LIMIT896-1;
			if (i == 0) p = LIMIT1024-1;

/* Load the appropriate EXE file and setup */

			fftlen = 0;
			if (! loadRightExe (p, &fftlen)) return (FALSE);

/* Do the self test */

			if (! selfTestInternal (fftlen, j)) return (FALSE);
		    }
		    if (pArg == 0) break;
		}
	}

/* Otherwise we are self-testing a specific word length. */

	else {
		char	iniName[32];
		unsigned long temp;

/* What fft length are we running? */

		fftlen = pick_fft_length (pArg);

/* If so, make sure we haven't done so already. */

		sprintf (iniName, SelfTestIniMask, (int) (fftlen/1024));
		if (IniGetInt (LOCALINI_FILE, iniName, 0)) return (TRUE);

/* Make sure the user really wants to spend an hour doing this now */

		OutputStr (SELFMSG1A);
		OutputStr (SELFMSG1B);
		OutputStr (SELFMSG1C);

/* Load the appropriate EXE file and setup */

		temp = 0;
		if (! loadRightExe (pArg, &temp)) return (FALSE);

/* Do the self test */

		if (! selfTestInternal (fftlen, 0)) return (FALSE);
	}

/* Self test completed! */

	return (TRUE);
}

int selfTestInternal (
	unsigned long fftlen,
	int	torture_count)	/* Number of times this fftlen has */
				/* been tested in a torture test */
{
	unsigned long k, limit;
	unsigned int i, num_iters, iters_left, countdown;
	double	timing;
	char	buf[100];
	char	iniName[32];

/* Determine the range from which we'll choose an exponent to test. */

	if (fftlen == 81920) limit = LIMIT80;
	if (fftlen == 98304) limit = LIMIT96;
	if (fftlen == 114688) limit = LIMIT112;
	if (fftlen == 131072) limit = LIMIT128;
	if (fftlen == 163840) limit = LIMIT160;
	if (fftlen == 196608) limit = LIMIT192;
	if (fftlen == 229376) limit = LIMIT224;
	if (fftlen == 262144) limit = LIMIT256;
	if (fftlen == 327680) limit = LIMIT320;
	if (fftlen == 393216) limit = LIMIT384;
	if (fftlen == 458752) limit = LIMIT448;
	if (fftlen == 524288) limit = LIMIT512;
	if (fftlen == 655360) limit = LIMIT640;
	if (fftlen == 786432) limit = LIMIT768;
	if (fftlen == 917504) limit = LIMIT896;
	if (fftlen == 1048576) limit = LIMIT1024;

/* Add 15% to iteration timing for the error checking. */
/* Only do an hour's worth of iterations. */

	timing = map_fftlen_to_timing (fftlen) * 1.15;
	num_iters = (int) (3600.0 / timing / 400.0);
	if (num_iters < 5) num_iters = 5;

/* Loop testing various exponents, testing one exponent from each run length */

	countdown = 1;
	iters_left = num_iters * (torture_count + 1);
	while (iters_left) {
	    for (i = 0; i < MAX_SELF_TEST_ITERS; i++) {
		unsigned long p, reshi, reslo;
		unsigned int ll_iters;

/* Now select the actual exponent */

		p = SELF_TEST_DATA[i].p;
		if (p > limit) continue;
		ll_iters = 400;

/* If this is a torture test, then do a different set of */
/* self-tests each hour. */

		if (iters_left > num_iters) {
			iters_left--;
			continue;
		}

/* Set the title */

		title ("Self-Test");

/* Output start message */

		sprintf (buf, SELF1, num_iters - iters_left + 1,
			 num_iters, ll_iters, p, fftlen / 1024);
		OutputStr (buf);

/* Now run Lucas setup */

		lucasSetup (p, fftlen);
		
/* Init data area with a pre-determined value */

		clear_memory (p, fftlen);
		set_fft_value (p, fftlen, 0, 4);

/* Do Lucas-Lehmer iterations */

		for (k = 0; k < ll_iters; k++) {
			int	fd;
			short	type;
			unsigned long counter;

/* One Lucas-Lehmer test with error checking */

			ERRCHK++;
			(*lucas)();
			ERRCHK--;

/* If the sum of the output values is an error (such as infinity) */
/* then raise an error.  For some reason these bad values are treated */
/* as zero by the C compiler.  There is probably a better way to */
/* check for this error condition. */

			if (SUMOUT <= 0.0) {
				OutputBoth (ERRMSG1A);
				OutputBoth (ERRMSG2);
				return (FALSE);
			}

/* Check that the sum of the input numbers squared is approximately */
/* equal to the sum of unfft results. */

			if (SUMOUT < SUMINP-128.0 || SUMOUT > SUMINP+128.0) {
				sprintf (buf, ERRMSG1B, SUMINP, SUMOUT);
				OutputBoth (buf);
				OutputBoth (ERRMSG2);
				return (FALSE);
			}

/* Make sure round off error is tolerable */

			if (MAXERR > 0.40) {
				sprintf (buf, ERRMSG1C, MAXERR);
				OutputBoth (buf);
				OutputBoth (ERRMSG2);
				return (FALSE);
			}

/* Abort if user demands it */

			if (escapeCheck ()) return (FALSE);

/* Test our ability to read and write files too. */

			if (k != 100) continue;
			if (--countdown) continue;
			if (! writeToFile (p, "ptemp", fftlen, 0)) {
				OutputBoth (SELFFAILW);
				return (FALSE);
			}
			countdown = 9;
			clear_memory (p, fftlen);
			if (! readFileHeader ("ptemp", &fd, &type, &counter) ||
			    ! readFileData (fd, p, fftlen)) {
				OutputBoth (SELFFAILR);
				return (FALSE);
			}
			_unlink ("ptemp");
		}

/* Compare final 32 bits with the pre-computed array of correct residues */

		generateResidue64 (p, fftlen, &reshi, &reslo);
		if (reshi != SELF_TEST_DATA[i].reshi) {
			sprintf (buf, SELFFAIL, p, reshi);
			OutputBoth (buf);
			OutputBoth (ERRMSG2);
			return (FALSE);
		}

/* One less iteration to do */

		if (--iters_left == 0) break;
	    }
	}

/* We've passed the self-test.  Remember this in the .INI file */
/* so that we do not need to do this again. */

	sprintf (buf, SELFPASS, (int) (fftlen/1024));
	OutputBoth (buf);
	sprintf (iniName, SelfTestIniMask, (int) (fftlen/1024));
	IniWriteInt (LOCALINI_FILE, iniName, 1);
	return (TRUE);
}

/* Generate random FFT data for timing the Lucas-Lehmer code */

void generateRandomData (
	unsigned long p,
	unsigned long fftlen)
{
	unsigned long i;

/* Fill data space with random values. */

	srand ((unsigned int) p);
	for (i = 0; i < fftlen; i++) {
		set_fft_value (p, fftlen, i, rand());
	}
}

void primeTime (
	unsigned long p,
	unsigned long iterations)
{
	unsigned long i, fftlen;
	unsigned long temp, ticks;
	double	totticks;

/* Load the appropriate EXE file */

	fftlen = 0;
	if (! loadRightExe (p, &fftlen)) return;
	lucasSetup (p, fftlen);

/* Fill data space with random values. */

	generateRandomData (p, fftlen);

/* Compute numbers in the lucas series */

	totticks = 0.0;
	for (i = 0; i < iterations; i++) {
		if (CPU_TYPE >= 5) RDTSC (temp);
		(*lucas)();
		if (CPU_TYPE >= 5) {
			RDTSC (ticks);
			ticks = ticks - temp;
			totticks += ticks;
		}
		OutputStr ("p: ");
		OutputNum (p);
		OutputStr (".  ");
		if (CPU_TYPE >= 5) OutputClocks ((double) ticks);
		else LineFeed ();
		if (escapeCheck ()) return;
	}
	OutputStr ("Iterations: ");
	OutputNum (iterations);
	OutputStr (".  ");
	if (CPU_TYPE >= 5) {
		char	buf[80];
		double	time;
		int	days, hours, minutes;

		OutputClocks (totticks);
		time = totticks / iterations * p / CPU_SPEED / 1000000;
		days = (int) (time / 86400.0); time -= (double) days * 86400.0;
		hours = (int) (time / 3600.0); time -= (double) hours * 3600.0;
		minutes = (int) (time / 60.0);
		OutputStr ("Estimated time to complete this exponent: ");
		sprintf (buf, days == 1 ? "%d day, " : "%d days, ", days);
		OutputStr (buf);
		sprintf (buf, hours == 1 ? "%d hour, " : "%d hours, ", hours);
		OutputStr (buf);
		sprintf (buf, minutes == 1 ? "%d minute, " : "%d minutes.\n",
			 minutes);
		OutputStr (buf);
	} else
		LineFeed ();
}

/* Trial factor a Mersenne number prior to running a Lucas-Lehmer test */

char FACMSG[] = "Factored M%ld through %lu*2^32 (pass %d of 16).  ";

int primeFactor (
	unsigned long p,
	unsigned int bits,
	int	*result,
	int	work_type,
	int	fd)			/* Set if this is a continuation */
{
	unsigned long write_counter, countdown;
	char	filename[20];
	char	buf[80], str[80];
	int	test;
	short	pass;
	unsigned long startpt, endpt;
	unsigned long iters, iters_r;
	double	totticks;
	int	stopping = 0;

/* Init */

	tempFileName (filename, p);

/* Write out intermediate results every DISK_WRITE_TIME minutes. */
/* 8000 iterations for p = 1300000 on a P90 takes 30 minutes. */

	write_counter = 115555555 / p * CPU_SPEED * DISK_WRITE_TIME / 30;
	if (CPU_TYPE <= 4) write_counter /= 7;
	countdown = write_counter;

/* Determine how much we should factor (in bits) */

	test = factorLimit (p, work_type == WORK_FACTOR);
	if (test == 64) endpt = 0xFFFFFFFF;	/* Compute endpoint */
	else endpt = 1L << (test - 32);
	if (bits < 32) startpt = 0;
	else startpt = 1L << (bits-32);

/* Is this a continuation?  If so, read continuation file. */

	if (fd) {
		short	res;
		if (_read (fd, &pass, sizeof (short)) != sizeof (short)) goto readerr;
		if (_read (fd, &res, sizeof (short)) != sizeof (short)) goto readerr;
		*result = res;
		if (_read (fd, &FACMSW, sizeof (long)) != sizeof (long)) goto readerr;
		if (_read (fd, &startpt, sizeof (long)) != sizeof (long)) goto readerr;
		if (_read (fd, &endpt, sizeof (long)) != sizeof (long)) goto readerr;
		_close (fd);
	} else {
		pass = 1;
		*result = FALSE;
		FACMSW = startpt;
		startRollingAverage ();
	}

/* We cannot handle factoring above 2^64 right now */

	if (bits >= 64) return (TRUE);

/* Is prime already factored enough? */

	if (startpt >= endpt) return (TRUE);

/* Init global vars for Test/Status and CommunicateWithServer */

	EXP_BEING_WORKED_ON = p;
	EXP_BEING_FACTORED = 1;
	EXP_PERCENT_COMPLETE = (double) (pass-1) / 16.0;

/* Load the factoring code */

	if (! loadfac ()) return (FALSE);

/* Sixteen passes! two for the 1 or 7 mod 8 factors times two for the */
/* 1 or 2 mod 3 factors times four for the 1, 2, 3, or 4 mod 5 factors. */

	iters_r = 0;
	iters = 0;
	totticks = 0.0;
	for ( ; pass <= 16; pass++) {

/* Only test for factors less than 2^32 on the first pass */

		if (FACMSW == 0 && pass > 1) FACMSW = 1;

/* Setup the factoring program */

		FACPASS = pass-1;
		FACLSW = p;
		setupf ();

/* Loop until all factors tested or factor found */

		for ( ; ; ) {
			unsigned long ticks, temp;

			if (CPU_TYPE >= 5) RDTSC (temp);
#ifdef SERVER_TESTING
			FACMSW=FACMSW*2+1000;
			if ((temp & 0x1FF) == 0x111) break;
#else
			if (factor64 () != 2) break;
#endif
			if (CPU_TYPE >= 5) {
				RDTSC (ticks);
				totticks += ticks - temp;
			}

/* Send queued messages to the server every so often */

			if ((iters_r & 0x7F) == 0) {
				if (!communicateWithServer ()) stopping = 1;

/* If this is a timed run, check if our time has expired */

				stopping |= timeCheck ();
			}

/* Set flag if we are stopping */

			stopping |= escapeCheck ();

/* Output informative message */

			if (++iters >= ITER_OUTPUT) {
				sprintf (buf, FACMSG, p, FACMSW, pass);
				OutputStr (buf);
				if (CPU_TYPE >= 5)
					OutputClocks (totticks / iters);
				else
					LineFeed ();
				sprintf (buf, "%ld", FACMSW);
				title (buf);
				iters = 0;
				totticks = 0.0;
			}

/* Output informative message */

			if (++iters_r >= ITER_OUTPUT_RES ||
			    (NO_GUI && stopping)) {
				sprintf (buf, FACMSG, p, FACMSW, pass);
				strcat (buf, "\n");
				writeResults (buf);
				iters_r = 0;
			}

/* Test for completion */

			if (FACMSW >= endpt) goto nextpass;

/* If an escape key was hit, write out the results and return */

			if (stopping || --countdown == 0) {
				short	res, two = 2;
				long	dummy;
				fd = _creat (filename, 0666);
				_close (fd);
				fd = _open (filename, _O_BINARY | _O_RDWR);
				_write (fd, &two, sizeof (short));
				_write (fd, &dummy, sizeof (long));
				_write (fd, &pass, sizeof (short));
				res = *result;
				_write (fd, &res, sizeof (short));
				_write (fd, &FACMSW, sizeof (long));
				_write (fd, &startpt, sizeof (long));
				_write (fd, &endpt, sizeof (long));
				_commit (fd);
				_close (fd);
				if (stopping) return (FALSE);
				countdown = write_counter;
			}
		}

/* Format the output message */

		makestr (FACMSW, FACLSW, str);

/* Output results to the screen, results file, and server */

		{
			struct primenetAssignmentResult pkt;
			memset (&pkt, 0, sizeof (pkt));
			pkt.exponent = p;
			pkt.resultType = PRIMENET_RESULT_FACTOR;
			strcpy (pkt.resultInfo.factor, str);
			spoolMessage (PRIMENET_ASSIGNMENT_RESULT, &pkt);
		}
		sprintf (buf, "M%ld has a factor: %s\n", p, str);
		OutputBoth (buf);
		spoolMessage (PRIMENET_RESULT_MESSAGE, buf);

/* Continue factoring (to find a smaller factor in a later pass). */

		*result = TRUE;
		if (FACMSW != 0xFFFFFFFF) endpt = FACMSW+1;

/* Do next pass */

nextpass:	FACMSW = startpt;	/* Start next pass at the start point*/

/* Every so often, set flag to see if we have enough work queued up */
/* Also, update the percent complete */

		ConditionallyUpdateEndDates ();
		CHECK_WORK_QUEUE = 1;
		if (*result) EXP_PERCENT_COMPLETE = 999.0;
		else EXP_PERCENT_COMPLETE = (double) pass / 16.0;
	}

/* Clear prime from the to-do list */

	if (*result)
		clearExponents (p, p, 0);

/* Output message if no factor found */

	else {
		struct primenetAssignmentResult pkt;
		struct work_unit w;

		memset (&pkt, 0, sizeof (pkt));
		pkt.exponent = p;
		pkt.resultType = PRIMENET_RESULT_NOFACTOR;
		pkt.resultInfo.how_far_factored = test;
		spoolMessage (PRIMENET_ASSIGNMENT_RESULT, &pkt);

		sprintf (buf, "M%ld no factor to 2^%d, WS%d: %08lX\n",
			 p, test, PORT, SEC3 (p));
		OutputBoth (buf);
		spoolMessage (PRIMENET_RESULT_MESSAGE, buf);

		clearExponents (p, p, test);

/* Update the rolling average when no factor is found */

		w.work_type = WORK_FACTOR;
		w.p = p;
		w.bits = bits;
		updateRollingAverage (raw_work_estimate (&w) * 24.0 / CPU_HOURS);
	}

/* Delete the continuation file */

	_unlink (filename);

/* Clear rolling average start time in case we've */
/* interrupted a Lucas-Lehmer test. */

	clearRollingStart ();

/* All done */

	return (TRUE);

/* Return a kludgy error code to indicate an error reading intermediate file */

readerr:_close (fd);
	*result = 999;
	return (TRUE);
}


/* Factor a range of primes using factors of the specified size */

char NOFAC[] = "M%ld no factor from 2^%d to 2^%d, WS%d: %08lX\n";

int primeSieve (
	unsigned long startp,
	unsigned long endp,
	unsigned short minbits,
	unsigned short maxbits,
	short	fromfile,
	int	fd)			/* Continuation file */
{
	unsigned long write_counter, countdown;
	char	filename[20], buf[80], str[80];
	int	increasing, continuation;
	unsigned long p, startpt, endpt, temp_endpt, iters, iters_r;
	short	pass, found_factor;
	double	totticks;
	int	stopping = 0;

/* Clear rolling average start time in case we've */
/* interrupted a Lucas-Lehmer test. */

	clearRollingStart ();

/* Special value indicates check all factors in a file named "factors" */

	if (startp == 8888) {
		primeSieveTest ();
		return (FALSE);
	}

/* Is this a continuation?  If so, read continuation file. */

	if (fd) {
		_read (fd, &startp, sizeof (long));
		_read (fd, &endp, sizeof (long));
		_read (fd, &minbits, sizeof (short));
		_read (fd, &maxbits, sizeof (short));
		_read (fd, &fromfile, sizeof (short));
		_read (fd, &p, sizeof (long));
		_read (fd, &FACMSW, sizeof (long));
		_read (fd, &pass, sizeof (short));
		_read (fd, &temp_endpt, sizeof (long));
		_read (fd, &found_factor, sizeof (short));
		_close (fd);
		continuation = TRUE;
	} else
		continuation = FALSE;

/* Init filename */

	strcpy (filename, "p0000000");

/* Write out intermediate results every DISK_WRITE_TIME minutes. */
/* 8000 iterations for p = 1300000 on a P90 takes 30 minutes. */

	write_counter = 115555555 / startp * CPU_SPEED * DISK_WRITE_TIME / 30;
	if (CPU_TYPE <= 4) write_counter /= 7;
	countdown = write_counter;

/* Determine how much we should factor */

	if (minbits > 32)		/* Compute startpoint */
		startpt = 1L << (minbits - 33);
	else
		startpt = 0;
	if (maxbits == 64)		/* Compute endpoint */
		endpt = 0xFFFFFFFF;
	else if (maxbits > 32)
		endpt = 1L << (maxbits - 32);
	else
		endpt = 1;

/* Load the factoring code */

	if (! loadfac ()) return (FALSE);

/* Loop until all the entire range is factored */

	increasing = (startp <= endp);
	if (!continuation) p = startp;
	if ((p & 1) == 0) p = increasing ? p + 1 : p - 1;
	for ( ; ; p = increasing ? p + 2 : p - 2) {
		unsigned long i;

/* Reached endp? */

		if (increasing) {
			if (p > endp) break;
		} else {
			if (p < endp) break;
		}

/* Is p a prime? */

		for (i = 3; i * i <= p; i += 2) {
			if (p % i == 0) goto nextp;
		}

/* Make sure p is in the primes file */

		if (fromfile) {
			unsigned long base = 0;
			fd = _open (DBFILE, _O_BINARY | _O_RDONLY);
			if (fd < 0) {
				OutputSomewhere (FILEERR);
				OutputSomewhere (FILEERR2);
				return (FALSE);
			}
			for ( ; ; ) {
				unsigned long testp;
				if (_read (fd, &testp, 4) < 4) {
					_close (fd);
					goto nextp;
				}
				if ((testp & 0xFF000000) == 0xFF000000) {
					base = testp >> 8;
					continue;
				}
				testp = base + (testp & 0xFFFFFF);
				if (p == testp) break;
				if (p < testp) {
					_close (fd);
					goto nextp;
				}
			}
			_close (fd);
		}

/* Sixteen passes! two for the 1 or 7 mod 8 factors times two for the */
/* 1 or 2 mod 3 factors times four for the 1, 2, 3, or 4 mod 5 factors. */

		if (!continuation) {
			pass = 1;
			temp_endpt = endpt;
			found_factor = FALSE;
		}
		for ( ; pass <= 16; pass++) {

/* Setup the factoring program */

		if (!continuation) {
			FACMSW = startpt;
			if (pass > 1 && FACMSW == 0) FACMSW = 1;
		}
		FACPASS = pass-1;
		FACLSW = p;
		setupf ();

/* Loop until all factors tested or factor found */

		iters = 0;
		iters_r = 0;
		totticks = 0.0;
		for ( ; ; ) {
			unsigned long temp, ticks;

/* Test for completion */

			if (FACMSW >= temp_endpt) goto nextpass;

/* Factor some more */

			if (CPU_TYPE >= 5) RDTSC (temp);
			if (factor64 () != 2) break;
			if (CPU_TYPE >= 5) {
				RDTSC (ticks);
				totticks += ticks - temp;
			}

/* Send queued messages to the server every so often */

			if ((iters & 0x7F) == 0) {
				if (!communicateWithServer ()) stopping = 1;

/* If this is a timed run, check if our time has expired */

				stopping |= timeCheck ();
			}

/* Set flag if we are stopping */

			stopping |= escapeCheck ();

/* Output informative message */

			if (++iters >= ITER_OUTPUT) {
				sprintf (buf, FACMSG, p, FACMSW, pass);
				OutputStr (buf);
				if (CPU_TYPE >= 5)
					OutputClocks (totticks / iters);
				else LineFeed ();
				sprintf (buf, "%ld", FACMSW);
				title (buf);
				iters = 0;
				totticks = 0.0;
			}

/* Output informative message */

			if (++iters_r >= ITER_OUTPUT_RES) {
				sprintf (buf, FACMSG, p, FACMSW, pass);
				strcat (buf, "\n");
				writeResults (buf);
				iters_r = 0;
			}

/* If an escape key was hit, write out the results and return */

			if (stopping || --countdown == 0) {
				short	three = 3;
				fd = _open (filename, _O_BINARY | _O_WRONLY | _O_TRUNC | _O_CREAT, 0666);
				_write (fd, &three, sizeof (short));
				_write (fd, &p, sizeof (long)); /* dummy */
				_write (fd, &startp, sizeof (long));
				_write (fd, &endp, sizeof (long));
				_write (fd, &minbits, sizeof (short));
				_write (fd, &maxbits, sizeof (short));
				_write (fd, &fromfile, sizeof (short));
				_write (fd, &p, sizeof (long));
				_write (fd, &FACMSW, sizeof (long));
				_write (fd, &pass, sizeof (short));
				_write (fd, &temp_endpt, sizeof (long));
				_write (fd, &found_factor, sizeof (short));
				_commit (fd);
				_close (fd);
				if (stopping) return (FALSE);
				countdown = write_counter;
			}
		}

/* Format the output message */

		makestr (FACMSW, FACLSW, str);

/* Output results */

		{
			struct primenetAssignmentResult pkt;
			memset (&pkt, 0, sizeof (pkt));
			pkt.exponent = p;
			pkt.resultType = PRIMENET_RESULT_FACTOR;
			strcpy (pkt.resultInfo.factor, str);
			spoolMessage (PRIMENET_ASSIGNMENT_RESULT, &pkt);
		}
		sprintf (buf, "M%ld has a factor: %s\n", p, str);
		OutputBoth (buf);
		spoolMessage (PRIMENET_RESULT_MESSAGE, buf);

/* Next pass */

		if (FACMSW != 0xFFFFFFFF) temp_endpt = FACMSW + 1;
		found_factor = TRUE;
nextpass:	continuation = FALSE;
		}

/* Output message if no factor found */

		if (! found_factor) {
			sprintf (buf, NOFAC,
				 p, minbits-1, maxbits, PORT, SEC4 (p));
			OutputBoth (buf);
			spoolMessage (PRIMENET_RESULT_MESSAGE, buf);
		}

/* Factor next prime */

nextp:		continuation = FALSE;
	}

/* Delete the continuation file */

	_unlink (filename);

/* All done */

	return (TRUE);
}


/* Test the factoring program */

void primeSieveTest ()
{
	char	buf[500];
	FILE	*fd;
	unsigned long p;

/* Load the factoring code */

	if (! loadfac ()) return;

/* Open factors file */

	fd = fopen ("factors", "r");

/* Loop until all the entire range is factored */

	while (fscanf (fd, "%ld", &p) && p) {
		unsigned long fachi, faclo;
		unsigned long i, facmid;
		char fac[480];
		char *f;

/* What is the factor? */

		fscanf (fd, "%s", fac);
		fachi = facmid = faclo = 0;
		for (f = fac; *f; f++) {
			if (*f < '0' || *f > '9') continue;
			if (fachi >= 429496729) {
				sprintf (buf, "%s factor too big.\n", fac);
				OutputBoth (buf);
				goto nextp;
			}
			faclo = faclo * 10 + *f - '0';
			facmid = facmid * 10 + (faclo >> 16); faclo &= 0xFFFF;
			fachi = fachi * 10 + (facmid >> 16); facmid &= 0xFFFF;
		}
		faclo = (facmid << 16) + faclo;

/* See if p is a prime */

		for (i = 3; i * i <= p; i += 2) {
			if (p % i == 0) {
				sprintf (buf, "%ld not a prime.\n", p);
				OutputBoth (buf);
				goto nextp;
			}
		}

/* Setup the factoring program */

		i = (fachi % 120 * 16 + faclo % 120) % 120;
		if (i == 1) FACPASS = 0;
		else if (i == 7) FACPASS = 1;
		else if (i == 17) FACPASS = 2;
		else if (i == 23) FACPASS = 3;
		else if (i == 31) FACPASS = 4;
		else if (i == 41) FACPASS = 5;
		else if (i == 47) FACPASS = 6;
		else if (i == 49) FACPASS = 7;
		else if (i == 71) FACPASS = 8;
		else if (i == 73) FACPASS = 9;
		else if (i == 79) FACPASS = 10;
		else if (i == 89) FACPASS = 11;
		else if (i == 97) FACPASS = 12;
		else if (i == 103) FACPASS = 13;
		else if (i == 113) FACPASS = 14;
		else if (i == 119) FACPASS = 15;
		else goto bad;
		FACMSW = fachi;
		FACLSW = p;
		setupf ();

/* Factor found, is it a match? */

		do {
			if (factor64 () != 2 &&
			    FACMSW == fachi && FACLSW == faclo) {
				sprintf (buf, "%ld factored OK.\n", p);
				OutputSomewhere (buf);
				goto nextp;
			}
		} while (FACMSW == fachi);

/* Uh oh. */

bad:		sprintf (buf, "%ld factor not found.\n", p);
		OutputBoth (buf);

/* If an escape key was hit, write out the results and return */

nextp:		if (escapeCheck ()) break;
		p = 0;
	}

/* All done */

	fclose (fd);
}
