/*----------------------------------------------------------------------
| This file contains routines and global variables that are common for
| all operating systems the program has been ported to.  It is included
| in one of the source code files of each port.  See common.h for the
| common #defines and common routine definitions.
|
| Commona contains information used only during setup
| Commonb contains information used only during execution
| Commonc contains information used during setup and execution
+---------------------------------------------------------------------*/

char	INI_FILE[20] = {0};
char	LOCALINI_FILE[20] = {0};
char	WORKTODO_FILE[20] = {0};
char	RESFILE[20] = {0};
char	SPOOL_FILE[20] = {0};
char	LOGFILE[20] = {0};

char	USERID[20] = {0};
char	USER_PWD[20] = {0};
char	OLD_USERID[20] = {0};
char	OLD_USER_PWD[20] = {0};
char	COMPID[20] = {0};
char	USER_NAME[80] = {0};
char	USER_ADDR[80] = {0};
int	SEND_EMAIL = 1;
int	USE_PRIMENET = 1;
int	USE_HTTP = 0;
int	DIAL_UP = 0;
short	WORK_PREFERENCE = 0;
unsigned int DAYS_OF_WORK = 20;
time_t	VACATION_END = 0;
int	ON_DURING_VACATION = 1;
int	ADVANCED_ENABLED = 0;
EXTERNC int volatile ERRCHK = 0;
unsigned int PRIORITY = 1;
unsigned int CPU_AFFINITY = 99;
int	MANUAL_COMM = 0;
EXTERNC unsigned int volatile CPU_TYPE = 0;
unsigned long volatile CPU_SPEED = 0;
unsigned int volatile CPU_HOURS = 0;
unsigned long volatile ITER_OUTPUT = 25;
unsigned long volatile ITER_OUTPUT_RES = 99999999;
unsigned long volatile DISK_WRITE_TIME = 30;
unsigned int MODEM_RETRY_TIME = 2;
unsigned int NETWORK_RETRY_TIME = 60;
unsigned int DAYS_BETWEEN_CHECKINS = 28;
int	TWO_BACKUP_FILES = 1;
int	SILENT_VICTORY = 0;
int	TRAY_ICON = FALSE;
int	HIDE_ICON = FALSE;
unsigned int ROLLING_AVERAGE = 0;

unsigned long EXP_BEING_WORKED_ON = 0;
int	EXP_BEING_FACTORED = 0;
double	EXP_PERCENT_COMPLETE = 0.0;

int	SPOOL_FILE_CHANGED = 0;
int	CHECK_WORK_QUEUE = 0;
int	GIMPS_QUIT = 0;
time_t	next_comm_time = 0;	/* Next time we should contact server */

char DBFILE[] = "database";
char FILEERR[] = "Database file not found.\n";
char FILEERR2[] = "Download database.zip from http://www.mersenne.org/range.htm\n";
char READFILEERR[] = "Error reading intermediate file: %s\n";

/* Try pentium-specific instruction (RDTSC) */

#if defined (__linux__) || defined (__EMX__)
#include <setjmp.h>
int	boom;
jmp_buf	env;
void sigboom_handler(int i)
{
	boom = TRUE;
	longjmp (env, 1);
}
int isPentium ()
{
	boom = FALSE;
	(void) signal (SIGILL, sigboom_handler);
	if (setjmp (env) == 0)
		__asm__ __volatile__ (".byte 0x0F\n .byte 0x31\n");
	(void) signal (SIGILL, SIG_DFL);
	return (!boom);
}
#elif defined (WIN31)
int isPentium ()
{
	return (TRUE);
}
#else
int isPentium ()
{
	int	rdtsc_succeeded;
	__try {
		__asm __emit 0x0F
		__asm __emit 0x31
		rdtsc_succeeded = TRUE;
	}
	__except (EXCEPTION_EXECUTE_HANDLER) {
		rdtsc_succeeded = FALSE;
	}
	return (rdtsc_succeeded);
}
#endif


/* Try pentium pro-specific instruction (CMOV) */

#if defined (__linux__) || defined (__EMX__)
int isPentiumPro ()
{
	boom = FALSE;
	(void) signal (SIGILL, sigboom_handler);
	if (setjmp (env) == 0)
		__asm__ __volatile__ (".byte 0x0F\n .byte 0x42\n .byte 0xC0\n");
	(void) signal (SIGILL, SIG_DFL);
	return (!boom);
}
#elif defined (WIN31)
int isPentiumPro ()
{
	return (TRUE);
}
#else
int isPentiumPro ()
{
	int	cmov_succeeded;
	__try {
		__asm __emit 0x0F
		__asm __emit 0x42
		__asm __emit 0xC0
		cmov_succeeded = TRUE;
	}
	__except (EXCEPTION_EXECUTE_HANDLER) {
		cmov_succeeded = FALSE;
	}
	return (cmov_succeeded);
}
#endif

/* Determine if a number is prime */

int isPrime (
	unsigned long p)
{
	unsigned long i;
	for (i = 2; i * i <= p; i = (i + 1) | 1)
		if (p % i == 0) return (FALSE);
	return (TRUE);
}

/* Determine the names of the INI files */

void nameIniFiles (
	int	named_ini_files)
{
	if (named_ini_files < 0) {
		strcpy (INI_FILE, "prime.ini");
		strcpy (LOCALINI_FILE, "local.ini");
		strcpy (SPOOL_FILE, "prime.spl");
		strcpy (WORKTODO_FILE, "worktodo.ini");
		strcpy (RESFILE, "results.txt");
		strcpy (LOGFILE, "prime.log");
	} else {
		sprintf (INI_FILE, "prim%04d.ini", named_ini_files);
		sprintf (LOCALINI_FILE, "loca%04d.ini", named_ini_files);
		sprintf (SPOOL_FILE, "prim%04d.spl", named_ini_files);
		sprintf (WORKTODO_FILE, "work%04d.ini", named_ini_files);
		sprintf (RESFILE, "resu%04d.txt", named_ini_files);
		sprintf (LOGFILE, "prim%04d.log", named_ini_files);
	}
}

/* Read the INI files */

void readIniFiles ()
{
	IniGetString (INI_FILE, "UserID", USERID, sizeof (USERID), NULL);
	USERID[PRIMENET_USER_ID_LENGTH] = 0;
	IniGetString (INI_FILE, "UserPWD", USER_PWD, sizeof (USER_PWD), NULL);
	USER_PWD[PRIMENET_USER_PW_LENGTH] = 0;
	IniGetString (INI_FILE, "OldUserID", OLD_USERID, sizeof (OLD_USERID), NULL);
	OLD_USERID[PRIMENET_USER_ID_LENGTH] = 0;
	IniGetString (INI_FILE, "OldUserPWD", OLD_USER_PWD, sizeof (OLD_USER_PWD), NULL);
	OLD_USER_PWD[PRIMENET_USER_PW_LENGTH] = 0;
	IniGetString (INI_FILE, "UserName", USER_NAME,
		      sizeof (USER_NAME), NULL);
	IniGetString (INI_FILE, "UserEmailAddr", USER_ADDR,
		      sizeof (USER_ADDR), NULL);
	IniGetString (LOCALINI_FILE, "ComputerID", COMPID,
		      sizeof (COMPID), NULL);
	COMPID[PRIMENET_COMPUTER_ID_LENGTH] = 0;
	SEND_EMAIL = (int) IniGetInt (INI_FILE, "SendEmail", 1);
	USE_PRIMENET = (int) IniGetInt (INI_FILE, "UsePrimenet", 1);
	USE_HTTP = (int) IniGetInt (INI_FILE, "UseHTTP", 99);
	if (USE_HTTP == 99) {	/* Guess value for users prior to V16.5 */
		int	fd;
		fd = _open ("PRIMENET.DLL", _O_BINARY | _O_RDONLY);
		if (fd < 0) USE_HTTP = 0;
		else {
			unsigned long filelen;
			filelen = _lseek (fd, 0L, SEEK_END);
			_close (fd);
			_unlink ("PRIMENET.DLL");
			USE_HTTP = (filelen < 40000) ? 0 : 1;
		}
		IniWriteInt (INI_FILE, "UseHTTP", USE_HTTP);
	}
	DIAL_UP = (int) IniGetInt (INI_FILE, "DialUp", 0);
	DAYS_OF_WORK = (unsigned int) IniGetInt (INI_FILE, "DaysOfWork", 14);
	if (DAYS_OF_WORK > 180) DAYS_OF_WORK = 180;
	WORK_PREFERENCE = (short) IniGetInt (INI_FILE, "WorkPreference", 0);

        VACATION_END = IniGetInt (LOCALINI_FILE, "VacationEnd", 0);
        ON_DURING_VACATION = (int) IniGetInt (LOCALINI_FILE, "VacationOn", 1);

	CPU_TYPE = (unsigned int) IniGetInt (LOCALINI_FILE, "CPUType", 0);
	CPU_SPEED = IniGetInt (LOCALINI_FILE, "CPUSpeed", 0);
	CPU_HOURS = (unsigned int) IniGetInt (LOCALINI_FILE, "CPUHours", 24);
	if (CPU_HOURS < 1) CPU_HOURS = 1;
	if (CPU_HOURS > 24) CPU_HOURS = 24;

	ROLLING_AVERAGE = (unsigned int) IniGetInt (LOCALINI_FILE, "RollingAverage", 1000);
	if (ROLLING_AVERAGE < 10) ROLLING_AVERAGE = 10;
	if (ROLLING_AVERAGE > 4000) ROLLING_AVERAGE = 4000;

	ITER_OUTPUT = IniGetInt (INI_FILE, "OutputIterations", 100);
	if (ITER_OUTPUT <= 0) ITER_OUTPUT = 1;
	ITER_OUTPUT_RES = IniGetInt (INI_FILE, "ResultsFileIterations",
				     99999999);
	if (ITER_OUTPUT_RES < 1000) ITER_OUTPUT_RES = 1000;
	DISK_WRITE_TIME = IniGetInt (INI_FILE, "DiskWriteTime", 30);
	MODEM_RETRY_TIME = (unsigned int) IniGetInt (INI_FILE, "NetworkRetryTime", 2);
	if (MODEM_RETRY_TIME < 1) MODEM_RETRY_TIME = 1;
	if (MODEM_RETRY_TIME > 300) MODEM_RETRY_TIME = 300;
	NETWORK_RETRY_TIME = (unsigned int)
		IniGetInt (INI_FILE, "NetworkRetryTime2",
			   MODEM_RETRY_TIME > 60 ? MODEM_RETRY_TIME : 60);
	if (NETWORK_RETRY_TIME < 1) NETWORK_RETRY_TIME = 1;
	if (NETWORK_RETRY_TIME > 300) NETWORK_RETRY_TIME = 300;
	DAYS_BETWEEN_CHECKINS = (unsigned int) IniGetInt (INI_FILE, "DaysBetweenCheckins",28);
	if (DAYS_BETWEEN_CHECKINS > 60) DAYS_BETWEEN_CHECKINS = 60;
	if (DAYS_BETWEEN_CHECKINS < 1) DAYS_BETWEEN_CHECKINS = 1;
	TWO_BACKUP_FILES = (int) IniGetInt (INI_FILE, "TwoBackupFiles", 1);
	SILENT_VICTORY = (int) IniGetInt (INI_FILE, "SilentVictory", 0);

	ADVANCED_ENABLED = (int) IniGetInt (INI_FILE, "Advanced", 0);
	ERRCHK = (int) IniGetInt (INI_FILE, "ErrorCheck", 0);
	PRIORITY = (unsigned int) IniGetInt (INI_FILE, "Priority", 1);
	CPU_AFFINITY = (unsigned int) IniGetInt (INI_FILE, "Affinity", 99);
	MANUAL_COMM = (int) IniGetInt (INI_FILE, "ManualComm", 0);
	HIDE_ICON = (int) IniGetInt (INI_FILE, "HideIcon", 0);
	TRAY_ICON = (int) IniGetInt (INI_FILE, "TrayIcon", 0);

/* Guess the CPU type if it isn't known.  Otherwise, validate it. */

	if (CPU_TYPE == 0 || CPU_SPEED == 0) guessCpuType ();
	if (CPU_TYPE == 6 && !isPentiumPro ()) CPU_TYPE = 5;
	if (CPU_TYPE >= 5 && !isPentium ()) CPU_TYPE = 4;
	if (CPU_SPEED > 3000) CPU_SPEED = 3000;
	if (CPU_SPEED < 25) CPU_SPEED = 25;

/* Make sure the Lock user info bit is sent to the server */

	if (IniGetInt (INI_FILE, "LockUserInfo", 0) &&
	    ! IniGetInt (INI_FILE, "LockUserInfoSent", 0)) {
		spoolMessage (PRIMENET_MAINTAIN_USER_INFO, NULL);
		IniWriteInt (INI_FILE, "LockUserInfoSent", 1);
	}
}

/*----------------------------------------------------------------------
| Portable routines to read and write ini files!  NOTE:  These only
| work if you open no more than 5 ini files.  Also you must not
| change the working directory at any time during program execution.
+---------------------------------------------------------------------*/

struct IniLine {
	char	*keyword;
	char	*value;
};
struct IniCache {
	char	*filename;
	int	immediate_writes;
	int	dirty;
	unsigned int num_lines;
	unsigned int array_size;
	struct IniLine **lines;
};

void growIniLineArray (
	struct IniCache *p)
{
	struct IniLine **newlines;

	if (p->num_lines != p->array_size) return;

	newlines = (struct IniLine **)
		malloc ((p->num_lines + 100) * sizeof (struct IniLine **));
	if (p->num_lines) {
		memcpy (newlines, p->lines, p->num_lines * sizeof (struct IniLine *));
		free (p->lines);
	}
	p->lines = newlines;
	p->array_size = p->num_lines + 100;
}

struct IniCache *openIniFile (
	char	*filename,
	int	forced_read)
{
static	struct IniCache *cache[10] = {0};
	struct IniCache *p;
	FILE	*fd;
	unsigned int i;
	char	line[80];
	char	*val;

/* See if file is cached */

	for (i = 0; i < 10; i++) {
		p = cache[i];
		if (p == NULL) {
			p = (struct IniCache *) malloc (sizeof (struct IniCache));
			p->filename = (char *) malloc (strlen (filename) + 1);
			strcpy (p->filename, filename);
			p->immediate_writes = 1;
			p->dirty = 0;
			p->num_lines = 0;
			p->array_size = 0;
			p->lines = NULL;
			forced_read = 1;
			cache[i] = p;
			break;
		}
		if (strcmp (filename, p->filename) == 0)
			break;
	}

/* Skip reading the ini file if appropriate */

	if (!forced_read) return (p);
	if (p->dirty) return (p);

/* Free the data if we've already read some in */

	for (i = 0; i < p->num_lines; i++) {
		free (p->lines[i]->keyword);
		free (p->lines[i]->value);
		free (p->lines[i]);
	}
	p->num_lines = 0;

/* Read the IniFile */
	
	fd = fopen (filename, "r");
	if (fd == NULL) return (p);

	while (fgets (line, 80, fd)) {
		if (line[strlen(line)-1] == '\n') line[strlen(line)-1] = 0;
		if (line[0] == 0) continue;
		if (line[strlen(line)-1] == '\r') line[strlen(line)-1] = 0;
		if (line[0] == 0) continue;

		val = strchr (line, '=');
		if (val == NULL) {
			char	buf[130];
			sprintf (buf, "Illegal line in INI file: %s\n", line);
			OutputSomewhere (buf);
			continue;
		}
		*val++ = 0;

		growIniLineArray (p);
		
/* Allocate and fill in a new line structure */

		i = p->num_lines++;
		p->lines[i] = (struct IniLine *) malloc (sizeof (struct IniLine));
		p->lines[i]->keyword = (char *) malloc (strlen (line) + 1);
		p->lines[i]->value = (char *) malloc (strlen (val) + 1);
		strcpy (p->lines[i]->keyword, line);
		strcpy (p->lines[i]->value, val);
	}
	fclose (fd);

	return (p);
}

void writeIniFile (
	struct IniCache *p)
{
	int	fd;
	unsigned int j;
	char	buf[100];

/* Delay writing the file unless this INI file is written */
/* to immediately */

	if (!p->immediate_writes) {
		p->dirty = 1;
		return;
	}

/* Create and write out the INI file */

	fd = _open (p->filename, _O_CREAT | _O_TRUNC | _O_WRONLY | _O_TEXT, 0666);
	if (fd < 0) return;
	for (j = 0; j < p->num_lines; j++) {
		strcpy (buf, p->lines[j]->keyword);
		strcat (buf, "=");
		strcat (buf, p->lines[j]->value);
		strcat (buf, "\n");
		_write (fd, buf, strlen (buf));
	}
	p->dirty = 0;
	_close (fd);
}

void truncated_strcpy (
	char	*buf,
	unsigned int bufsize,
	char	*val)
{
	if (strlen (val) >= bufsize) {
		memcpy (buf, val, bufsize-1);
		buf[bufsize-1] = 0;
	} else {
		strcpy (buf, val);
	}
}

void IniGetString (
	char	*filename,
	char	*keyword,
	char	*val,
	unsigned int val_bufsize,
	char	*default_val)
{
	struct IniCache *p;
	unsigned int i;

/* Open ini file */

	p = openIniFile (filename, 0);

/* Look for the keyword */

	for (i = 0; ; i++) {
		if (i == p->num_lines) {
			if (default_val == NULL) {
				val[0] = 0;
			} else {
				truncated_strcpy (val, val_bufsize, default_val);
			}
			return;
		}
		if (stricmp (keyword, p->lines[i]->keyword) == 0) break;
	}

/* Copy info from the line structure to the user buffers */

	truncated_strcpy (val, val_bufsize, p->lines[i]->value);
}

long IniGetInt (
	char	*filename,
	char	*keyword,
	long	default_val)
{
	char	buf[20], defval[20];
	sprintf (defval, "%ld", default_val);
	IniGetString (filename, keyword, buf, 20, defval);
	return (atol (buf));
}

void IniWriteString (
	char	*filename,
	char	*keyword,
	char	*val)
{
	struct IniCache *p;
	unsigned int i;

/* Open ini file */

	p = openIniFile (filename, 1);

/* Look for the keyword */

	for (i = 0; ; i++) {
		if (i == p->num_lines) {

/* Make sure the line array has room for the new line */

			growIniLineArray (p);

/* Allocate and fill in a new line structure */

			p->lines[i] = (struct IniLine *) malloc (sizeof (struct IniLine));
			p->lines[i]->keyword = (char *) malloc (strlen (keyword) + 1);
			strcpy (p->lines[i]->keyword, keyword);
			p->lines[i]->value = NULL;
			p->num_lines++;
			break;
		}
		if (stricmp (keyword, p->lines[i]->keyword) == 0) {
			if (strcmp (val, p->lines[i]->value) == 0) return;
			break;
		}
	}

/* Replace the value associated with the keyword */

	free (p->lines[i]->value);
	p->lines[i]->value = (char *) malloc (strlen (val) + 1);
	strcpy (p->lines[i]->value, val);

/* Write the INI file back to disk */

	writeIniFile (p);
}

void IniWriteInt (
	char	*filename,
	char	*keyword,
	long	val)
{
	char	buf[20];
	sprintf (buf, "%ld", val);
	IniWriteString (filename, keyword, buf);
}

void IniFileOpen (
	char	*filename,
	int	immediate_writes)
{
	struct IniCache *p;
	p = openIniFile (filename, 1);
	p->immediate_writes = immediate_writes;
}

void IniFileClose (
	char	*filename)
{
	struct IniCache *p;
	p = openIniFile (filename, 0);
	p->immediate_writes = 1;
	writeIniFile (p);
}

unsigned int IniGetNumLines (
	char	*filename)
{
	struct IniCache *p;
	p = openIniFile (filename, 0);
	return (p->num_lines);
}

void IniGetLineAsString (
	char	*filename,
	unsigned int line,
	char	*keyword,
	unsigned int keyword_bufsize,
	char	*val,
	unsigned int val_bufsize)
{
	struct IniCache *p;

/* Open ini file */

	p = openIniFile (filename, 0);

/* Copy info from the line structure to the user buffers */

	truncated_strcpy (keyword, keyword_bufsize, p->lines[line-1]->keyword);
	truncated_strcpy (val, val_bufsize, p->lines[line-1]->value);
}

void IniGetLineAsInt (
	char	*filename,
	unsigned int line,
	char	*keyword,
	unsigned int keyword_bufsize,
	long	*val)
{
	char	buf[20];
	IniGetLineAsString (filename, line, keyword, keyword_bufsize, buf, 20);
	*val = atol (buf);
}

void IniReplaceLineAsString (
	char	*filename,
	unsigned int line,
	char	*keyword,
	char	*val)
{
	IniDeleteLine (filename, line);
	IniInsertLineAsString (filename, line, keyword, val);
}

void IniReplaceLineAsInt (
	char	*filename,
	unsigned int line,
	char	*keyword,
	long	val)
{
	char	buf[20];
	sprintf (buf, "%ld", val);
	IniReplaceLineAsString (filename, line, keyword, buf);
}

void IniInsertLineAsString (
	char	*filename,
	unsigned int line,
	char	*keyword,
	char	*val)
{
	struct IniCache *p;
	unsigned int i;

/* Open ini file, do not reread it as that could change the line numbers! */

	p = openIniFile (filename, 0);

/* Adjust line number if it doesn't make sense */

	if (line == 0) line = 1;
	if (line > p->num_lines+1) line = p->num_lines+1;

/* Make sure the line array has room for the new line */

	growIniLineArray (p);

/* Shuffle lines down in the array to make room for the new line */

	for (i = p->num_lines; i >= line; i--) p->lines[i] = p->lines[i-1];
	p->num_lines++;

/* Allocate and fill in a new line structure */

	p->lines[line-1] = (struct IniLine *) malloc (sizeof (struct IniLine));
	p->lines[line-1]->keyword = (char *) malloc (strlen (keyword) + 1);
	p->lines[line-1]->value = (char *) malloc (strlen (val) + 1);
	strcpy (p->lines[line-1]->keyword, keyword);
	strcpy (p->lines[line-1]->value, val);

/* Write the INI file back to disk */

	writeIniFile (p);
}

void IniInsertLineAsInt (
	char	*filename,
	unsigned int line,
	char	*keyword,
	long	val)
{
	char	buf[20];
	sprintf (buf, "%ld", val);
	IniInsertLineAsString (filename, line, keyword, buf);
}

void IniAppendLineAsString (
	char	*filename,
	char	*keyword,
	char	*val)
{
	struct IniCache *p;
	p = openIniFile (filename, 0);
	IniInsertLineAsString (filename, p->num_lines+1, keyword, val);
}

void IniAppendLineAsInt (
	char	*filename,
	char	*keyword,
	long	val)
{
	char	buf[20];
	sprintf (buf, "%ld", val);
	IniAppendLineAsString (filename, keyword, buf);
}

void IniDeleteLine (
	char	*filename,
	unsigned int line)
{
	struct IniCache *p;
	unsigned int i;

/* Open ini file, do not reread it as that could change the line numbers! */

	p = openIniFile (filename, 0);
	if (line == 0 || line > p->num_lines) return;

/* Free the data associated with the given line */

	free (p->lines[line-1]->keyword);
	free (p->lines[line-1]->value);
	free (p->lines[line-1]);

/* Delete the line from the lines array */

	for (i = line; i < p->num_lines; i++) p->lines[i-1] = p->lines[i];
	p->num_lines--;

/* Write the INI file back to disk */

	writeIniFile (p);
}

void IniDeleteAllLines (
	char	*filename)
{
	struct IniCache *p;
	unsigned int i;

/* Open ini file! */

	p = openIniFile (filename, 0);

/* Free the data associated with the given line */

	for (i = 0; i < p->num_lines; i++) {
		free (p->lines[i]->keyword);
		free (p->lines[i]->value);
		free (p->lines[i]);
	}
	p->num_lines = 0;

/* Write the INI file back to disk */

	writeIniFile (p);
}

/*----------------------------------------------------------------------+
| Portable routines to read and write messages in the spool file.	|
+----------------------------------------------------------------------*/

#include "security.c"
#ifndef IPSHASH
void hash_packet (short operation, void *pkt)
{
	struct primenetUserInfo *z = (struct primenetUserInfo *) pkt;
	if (operation == PRIMENET_PING_SERVER_INFO) return;
	z->salt = 0;
	z->hash = 0;
}
#endif

/* Update completion dates on the server.  Set a flag in */
/* the spool file saying this is necessary. */

void UpdateEndDates ()
{
	spoolMessage (PRIMENET_COMPLETION_DATE, NULL);
}

/* Update completion dates on the server if it has been a */
/* month since we last updated the server. */

void ConditionallyUpdateEndDates ()
{
#ifndef SERVER_TESTING
	time_t	start_time, current_time;

/* Get the current time and when the completion dates were last sent */

	time (&current_time);
	start_time = IniGetInt (LOCALINI_FILE, "LastEndDatesSent", 0);

/* If it's been the correct number of days, then update the end dates */

	if (current_time < start_time ||
	    current_time > (time_t) (start_time + DAYS_BETWEEN_CHECKINS * 86400))
		UpdateEndDates ();
#endif
}

/* Write a message to the spool file */

void spoolMessage (
	short	msgType,
	void	*msg)
{
	int	fd;
	char	header_byte;

/* If we're not using primenet, ignore this call */

	if (!USE_PRIMENET) goto leave;

/* Open the spool file */

	fd = _open (SPOOL_FILE, _O_RDWR | _O_BINARY | _O_CREAT, 0666);
	if (fd < 0) {
		LogMsg ("ERROR: Unable to open spool file.\n");
		goto leave;
	}

/* Get the current header byte */

	header_byte = 0;
	_read (fd, &header_byte, 1);

/* If this is a maintain user info message, then set */
/* the header byte appropriately */

	if (msgType == PRIMENET_MAINTAIN_USER_INFO)
		header_byte |= 0x4;

/* If this is a set computer info message, then set */
/* the header byte appropriately */

	else if (msgType == PRIMENET_SET_COMPUTER_INFO)
		header_byte |= 0x20;

/* If this is an update completion dates message, then set */
/* the header byte appropriately */

	else if (msgType == PRIMENET_COMPLETION_DATE)
		header_byte |= 0x8;

/* Ugly little hack when quitting GIMPS */

	else if (msgType == 999)
		header_byte |= 0x10;

/* See if this is an important message */

	else if (msgType == PRIMENET_ASSIGNMENT_RESULT)
		header_byte |= 0x2;
	else
		header_byte |= 0x1;

/* Write the new header byte */

	_lseek (fd, 0, SEEK_SET);
	_write (fd, &header_byte, 1);

/* Write out a full message */

	if (msgType == PRIMENET_RESULT_MESSAGE ||
	    msgType == PRIMENET_ASSIGNMENT_RESULT) {
		char	buf[1024];
		short	datalen;
		union {
			struct primenetResultMessage msg;
			struct primenetAssignmentResult res;
		} x;

/* Skip the remaining messages */

		while (_read (fd, buf, sizeof (buf)));

/* Append the latest message */

		if (msgType == PRIMENET_RESULT_MESSAGE) {
			memset (&x.msg, 0, sizeof (x.msg));
			if (strlen ((char *) msg) >= sizeof (x.msg.message)) {
				memcpy (x.msg.message, msg, sizeof (x.msg.message));
				x.msg.message[sizeof(x.msg.message)-1] = 0;
			} else
				strcpy (x.msg.message, (char *) msg);
			datalen = sizeof (struct primenetResultMessage);
		} else {
			datalen = sizeof (struct primenetAssignmentResult);
			memcpy (&x, msg, datalen);
		}
		_write (fd, &msgType, sizeof (short));
		_write (fd, &datalen, sizeof (short));
		_write (fd, &x, datalen);
	}

/* Set flag that the spool file has changed */

	SPOOL_FILE_CHANGED = 1;

/* Close the spool file */

	_close (fd);

/* If this is a maintain user info message, then also */
/* write a message to the results file */

leave:	if (msgType == PRIMENET_MAINTAIN_USER_INFO &&
	    (USER_NAME[0] != 0 || USER_ADDR[0] != 0)) {
		char	buf[200];
		if (USERID[0])
			sprintf (buf, "UID: %s, User: %s, %s\n",
				 USERID, USER_NAME, USER_ADDR);
		else
			sprintf (buf, "User: %s, %s\n", USER_NAME, USER_ADDR);
		writeResults (buf);
		spoolMessage (PRIMENET_RESULT_MESSAGE, buf);
		if (OLD_USERID[0]) {
			sprintf (buf, "Change UID from %s to %s\n",
				 OLD_USERID, USERID);
			writeResults (buf);
			spoolMessage (PRIMENET_RESULT_MESSAGE, buf);
		}
	}
}

/* Read a spooled message */

void readMessage (
	int	fd,
	long	*offset,	/* Offset of the message */
	short	*msgType,	/* 0 = no message */
	void	*msg)
{
	short	datalen;

/* Loop until a message that hasn't already been sent is found */

	for ( ; ; ) {
		*offset = _lseek (fd, 0, SEEK_CUR);
		if (_read (fd, msgType, sizeof (short)) != sizeof (short))
			goto err;
		if (_read (fd, &datalen, sizeof (short)) != sizeof (short))
			goto err;

/* Upgrade old style spooled messages */

		if (*msgType == PRIMENET_RESULT_MESSAGE &&
		    datalen == sizeof (struct primenet2ResultMessage)) {
			struct primenet2ResultMessage v2msg;
			struct primenetResultMessage *v3msg;
			v3msg = (struct primenetResultMessage *) msg;
			if (_read (fd, &v2msg, datalen) != datalen) goto err;
			memset (msg, 0, sizeof (struct primenetResultMessage));
			strcpy (v3msg->message, v2msg.message);
		} else if (*msgType == PRIMENET_ASSIGNMENT_RESULT &&
			   datalen==sizeof (struct primenet2AssignmentResult)){
			struct primenet2AssignmentResult v2msg;
			struct primenetAssignmentResult *v3msg;
			v3msg = (struct primenetAssignmentResult *) msg;
			if (_read (fd, &v2msg, datalen) != datalen) goto err;
			memset (msg, 0,
				sizeof (struct primenetAssignmentResult));
			v3msg->exponent = v2msg.exponent;
			v3msg->resultType = v2msg.resultType;
			memcpy (&v3msg->resultInfo, &v2msg.resultInfo,
				sizeof (v2msg.resultInfo));
		}

/* Read the body of the message */

		else
			if (_read (fd, msg, datalen) != datalen) goto err;

/* Loop if message has already been sent */

		if (*msgType != -1) break;
	}
	return;
err:	*msgType = 0;
}


/* Send a message that was read from the spool file */

int sendMessage (
	short	msgType,
	void	*msg)
{
	struct primenetResultMessage *pkt;
	struct primenetResultMessage local_pkt;
	short	structSize;
	int	return_code;
	char	buf[200];

/* Load the primenet library and do any special handling */
/* required prior to calling primenet */

	if (!LoadPrimeNet ()) return (PRIMENET_ERROR_MODEM_OFF);

/* Prepend all messages with the userid and computer id */

	pkt = (struct primenetResultMessage *) msg;
	if (msgType == PRIMENET_RESULT_MESSAGE) {
		memcpy (&local_pkt, pkt, sizeof(struct primenetResultMessage));
		if (USERID[0] == 0) 
			sprintf (local_pkt.message, "%s", pkt->message);
		else if (COMPID[0] == 0)
			sprintf (local_pkt.message, "UID: %s, %s",
				 USERID, pkt->message);
		else
			sprintf (local_pkt.message, "UID: %s/%s, %s",
				 USERID, COMPID, pkt->message);
		pkt = &local_pkt;
	}

/* Print a message on the screen and in the log file */

	switch (msgType) {
	case PRIMENET_PING_SERVER_INFO:
		OutputStr ("Contacting PrimeNet Server.\n");
		structSize = sizeof (struct primenetPingServerInfo);
		break;
	case PRIMENET_MAINTAIN_USER_INFO:
		LogMsg ("Updating user information on the server\n");
		structSize = sizeof (struct primenetUserInfo);
		break;
	case PRIMENET_SET_COMPUTER_INFO:
		LogMsg ("Updating computer information on the server\n");
		structSize = sizeof (struct primenetComputerInfo);
		break;
	case PRIMENET_GET_ASSIGNMENT:
		LogMsg ("Getting exponents from server\n");
		structSize = sizeof (struct primenetGetAssignment);
		break;
	case PRIMENET_COMPLETION_DATE:
		{
			time_t	this_time;
			char	timebuf[30];
			time (&this_time);
			this_time += ((struct primenetCompletionDate *)pkt)->days * 86400;
			strcpy (timebuf, ctime (&this_time)+4);
			strcpy (timebuf+12, timebuf+15);
			sprintf (buf, "Sending expected completion date for M%ld: %s",
				((struct primenetCompletionDate *)pkt)->exponent, timebuf);
			LogMsg (buf);
		}
		structSize = sizeof (struct primenetCompletionDate);
		break;
	case PRIMENET_RESULT_MESSAGE:
		LogMsg ("Sending text message to server:\n");
		LogMsg (pkt->message);
		structSize = sizeof (struct primenetResultMessage) - 200 +
			strlen (pkt->message) + 1;
		break;
	case PRIMENET_ASSIGNMENT_RESULT:
		if (((struct primenetAssignmentResult *)pkt)->resultType == PRIMENET_RESULT_UNRESERVE)
			sprintf (buf, "Unreserving exponent %ld\n",
				 ((struct primenetAssignmentResult *)pkt)->exponent);
		else
			sprintf (buf, "Sending result to server for exponent %ld\n",
				 ((struct primenetAssignmentResult *)pkt)->exponent);
		LogMsg (buf);
		structSize = sizeof (struct primenetAssignmentResult);
		break;
	}

/* Fill in the common header fields */

	pkt->versionNumber = PRIMENET_VERSION;
	pkt->structSize = structSize;
	if (msgType != PRIMENET_PING_SERVER_INFO) {
		strcpy (pkt->userID, USERID);
		strcpy (pkt->userPW, USER_PWD);
		strcpy (pkt->computerID, COMPID);
		pkt->hashFunction = PRIMENET_HASHFUNC_WOLTMAN_1;
	}
	hash_packet (msgType, pkt);

/* Send the message */

	return_code = (*PRIMENET)(msgType, pkt);

/* Print message based on return code */

	switch (return_code) {
	case PRIMENET_ERROR_OK:
		break;
	case PRIMENET_ERROR_CONNECT_FAILED:
		sprintf (buf, "ERROR %d: Server unavailable\n", return_code);
		if (msgType == PRIMENET_PING_SERVER_INFO)
			OutputStr (buf);
		else
			LogMsg (buf);
		break;
	case PRIMENET_ERROR_SERVER_BUSY:
		sprintf (buf, "ERROR %d: Server busy\n", return_code);
		if (msgType == PRIMENET_PING_SERVER_INFO)
			OutputStr (buf);
		else
			LogMsg (buf);
		return_code = PRIMENET_ERROR_CONNECT_FAILED;
		break;
	case PRIMENET_ERROR_NO_ASSIGNMENT:
		sprintf (buf, "ERROR %d: Server has run out of exponents to assign.\n", return_code);
		LogMsg (buf);
		break;
	case PRIMENET_ERROR_ACCESS_DENIED:
		sprintf (buf, "ERROR %d: UserID/Password error.\n", return_code);
		LogMsg (buf);
		if (OLD_USERID[0]) {
			LogMsg ("Old userID and password will be used.\n");
			strcpy (USERID, OLD_USERID);
			strcpy (USER_PWD, OLD_USER_PWD);
			OLD_USERID[0] = 0;
			OLD_USER_PWD[0] = 0;
		} else {
			LogMsg ("A new userID and password will be generated.\n");
			USERID[0] = 0;
			USER_PWD[0] = 0;
		}
		spoolMessage (PRIMENET_MAINTAIN_USER_INFO, NULL);
		break;
	case PRIMENET_ERROR_ALREADY_TESTED:
		sprintf (buf, "ERROR %d: Exponent already tested.\n", return_code);
		LogMsg (buf);
		if (msgType == PRIMENET_ASSIGNMENT_RESULT) return_code = 0;
		break;
	case PRIMENET_ERROR_UNASSIGNED:
	case PRIMENET_ERROR_NOT_PERMITTED:
		sprintf (buf, "ERROR %d: Exponent not assigned to us.\n", return_code);
		LogMsg (buf);
		if (msgType == PRIMENET_ASSIGNMENT_RESULT) return_code = 0;
		break;
	default:
		sprintf (buf, "ERROR: Primenet error: %d\n", return_code);
		LogMsg (buf);
		if (msgType == PRIMENET_ASSIGNMENT_RESULT) return_code = 0;
		break;
	}

/* Return the return code */

	return (return_code);
}


/* Copy an existing results file to the spool file */
/* This is used when converting from manual to automatic mode */
/* It provides an extra chance that the existing results file */
/* will get sent to us. */

void spoolExistingResultsFile ()
{
	int	i;
	char	*filename;
	char	line[256];
	FILE	*fd;

	for (i = 0; i < 2; i++) {
		if (i == 0) filename = "results";
		if (i == 1) filename = "results.txt";
		if (i == 2) {
			if (!strcmp (RESFILE, "results")) continue;
			if (!strcmp (RESFILE, "results.txt")) continue;
			filename = RESFILE;
		}
		fd = fopen (filename, "r");
		if (fd == NULL) continue;
		while (fgets (line, sizeof (line) - 1, fd)) {
			if (line[0] == '[') continue;
			if (strstr (line, "Self-test")) continue;
			if (strstr (line, "Factored")) continue;
			if (strstr (line, " / ")) continue;
			if (strstr (line, "Possible")) continue;
			if (strstr (line, "Continuing")) continue;
			if (strstr (line, "no more")) continue;
			if (strstr (line, "Please send")) continue;
			if (strstr (line, "new range")) continue;
			if (strstr (line, "Quit")) continue;
			if (line[0] == 'U' && strchr (line, ',') != NULL)
				strcpy (line, strchr (line, ',') + 2);
			spoolMessage (PRIMENET_RESULT_MESSAGE, line);
		}
		fclose (fd);
	}
}

/* Output string to screen or results file */

void OutputSomewhere (
	char	*buf)
{
	if (NO_GUI) writeResults (buf);
	else OutputStr (buf);
}

/* Output string to both the screen and results file */

void OutputBoth (
	char	*buf)
{
	OutputStr (buf);
	writeResults (buf);
}

/* Output message to screen and prime.log file */

void LogMsg (
	char	*str)
{
	int	fd;
	unsigned long filelen;
static	time_t	last_time = 0;
	time_t	this_time;

/* Output it to the screen */

	OutputStr (str);

/* Open the log file and position to the end */

	fd = _open (LOGFILE, _O_TEXT | _O_RDWR | _O_CREAT, 0666);
	if (fd < 0) {
		OutputStr ("Unable to open log file.\n");
		return;
	}
	filelen = _lseek (fd, 0L, SEEK_END);

/* Output to the log file only if it hasn't grown too big */

	if (filelen < 250000) {

/* If it has been at least 5 minutes since the last time stamp */
/* was output, then output a new timestamp */

		time (&this_time);
		if (this_time - last_time > 300) {
			char	buf[48];
			last_time = this_time;
			buf[0] = '[';
			strcpy (buf+1, ctime (&this_time));
			sprintf (buf+25, " - ver %s]\n", VERSION);
			_write (fd, buf, strlen (buf));
		}

/* Output the message */

		_write (fd, str, strlen (str));
	}

/* Display message about full log file */
	
	else {
		char	*fullmsg = "Prime.log file full.  Please delete it.\n";
		OutputStr (fullmsg);
		if (filelen < 251000)
			_write (fd, fullmsg, strlen (fullmsg));
	}
	_close (fd);
}


/* Determine the preferred work type (based on CPU speed) */
/* Anything slower than a 45 MHz Pentium should */
/* probably be relegated to just factoring */

short default_work_type ()
{
	if (map_fftlen_to_timing (262144) *
			24.0 / CPU_HOURS *
			1000.0 / ROLLING_AVERAGE > 1.556)
		return (PRIMENET_ASSIGN_FACTOR);
	return (PRIMENET_ASSIGN_TEST);
}

/* Parse a line from the work-to-do file */

int parseWorkToDoLine (
	unsigned int line,	/* Line number to parse */
	struct work_unit *w)	/* Returned type of work */
{
	char	keyword[20];
	char	value[80];
	char	*comma;
		
/* Get the line from work-to-do file */

loop:	if (line > IniGetNumLines (WORKTODO_FILE)) return (FALSE);
	IniGetLineAsString (WORKTODO_FILE, line, keyword, sizeof (keyword), value, sizeof (value));

/* Parse the keyword */

	if (stricmp (keyword, "Factor") == 0)
		w->work_type = WORK_FACTOR;
	else if (stricmp (keyword, "Test") == 0)
		w->work_type = WORK_TEST;
	else if (stricmp (keyword, "AdvancedTest") == 0)
		w->work_type = WORK_ADVANCEDTEST;
	else if (stricmp (keyword, "DoubleCheck") == 0)
		w->work_type = WORK_DBLCHK;
	else if (stricmp (keyword, "ECM") == 0) {
		int	i;
		char	*q;
		w->work_type = WORK_ECM;
		sscanf (value, "%ld,%ld,%ld,%ld,%ld",
			 &w->p, &w->B1, &w->B2, &w->curves_to_do,
			 &w->curves_completed);
		w->curve = 0;
		for (i = 1, q = value; i <= 5; i++, q++)
			if ((q = strchr (q, ',')) == NULL) return (TRUE);
		w->curve = atof (q);
		return (TRUE);
	} else {
		char	buf[130];
		sprintf (buf, "Illegal line in INI file: %s=%s\n", keyword, value);
		OutputSomewhere (buf);
		IniDeleteLine (WORKTODO_FILE, line);
		goto loop;
	}

/* If work type involves factoring, get the number of bits factored */

	comma = strchr (value, ',');
	if (comma == NULL) w->bits = 0;
	else {
		*comma++ = 0;
		w->bits = (unsigned int) atol (comma);
	}

/* Get the exponent being tested */

	w->p = atol (value);
	return (TRUE);
}

/* Add a line of work to the work-to-do INI file.  */
/* This assumes the caller will open and close the ini file. */

void addWorkToDoLine (
	struct work_unit *w)	/* Type of work */
{
	char	buf[80];
	unsigned int i;

/* Add new ECM work after all other ECM work, but before */
/* any other type of work */

	if (w->work_type == WORK_ECM) {
		for (i = 1; ; i++) {
			struct work_unit temp;
			if (! parseWorkToDoLine (i, &temp)) break;
			if (temp.work_type != WORK_ECM) break;
		}

/* Format and insert line into ini file */

		sprintf (buf, "%lu,%lu,%lu,%lu,%lu,%.0f",
			 w->p, w->B1, w->B2, w->curves_to_do,
			 w->curves_completed, w->curve);
		IniAppendLineAsString (WORKTODO_FILE, "ECM", buf);
		return;
	}
	
/* See if this exponent is already in the worktodo file. */

	for (i = 1; ; i++) {
		struct work_unit temp;

/* Read the line of the work file */

		if (! parseWorkToDoLine (i, &temp)) break;

/* Skip the line if p is not in the range */

		if (w->p != temp.p) continue;

/* Ignore this add request if it is a subset of the */
/* entry that is already in the work-to-do file */

		if (w->work_type == temp.work_type) return;
		if (w->work_type == WORK_FACTOR) return;

/* Delete the line */

		IniDeleteLine (WORKTODO_FILE, i);
		break;
	}

/* Format the number of bits factored */

	sprintf (buf, "%ld,%d", w->p, w->bits);

/* Append a line */

	if (w->work_type == WORK_FACTOR)
		IniAppendLineAsString (WORKTODO_FILE, "Factor", buf);
	if (w->work_type == WORK_TEST)
		IniAppendLineAsString (WORKTODO_FILE, "Test", buf);
	if (w->work_type == WORK_DBLCHK)
		IniAppendLineAsString (WORKTODO_FILE, "DoubleCheck", buf);
}

/* Open the results file and see if a given exponent has */
/* already been worked on. */ 

void checkResultsFile (
	unsigned long p,
	int	*tested,
	int	*factored)
{
	int	fd;
	char	buf[80];
	long	offset = 0;

/* Assume prime has not been lucas tested or factored. */

	*tested = FALSE;
	*factored = FALSE;

/* Loop until entire file has been processed */

	fd = _open (RESFILE, _O_BINARY | _O_RDONLY);
	if (fd < 0) return;
	for ( ; ; ) {
		unsigned long file_p = 0;
		int	i = 0;
		_lseek (fd, offset, SEEK_SET);
		buf[_read (fd, buf, sizeof (buf) - 1)] = 0;
		while (buf[i]) if (buf[i++] == 'M') break;
		if (buf[i] == 0) break;
		for (file_p = 0; isdigit(buf[i]); i++)
			file_p = file_p * 10 + buf[i] - '0';
		while (isspace(buf[i])) i++;
		if (file_p == p) {
			if (buf[i] == 'h') *tested = TRUE;
			if (buf[i] == 'n') *factored = TRUE;
			if (buf[i] == 'i') *tested = TRUE;
		}
		while (buf[i] && buf[i] != '\n') i++;
		offset += i;
	}
	_close (fd);
}


/* Get exponents from the database file and add them to the */
/* work-to-do file */

void getWorkFromDatabase (
	unsigned long range_start,
	unsigned long range_end,
	int	factoring,
	int	dblchk)
{
	unsigned long base = 0;
	unsigned long p;
	int	fd;
	unsigned int bits, max_exponents;

/* Place an upper limit on the number of exponents to be checked out */

	max_exponents = (unsigned int) IniGetInt (INI_FILE, "MaxExponents", 10);
	if (factoring) max_exponents *= 4;

/* Open the database file, if this fails raise error and return */

	fd = _open (DBFILE, _O_BINARY | _O_RDONLY);
	if (fd < 0) {
		OutputSomewhere (FILEERR);
		OutputSomewhere (FILEERR2);
		return;
	}

/* Open the work-to-do file */

	IniFileOpen (WORKTODO_FILE, 0);

/* Loop reading database file */

	for ( ; ; ) {
		struct work_unit w;

/* Read next entry, break if EOF */

		if (_read (fd, &p, 4) != 4) break;

/* Check for special "set base" record */

		if ((p & 0xFF000000) == 0xFF000000) {
			base = p >> 8;
			continue;
		}

/* See if this exponent meets our criteria */

		if (p == 0) continue;
		if ((p & 0x80000000) && !dblchk) continue;
		if (!(p & 0x80000000) && dblchk) continue;
		bits = (int) ((p >> 24) & 0x3F);
		p = base + (p & 0xFFFFFF);
		if (p < range_start) continue;
		if (p > range_end) continue;

/* See if prime is already in the results file.  This can happen when a user */
/* downloads a new database in the middle of testing a range. */

		if (!factoring) {
			int	tested, factored;
			checkResultsFile (p, &tested, &factored);
			if (tested) continue;
			if (factored) bits = 63;
		}

/* If we're only finding factors, see if this has been factored enough */

		if (factoring && bits >= factorLimit (p, 1))
			continue;

/* Only allow a small number of entries in the work-to-do file */
/* That is don't let someone accidentally check out a large number */
/* of exponents. */

		if (IniGetNumLines (WORKTODO_FILE) >= max_exponents) {
			OutputSomewhere ("Range truncated.  Maximum number of exponents exceeded.\n");
			break;
		}

/* Add a line to the work-to-do file */

		w.work_type = factoring ? WORK_FACTOR :
				 dblchk ? WORK_DBLCHK : WORK_TEST;
		w.p = p;
		w.bits = bits;
		addWorkToDoLine (&w);
	}

/* Close files and return */

	_close (fd);
	IniFileClose (WORKTODO_FILE);
}

/* Clear exponent(s) from the work-to-do list and optional database file */

void clearExponents (
	unsigned long start,
	unsigned long end,
	int	type)		/* zero OR how_far_factored */
{
	unsigned long base = 0;
	unsigned int i;
	int	fd;

/* Clear the exponent from the work-to-do-file */

	IniFileOpen (WORKTODO_FILE, 0);
	for (i = 1; ; i++) {
		struct work_unit w;

/* Read the line of the work file */

		if (! parseWorkToDoLine (i, &w)) break;

/* Skip the line if p is not in the range */

		if (w.p < start || w.p > end) continue;

/* Delete the line */

		if (type == 0 || w.work_type == WORK_FACTOR) {
			IniDeleteLine (WORKTODO_FILE, i);
			i--;
		}

/* Otherwise, replace the line with the new how-far-factored info */

		else {
			char	buf[20];
			sprintf (buf, "%ld,%d", w.p, type);
			IniReplaceLineAsString (
				WORKTODO_FILE, i,
				(w.work_type == WORK_TEST) ? "Test" : "DoubleCheck",
				buf);
		}
	}
	IniFileClose (WORKTODO_FILE);

/* Clear a prime from the local database */

	fd = _open (DBFILE, _O_BINARY | _O_RDWR);
	if (fd >= 0) {
		for ( ; ; ) {
			unsigned long temp;
			if (_read (fd, &temp, 4) != 4) break;
			if ((temp & 0xFF000000) == 0xFF000000) {
				base = temp >> 8;
				continue;
			}
			if (base + (temp & 0xFFFFFF) < start) continue;
			if (base + (temp & 0xFFFFFF) > end) break;
			_lseek (fd, -4L, SEEK_CUR);
			temp = type ? ((type << 24) + (temp & 0x80FFFFFF)) : 0;
			_write (fd, &temp, 4);
		}
		_close (fd);
	}
}

/* See if a new database has been downloaded with more */
/* recent factoring information */

unsigned int checkDatabaseForNewInfo (
	unsigned long p,
	unsigned int bits)
{
	int	fd;
	unsigned long base = 0;

	fd = _open (DBFILE, _O_BINARY | _O_RDONLY);
	if (fd < 0) return (bits);
	for ( ; ; ) {
		unsigned long temp;
		if (_read (fd, &temp, 4) != 4) break;
		if ((temp & 0xFF000000) == 0xFF000000) {
			base = temp >> 8;
			continue;
		}
		if (base + (temp & 0xFFFFFF) < p) continue;
		if (base + (temp & 0xFFFFFF) > p) break;
		temp = temp >> 24;
		temp &= 0x7F;
		if ((unsigned int) temp > bits) {
			bits = (unsigned int) temp;
			clearExponents (p, p, bits);
		}
		break;
	}
	_close (fd);
	return (bits);
}


/* Return the number of seconds until vacation is over */

unsigned long secondsUntilVacationEnds ()
{
	time_t	current_time;

	if (VACATION_END == 0) return (0);
	time (&current_time);
	if (current_time >= VACATION_END) {
		VACATION_END = 0;
		IniWriteInt (LOCALINI_FILE, "VacationEnd", 0);
		return (0);
	}
	return (VACATION_END - current_time);
}

/* Return a double estimating current exponent's percentage complete. */

double pct_complete (
	int	work_type,
	unsigned long p,
	unsigned long *iteration)
{
	unsigned long counter;
	short	type;
	int	fd;
	char	filename[20];

	if (p == EXP_BEING_WORKED_ON) {
		if (EXP_PERCENT_COMPLETE == 999.0) {
			*iteration = p;
			return (0.999);
		}
		if (work_type == WORK_FACTOR && EXP_BEING_FACTORED) {
			*iteration = (unsigned long) (EXP_PERCENT_COMPLETE * 16 + 1);
			return (EXP_PERCENT_COMPLETE);
		}
		if (work_type != WORK_FACTOR && !EXP_BEING_FACTORED) {
			*iteration = (unsigned long) (EXP_PERCENT_COMPLETE * p);
			return (EXP_PERCENT_COMPLETE);
		}
	}

	tempFileName (filename, p);
	if (fileExists (filename) &&
	    readFileHeader (filename, &fd, &type, &counter)) {
		close (fd);
		if (type >= 4) {
			*iteration = counter;
			return ((double) counter / (double) p);
		}
		if (type == 2) {
			*iteration = 1;
			return (0.01);
		}
	}

	*iteration = 0;
	return (0.0);
}

/* Make a guess as to how long a chore should take. */

double raw_work_estimate (
	struct work_unit *w)
{
	double	timing;
	double	est;

/* Not very accurate estimates for ECM! */

	if (w->work_type == WORK_ECM) est = 86400.0;

/* If factoring, guess how long that will take */
/* Timings are based on how long it takes my P-90 */
/* to process a range 9888*2^32 factors when p is 2.7 million */
/* Compute 2^limit * 16 passes / (9888 * 2^32) * 2,700,000 / p */
/* Which simplifies to: 2^(limit-32) * 4368.9 / p which equals */
/* 2^(limit-40) * 1118438 / p */

	if (w->work_type == WORK_FACTOR) {
		unsigned int limit;
		limit = factorLimit (w->p, 1);
		timing = (limit > 62) ? 0.217 : 0.189;
		est = timing * 1118438 * (1L << (limit - 40)) / w->p;
		est = est * 90.0 / CPU_SPEED;
		if (CPU_TYPE <= 4) est *= REL_486_SPEED;
		if (CPU_TYPE == 6) est *= (limit > 62) ? 1.17 : 0.75;
		if (CPU_TYPE == 7) est *= REL_K6_SPEED;
		if (w->bits >= limit) est = 0.0;
		else if (w->bits == limit-1) est *= 0.5;
		else if (w->bits == limit-2) est *= 0.75;
		else if (w->bits == limit-3) est *= 0.875;
	}

/* If testing add in the Lucas-Lehmer testing time */

	if (w->work_type == WORK_TEST ||
	    w->work_type == WORK_ADVANCEDTEST ||
	    w->work_type == WORK_DBLCHK) {
		est = w->p * map_fftlen_to_timing (pick_fft_length (w->p));
	}

/* Return the total estimated time in seconds */

	return (est);
}


/* Make a guess as to how long a chore should take. */

double work_estimate (
	struct work_unit *w)
{
	return (raw_work_estimate (w) *
		24.0 / CPU_HOURS * 1000.0 / ROLLING_AVERAGE);
}


/* Determine how much we should factor (in bits) */
/* Don't let feeble 486 any Cyrix machines factor higher than 2^62 */
/* as this would execute some really slow floating point code */

unsigned int factorLimit (
	unsigned long p,
	int	factoring)
{
	unsigned int test, override;

	if (p > FAC64) test = 64;	/* Test all 64 bit factors */
	else if (p > FAC63) test = 63;	/* Test all 63 bit factors */
	else if (p > FAC62) test = 62;	/* Test all 62 bit factors */
	else if (p > FAC61) test = 61;	/* Test all 61 bit factors */
	else if (p > FAC60) test = 60;	/* Test all 60 bit factors */
	else if (p > FAC59) test = 59;	/* Test all 59 bit factors */
	else if (p > FAC58) test = 58;	/* Test all 58 bit factors */
	else if (p > FAC57) test = 57;	/* Test all 57 bit factors */
	else test = 40;			/* Test all 40 bit factors */
	if (factoring) {
/*		if (CPU_TYPE < 5 && test > 62) test = 62; */
		override = (unsigned int)
			IniGetInt (INI_FILE, "FactorOverride", 99);
		if (test > override) test = override;
	}
	return (test);
}

/* Given an exponent, pick the right fft length */

unsigned long pick_fft_length (
	unsigned long p) 
{
	if (p < LIMIT80) return (81920);
	if (p < LIMIT96) return (98304);
	if (p < LIMIT112) return (114688);
	if (p < LIMIT128) return (131072);
	if (p < LIMIT160) return (163840);
	if (p < LIMIT192) return (196608);
	if (p < LIMIT224) return (229376);
	if (p < LIMIT256) return (262144);
	if (p < LIMIT320) return (327680);
	if (p < LIMIT384) return (393216);
	if (p < LIMIT448) return (458752);
	if (p < LIMIT512) return (524288);
	if (p < LIMIT640) return (655360);
	if (p < LIMIT768) return (786432);
	if (p < LIMIT896) return (917504);
	return (1048576);
}


/* Make a guess as to how long each iteration should take. */

double map_fftlen_to_timing (
	unsigned long fftlen)
{
	double	timing;

/* Use my P-90 timings as a guide */

	if (fftlen == 81920) timing = 0.227;
	if (fftlen == 98304) timing = 0.272;
	if (fftlen == 114688) timing = 0.332;
	if (fftlen == 131072) timing = 0.371;
	if (fftlen == 163840) timing = 0.481;
	if (fftlen == 196608) timing = 0.589;
	if (fftlen == 229376) timing = 0.699;
	if (fftlen == 262144) timing = 0.778;
	if (fftlen == 327680) timing = 1.034;
	if (fftlen == 393216) timing = 1.244;
	if (fftlen == 458752) timing = 1.498;
	if (fftlen == 524288) timing = 1.680;
	if (fftlen == 655360) timing = 2.470;
	if (fftlen == 786432) timing = 3.074;
	if (fftlen == 917504) timing = 3.796;
	if (fftlen == 1048576) timing = 4.202;

/* Since the program is about 10% memory bound, the program will not */
/* speed up linearly with increase in chip speed. */

	timing = 0.10 * timing + 0.90 * timing * 90.0 / CPU_SPEED;
	if (CPU_TYPE <= 4) timing = timing * REL_486_SPEED;
	if (CPU_TYPE == 6) timing = timing * REL_PRO_SPEED;
	if (CPU_TYPE == 7) timing = timing * REL_K6_SPEED;
	return (timing);
}

/* Generate temporary file name */

void tempFileName (
	char	*buf,
	unsigned long p)
{
	sprintf (buf, "p%07li", p % 10000000);
	if (p >= 10000000) buf[1] = (char) ('A' + (p / 1000000) - 10);
}

/* See if the given file exists */

int fileExists (
	char	*filename)
{
	int	fd;
	fd = _open (filename, _O_RDONLY | _O_BINARY);
	if (fd < 0) return (0);
	_close (fd);
	return (1);
}

/* Read the header of an intermediate Lucas-Lehmer results file */

int readFileHeader (
	char	*filename,
	int	*fd,
	short	*type,
	unsigned long *counter)
{
	*fd = _open (filename, _O_BINARY | _O_RDONLY);
	if (*fd <= 0) return (FALSE);
	if (_read (*fd, type, sizeof (short)) != sizeof (short)) goto readerr;
	if (_read (*fd, counter, sizeof (long)) != sizeof (long)) goto readerr;
	return (TRUE);
readerr:_close (*fd);
	return (FALSE);
}

/* Open the results file and write a line to the end of it. */

int writeResults (
	char	*msg)
{
static	time_t	last_time = 0;
	time_t	this_time;
	int	fd;

/* Open file, position to end */

	fd = _open (RESFILE, _O_TEXT | _O_RDWR | _O_CREAT | _O_APPEND, 0666);
	if (fd < 0) {
		LogMsg ("Error opening the results file.\n");
		return (FALSE);
	}

/* If it has been at least 5 minutes since the last time stamp */
/* was output, then output a new timestamp */

	time (&this_time);
	if (this_time - last_time > 300) {
		char	buf[32];
		last_time = this_time;
		buf[0] = '[';
		strcpy (buf+1, ctime (&this_time));
		buf[25] = ']';
		buf[26] = '\n';
		_write (fd, buf, 27);
	}

/* Output a USERID prefix for result messages */

	if (msg[0] == 'M' && isdigit (msg[1]) && USERID[0]) {
		if (_write (fd, "UID: ", 5) < 0) return (FALSE);
		if (_write (fd, USERID, strlen (USERID)) < 0) return (FALSE);
		if (COMPID[0]) {
			if (_write (fd, "/", 1) < 0) return (FALSE);
			if (_write (fd, COMPID, strlen (COMPID)) < 0)
				return (FALSE);
		}
		if (_write (fd, ", ", 2) < 0) return (FALSE);
	}

/* Output the message */

	if (_write (fd, msg, strlen (msg)) < 0) return (FALSE);
	_close (fd);
	return (TRUE);
}


/* Send any queued up messages to the server.  See if we have enough */
/* work queued up.  If we have too much work, give some back */

int communicateWithServer ()
{
static	int	msgs_to_send = 0;/* 0 = no messages, 0x1 = messages */
				/* 0x2 = critical messages */
	char	header_byte;	/* Cached flag byte from spool file */
				/* 0 = no spool file */
				/* 0x1 = informational msgs in file */
				/* 0x2 = important msgs in file */
				/* 0x4 = user info has changed */
				/* 0x8 = completion dates need updating */
				/* 0x10 = return all work */
				/* 0x20 = computer info has changed */
	int	fd;		/* Spool file handle */
	time_t	this_time;
	unsigned int i;
	int	have_enough_work, retries;
	double	est, work_to_get;
	unsigned long vacation_time;
	int	rc;
	int	talked_to_server = 0;
	int	manual_comm = 0;

/* Use this oppurtunity to perform other miscellaneous tasks that may */
/* be required by this particular OS port */

	doMiscTasks ();

/* If were operating manually, return */

	if (!USE_PRIMENET) return (TRUE);

/* If all communication with the server is manually initiated, */
/* then handle that case here */

	if (MANUAL_COMM & 0x2) {
		MANUAL_COMM &= 0xFD;
		manual_comm = 1;
		next_comm_time = 0;
		talked_to_server = 1;
	} else if (MANUAL_COMM & 0x1) {
		return (TRUE);
	}

/* Open and read the spool file to set msgs_to_send correctly */

	if (SPOOL_FILE_CHANGED) {
		SPOOL_FILE_CHANGED = 0;
		fd = _open (SPOOL_FILE, _O_RDONLY | _O_BINARY);
		if (fd >= 0) {

/* Read the first byte.  This indicates if there are */
/* "important" messages queued up.  Important messages */
/* will cause us to try and contact the server every */
/* few minutes until we get through. */

			_read (fd, &header_byte, 1);
			if (header_byte > 1) msgs_to_send |= 0x1;
			if (header_byte & 0x10) msgs_to_send |= 0x2;
			_close (fd);
		}
	}

/* If no important messages queued up AND we're not */
/* checking to see if we have enough work to do, then return */

	if (!msgs_to_send && !CHECK_WORK_QUEUE) return (TRUE);

/* Make sure MODEM_RETRY_TIME or NETWORK_RETRY_TIME minutes have */
/* passed since our last attempt to communicate with the server. */

loop:	time (&this_time);
	if (this_time < next_comm_time) goto exit_or_loop_nomsg;

/* Make sure we don't pummel the server with data.  Suppose user uses */
/* Advanced/Factor to find tiny factors again.  At least, make sure */
/* he only sends the data once every 5 minutes. */

	next_comm_time = this_time + 300;

/* If we have messages to send, see if the server is available */
/* Use Ping so that we don't write to the log file every few minutes */
/* when the server is unavailable */

	if (msgs_to_send) {
		struct primenetPingServerInfo pkt;
		memset (&pkt, 0, sizeof (pkt));
		rc = sendMessage (PRIMENET_PING_SERVER_INFO, &pkt);
		if (rc) goto exit_or_loop;
		talked_to_server = 1;
	}

/* Send all the queued up messages */

	header_byte = 0;
	fd = _open (SPOOL_FILE, _O_RDWR | _O_BINARY);
	if (fd >= 0) {

/* Read the first byte.  This indicates if there are */
/* "important" messages queued up.  Important messages */
/* will cause us to try and contact the server every */
/* few minutes until we get through. */

		_read (fd, &header_byte, 1);

/* Send any changes to the user info first */

		if (header_byte & 0x4) {
			struct primenetUserInfo pkt;

			memset (&pkt, 0, sizeof (pkt));
			strcpy (pkt.userName, USER_NAME);
			strcpy (pkt.userEmailAddr, USER_ADDR);
			strcpy (pkt.oldUserID, OLD_USERID);
			strcpy (pkt.oldUserPW, OLD_USER_PWD);
			if (SEND_EMAIL)
				pkt.bUserOptions |= PRIMENET_USER_OPTION_SENDEMAIL;
			if (IniGetInt (INI_FILE, "LockUserInfo", 0))
				pkt.bUserOptions |= PRIMENET_USER_OPTION_TEAMACCT;
			rc = sendMessage (PRIMENET_MAINTAIN_USER_INFO, &pkt);
			if (rc) {
				_close (fd);
				goto exit_or_loop;
			}
			talked_to_server = 1;

/* Delete old User id info */

			OLD_USERID[0] = 0;
			OLD_USER_PWD[0] = 0;
			IniWriteString (INI_FILE, "OldUserID", OLD_USERID);
			IniWriteString (INI_FILE, "OldUserPWD", OLD_USER_PWD);

/* If a user ID has just been assigned to us, then write it the INI file */

			if (strcmp (USERID, pkt.userID)) {
				strcpy (USERID, pkt.userID);
				IniWriteString (INI_FILE, "UserID", USERID);
			}
			if (strcmp (USER_PWD, pkt.userPW)) {
				strcpy (USER_PWD, pkt.userPW);
				IniWriteString (INI_FILE, "UserPWD", USER_PWD);
			}

/* Flag the message as successfully sent */

			header_byte &= 0xFB;
			_lseek (fd, 0, SEEK_SET);
			_write (fd, &header_byte, 1);
		}

/* Send any changes to the computer info next */

		if (header_byte & 0x20) {
			struct primenetComputerInfo pkt;

			memset (&pkt, 0, sizeof (pkt));
			pkt.cpu_type = CPU_TYPE;
			pkt.speed = (short) CPU_SPEED;
			pkt.hours_per_day = CPU_HOURS;
			rc = sendMessage (PRIMENET_SET_COMPUTER_INFO, &pkt);
			if (rc) {
				_close (fd);
				goto exit_or_loop;
			}
			talked_to_server = 1;

/* Flag the message as successfully sent */

			header_byte &= 0xDF;
			_lseek (fd, 0, SEEK_SET);
			_write (fd, &header_byte, 1);
		}

/* Send the messages */

		for ( ; ; ) {
			short	msgType;
			union {
				struct primenetResultMessage msg;
				struct primenetAssignmentResult res;
			} msg;
			long	old_offset, new_offset;

			readMessage (fd, &old_offset, &msgType, &msg);
			if (msgType == 0) break;
			rc = sendMessage (msgType, &msg);
			if (rc) {
				_close (fd);
				goto exit_or_loop;
			}
			talked_to_server = 1;

/* Flag the message as successfully sent */

			new_offset = _lseek (fd, 0, SEEK_CUR);
			_lseek (fd, old_offset, SEEK_SET);
			msgType = -1;
			_write (fd, &msgType, sizeof (short));
			_lseek (fd, new_offset, SEEK_SET);

/* Stop sending to server if so requested */

			if (escapeCheck ()) {
				_close (fd);
				return (FALSE);
			}
		}
		header_byte &= 0xFC;

/* Close the spool file */

		_close (fd);
	}

/* At this point, we can usually delete the spool file */

	if (header_byte == 0) _unlink (SPOOL_FILE);

/* Clear flag indicating work queue needs checking */
/* We will set msgs_to_send if we find we need to return */
/* or request exponents.  NOTE: We must be extra careful here */
/* to avoid opening the work-to-do file every few minutes when */
/* the server is unavailable. */

	CHECK_WORK_QUEUE = 0;

/* Get work to do until we've accumulated enough to keep */
/* us busy for a while (DAYS_OF_WORK or the length of a vacation). */
/* If we have too much work to do, lets give some back. */

	have_enough_work = 0;
	est = 0.0;
	vacation_time = secondsUntilVacationEnds ();
	work_to_get = DAYS_OF_WORK * 86400.0;
	if (ON_DURING_VACATION && work_to_get < vacation_time)
		work_to_get = vacation_time;
	IniFileOpen (WORKTODO_FILE, 0);
	for (i = 1; ; i++) {
		struct work_unit w;
		unsigned long iteration;
		double	work, pct;

/* Read the line of the work file */

		if (! parseWorkToDoLine (i, &w)) break;

/* Adjust our time estimate */

		work = work_estimate (&w);
		pct = pct_complete (w.work_type, w.p, &iteration);
		work *= (1.0 - pct);
		est += work;

/* If we have too much work queued, give it back */

		if (header_byte & 0x10 ||
		    (have_enough_work &&
		     est >= work_to_get + 30.0 * 86400.0 &&
		     pct == 0.0)) {
			struct primenetAssignmentResult pkt;
			memset (&pkt, 0, sizeof (pkt));
			pkt.exponent = w.p;
			pkt.resultType = PRIMENET_RESULT_UNRESERVE;
			msgs_to_send |= 0x1;
			rc = sendMessage (PRIMENET_ASSIGNMENT_RESULT, &pkt);
			if (rc) {
				IniFileClose (WORKTODO_FILE);
				goto exit_or_loop;
			}
			talked_to_server = 1;
			IniDeleteLine (WORKTODO_FILE, i);
			i--;
		}

/* Acknowledge the exponent by sending a projected completion date */
/* If an unexpected error occurs, ignore it.  We can get "not-assigned" */
/* errors when testing manually reserved exponents and we've just */
/* enabled primenet. */

		else if (header_byte & 0x8) {
			struct primenetCompletionDate pkt2;
			double	comp_date;
			memset (&pkt2, 0, sizeof (pkt2));
			pkt2.exponent = w.p;
			if (!ON_DURING_VACATION)
				comp_date = est + vacation_time;
			else if (est < vacation_time)
				comp_date = vacation_time;
			else
				comp_date = est;
			pkt2.days = (unsigned long) (comp_date / 86400.0) + 1;
			pkt2.requestType =
				(w.work_type == WORK_FACTOR) ? PRIMENET_ASSIGN_FACTOR :
				(w.work_type == WORK_TEST) ? PRIMENET_ASSIGN_TEST :
					PRIMENET_ASSIGN_DBLCHK;
			pkt2.programType = PRIMENET_PROGRAM_WOLTMAN;
			pkt2.iteration = iteration;
			if (ON_DURING_VACATION && vacation_time > (unsigned long) DAYS_BETWEEN_CHECKINS * 86400)
				pkt2.nextMsg = vacation_time / 86400;
			else
				pkt2.nextMsg = DAYS_BETWEEN_CHECKINS;
			rc = sendMessage (PRIMENET_COMPLETION_DATE, &pkt2);
			if (rc == PRIMENET_ERROR_CONNECT_FAILED ||
			    rc == PRIMENET_ERROR_ACCESS_DENIED) {
				IniFileClose (WORKTODO_FILE);
				goto exit_or_loop;
			}
			if (i > 1 &&
			    (rc == PRIMENET_ERROR_ALREADY_TESTED ||
			     rc == PRIMENET_ERROR_NOT_PERMITTED)) {
				est -= work;
				IniDeleteLine (WORKTODO_FILE, i);
				i--;
			}
			talked_to_server = 1;
		}

/* Set flag if we have enough work */

		if (est >= work_to_get)
			have_enough_work = 1;
	}

/* Delete the spool file. */

	if (header_byte) _unlink (SPOOL_FILE);

/* If we just returned all work to the server, enter manual mode */
/* and exit - the user is quitting GIMPS */

	if (header_byte & 0x10) {
		USE_PRIMENET = 0;
		IniWriteInt (INI_FILE, "UsePrimenet", 0);
		GIMPS_QUIT = 1;
		OutputSomewhere ("Successfully quit GIMPS.\n");
		goto exit;
	}

/* After sending new completion dates, remember the current rolling */
/* average so that if it changes by more than 10% we can send new */
/* completion dates.  Also remember the current time, so that we can */
/* send new completion dates in a month. */

	if (header_byte & 0x8) {
		time_t	current_time;
		IniWriteInt (LOCALINI_FILE, "OldRollingAverage", ROLLING_AVERAGE);
		time (&current_time);
		IniWriteInt (LOCALINI_FILE, "LastEndDatesSent", current_time);
	}

/* If we don't have enough work to do, */
/* get more work from the server. */

	retries = 0;
	while (!have_enough_work &&
	       IniGetNumLines (WORKTODO_FILE) < (unsigned int)
			IniGetInt (INI_FILE, "MaxExponents", 20)) {
		struct primenetGetAssignment pkt1;
		struct primenetCompletionDate pkt2;
		struct work_unit w;
		double	new_est, comp_date;

/* Check for ESC */

		if (escapeCheck ()) {
			IniFileClose (WORKTODO_FILE);
			return (FALSE);
		}

/* Get an exponent to work on */

		memset (&pkt1, 0, sizeof (pkt1));
		if (WORK_PREFERENCE)
			pkt1.requestType = WORK_PREFERENCE;
		else
			pkt1.requestType = default_work_type ();
		pkt1.programType = PRIMENET_PROGRAM_WOLTMAN;
		pkt1.how_far_factored = (CPU_TYPE < 5) ? 486.0 : 64.0;
		msgs_to_send |= 0x1;
		rc = sendMessage (PRIMENET_GET_ASSIGNMENT, &pkt1);
		if (rc) {
			IniFileClose (WORKTODO_FILE);
			goto exit_or_loop;
		}
		talked_to_server = 1;

/* For some strange reason the server has been known to send us */
/* bogus exponents (zero) to test */

		if (pkt1.exponent < 1000000) {
			char	buf[80];
			sprintf (buf, "Server sent bad exponent: %ld.\n",
				 pkt1.exponent);
			LogMsg (buf);
			if (++retries < 3) continue;
			IniFileClose (WORKTODO_FILE);
			goto exit_or_loop;
		}

/* Add the exponent to our time estimate */

		w.work_type = (pkt1.requestType == PRIMENET_ASSIGN_FACTOR) ?
						WORK_FACTOR :
			      (pkt1.requestType == PRIMENET_ASSIGN_TEST) ?
						WORK_TEST : WORK_DBLCHK;
		w.p = pkt1.exponent;
		w.bits = (unsigned int) pkt1.how_far_factored;
		new_est = est + work_estimate (&w);

/* Acknowledge the exponent by sending a projected completion date */

		memset (&pkt2, 0, sizeof (pkt2));
		pkt2.exponent = pkt1.exponent;
		if (!ON_DURING_VACATION)
			comp_date = new_est + vacation_time;
		else if (new_est < vacation_time)
			comp_date = vacation_time;
		else
			comp_date = new_est;
		pkt2.days = (unsigned long) (comp_date / 86400.0) + 1;
		pkt2.requestType =
			(w.work_type == WORK_FACTOR) ? PRIMENET_ASSIGN_FACTOR :
			(w.work_type == WORK_TEST) ? PRIMENET_ASSIGN_TEST :
				PRIMENET_ASSIGN_DBLCHK;
		pkt2.programType = PRIMENET_PROGRAM_WOLTMAN;
		pkt2.iteration = 0;
		if (ON_DURING_VACATION && vacation_time > (unsigned long) DAYS_BETWEEN_CHECKINS * 86400)
			pkt2.nextMsg = vacation_time / 86400;
		else
			pkt2.nextMsg = DAYS_BETWEEN_CHECKINS;
		rc = sendMessage (PRIMENET_COMPLETION_DATE, &pkt2);
		if (rc) {
			if (++retries < 3) continue;
			IniFileClose (WORKTODO_FILE);
			goto exit_or_loop;
		}
		talked_to_server = 1;

/* Once the server accepts our acknowledgement, write the exponent */
/* to our worktodo file */

		addWorkToDoLine (&w);

/* Remember the current rolling average so that if it changes */
/* by more than 10% we can send new completion dates */

		IniWriteInt (LOCALINI_FILE, "OldRollingAverage", ROLLING_AVERAGE);

/* Set flag if we have enough work */

		est = new_est;
		if (est >= work_to_get)
			have_enough_work = 1;
	}

/* Close the work-to-do file */

exit:	IniFileClose (WORKTODO_FILE);

/* Tell user we're done communicating */

	if (talked_to_server)
		OutputStr ("Done communicating with server.\n");

/* Clear flag indicating msgs need to be sent to primenet server */

	msgs_to_send = 0;
	return (TRUE);

/* We could not contact server, either exit or loop */
/* We loop when we are quitting GIMPS */

exit_or_loop:
	if (!manual_comm) {
		char	buf[80];
		unsigned int retry_time;
		retry_time = (rc == PRIMENET_ERROR_MODEM_OFF) ?
			MODEM_RETRY_TIME : NETWORK_RETRY_TIME;
		sprintf (buf, "Will try contacting server again in %d minutes.\n",
			 retry_time);
		OutputStr (buf);
		next_comm_time = this_time + 60 * retry_time;
	}
exit_or_loop_nomsg:
	if (msgs_to_send & 0x2) {
		Sleep (5000);
		if (escapeCheck ()) return (FALSE);
		next_comm_time = 0;
		goto loop;
	}
	return (TRUE);
}
