/* Constants */

#define VERSION		"16.5"

#define MIN_PRIME	5L		/* Smallest testable prime */
#define MAX_PRIME	LIMIT1024	/* Largest testable prime */
#define MAX_FACTOR	LIMIT1024	/* Largest factorable number */

/* FFT Limits based on 1000 iterations with an error no greater than 0.30. */

#define LIMIT1024	20500000L	/* Crossover points */
#define LIMIT896	17945000L
#define LIMIT768	15400000L
#define LIMIT640	12895000L
#define LIMIT512	10380000L
#define LIMIT448	9090000L
#define LIMIT384	7820000L
#define LIMIT320	6540000L
#define LIMIT256	5260000L
#define LIMIT224	4605000L
#define LIMIT192	3960000L
#define LIMIT160	3310000L
#define LIMIT128	2655000L
#define LIMIT112	2330000L
#define LIMIT96		2000000L
#define LIMIT80		1675000L

/* Factoring limits based on complex formulas given the speed of the */
/* factoring code vs. the speed of the Lucas-Lehmer code */

#define FAC64	9150000L		/* How far to factor */
#define FAC63	7270000L
#define FAC62	5160000L
#define FAC61	LIMIT192
#define FAC60	2950000L
#define FAC59	LIMIT128
#define FAC58	2135000L
#define FAC57	LIMIT80

/* Speed of a 486 and Pentium Pro compared to a Pentium of same clock speed */

#define REL_486_SPEED	7.0	/* 486 is 7 times slower than Pentium */
#define REL_K6_SPEED	2.5	/* K6 is 2.5 times slower than Pentium */
#define REL_PRO_SPEED	0.87	/* Pro is little faster than Pentium */


/* Global variables */

extern char INI_FILE[20];		/* Name of the prime INI file */
extern char LOCALINI_FILE[20];		/* Name of the local INI file */
extern char WORKTODO_FILE[20];		/* Name of the work-to-do INI file */
extern char RESFILE[20];		/* Name of the results file */
extern char SPOOL_FILE[20];		/* Name of the spool file */
extern char LOGFILE[20];		/* Name of the server log file */

extern char USERID[20];			/* User's ID */
extern char USER_PWD[20];		/* User's Password */
extern char OLD_USERID[20];		/* Previous User's ID */
extern char OLD_USER_PWD[20];		/* Previous User's Password */
extern char COMPID[20];			/* Computer ID */
extern char USER_NAME[80];		/* User's real name */
extern char USER_ADDR[80];		/* User's email address */
extern int SEND_EMAIL;			/* Send email about expirations */
extern int USE_PRIMENET;		/* TRUE if we're using PrimeNet */
extern int USE_HTTP;			/* Use http rather than rpc DLL */
extern int DIAL_UP;			/* TRUE if we're dialing into */
					/* PrimeNet server */
extern short WORK_PREFERENCE;		/* Type of work (factoring, testing, */
					/* etc.) to get from the server. */
extern unsigned int DAYS_OF_WORK;	/* How much work to retrieve from */
					/* the primenet server */
extern time_t VACATION_END;		/* Date vacation is expected to end */
extern int ON_DURING_VACATION;		/* TRUE if computer is on while */
					/* vacationing */
extern int ADVANCED_ENABLED;		/* 1 if advanced menu is enabled */
EXTERNC int volatile ERRCHK;		/* 1 to turn on error checking */
extern unsigned int PRIORITY;		/* Desired priority level */
extern unsigned int CPU_AFFINITY;	/* NT Processor affinity */
extern int MANUAL_COMM;			/* Set on if user explicitly starts */
					/* all communication with the server */
EXTERNC unsigned int volatile CPU_TYPE;	/* 3=Cyrix, 4=486, 5=Pentium, */
					/* 6=Pro, 7=K6 */
extern unsigned long volatile CPU_SPEED;/* Speed in MHz */
extern unsigned int volatile CPU_HOURS;	/* Hours per day program will run */
extern unsigned long volatile ITER_OUTPUT;/* Iterations between outputs */
extern unsigned long volatile ITER_OUTPUT_RES;/* Iterations between results */
					/* file outputs */
extern unsigned long volatile DISK_WRITE_TIME;
					/* Number of minutes between writing */
					/* intermediate results to disk */
extern unsigned int MODEM_RETRY_TIME;	/* How often to try sending msgs */
					/* to primenet server whem modem off */
extern unsigned int NETWORK_RETRY_TIME;	/* How often to try sending msgs */
					/* to primenet server */
extern unsigned int DAYS_BETWEEN_CHECKINS; /* Days between sending updated */
					/* completion dates to the server */
extern int TWO_BACKUP_FILES;		/* TRUE for 2 backup files(qXXXXXXX) */
extern int SILENT_VICTORY;		/* Quiet find of new Mersenne prime */
extern int TRAY_ICON;			/* Display tiny tray icon */
extern int HIDE_ICON;			/* Display no icon */
extern unsigned int ROLLING_AVERAGE;	/* Ratio of this computer's speed */
					/* compared to the expected speed */
					/* for this CPU */

extern unsigned long EXP_BEING_WORKED_ON; /* Exponent being tested */
extern int EXP_BEING_FACTORED;		/* TRUE is exp is being factored */
extern double EXP_PERCENT_COMPLETE;	/* Percent complete of factoring */
					/* and/or LL test */

extern int SPOOL_FILE_CHANGED;		/* Flag indicating the data in the */
					/* spool file has been updated. */
extern int CHECK_WORK_QUEUE;		/* Flag indicating we need to check */
					/* if our work queue is full. */
extern int GIMPS_QUIT;			/* TRUE if we just successfully */
					/* quit the GIMPS project */
extern time_t next_comm_time;		/* Next time to contact server */

extern char DBFILE[];			/* Name of the database file */
extern char FILEERR[];
extern char FILEERR2[];
extern char READFILEERR[];

/* Common routines */

int isPentium ();
int isPentiumPro ();
int isPrime (unsigned long p);

void nameIniFiles (int named_ini_files);
void readIniFiles ();

void IniGetString (char *, char *, char *, unsigned int, char *);
long IniGetInt (char *, char *, long);
void IniWriteString (char *, char *, char *);
void IniWriteInt (char *, char *, long);

void IniFileOpen (char *, int);
void IniFileClose (char *);
unsigned int IniGetNumLines (char *);
void IniGetLineAsString (char *, unsigned int, char *, unsigned int,
			 char *, unsigned int);
void IniGetLineAsInt (char *, unsigned int, char *, unsigned int, long *);
void IniReplaceLineAsString (char *, unsigned int, char *, char *);
void IniReplaceLineAsInt (char *, unsigned int, char *, long);
void IniInsertLineAsString (char *, unsigned int, char *, char *);
void IniInsertLineAsInt (char *, unsigned int, char *, long);
void IniAppendLineAsString (char *, char *, char *);
void IniAppendLineAsInt (char *, char *, long);
void IniDeleteLine (char *, unsigned int);
void IniDeleteAllLines (char *);

void UpdateEndDates ();
void ConditionallyUpdateEndDates ();
void spoolMessage (short, void *);
void readMessage (int, long *, short *, void *);
int sendMessage (short, void *);
void spoolExistingResultsFile ();
void OutputBoth (char *);
void OutputSomewhere (char *);
void LogMsg (char *);

struct work_unit {
	int	work_type;	/* Type of work to do */
	unsigned long p;	/* Exponent to work on */
	unsigned int bits;	/* How far factored */
	unsigned int B1;	/* ECM - Bound #1 */
	unsigned int B2;	/* ECM - Bound #2 */
	unsigned int curves_to_do; /* ECM - curves to try */
	unsigned int curves_completed; /* ECM - curves done */
	double	curve;		/* ECM - Specific curve to test */
};
short default_work_type ();
#define WORK_FACTOR		0
#define WORK_TEST		1
#define WORK_ADVANCEDTEST	2
#define WORK_DBLCHK		3
#define WORK_ECM		4
#define WORK_OTHER		5
int parseWorkToDoLine (unsigned int, struct work_unit *);
void addWorkToDoLine (struct work_unit *);
void checkResultsFile (unsigned long, int *, int *);
void getWorkFromDatabase (unsigned long, unsigned long, int, int);
void clearExponents (unsigned long, unsigned long, int);
unsigned int checkDatabaseForNewInfo (unsigned long, unsigned int);

unsigned long secondsUntilVacationEnds ();
double pct_complete (int, unsigned long, unsigned long *);
double raw_work_estimate (struct work_unit *);
double work_estimate (struct work_unit *);
unsigned int factorLimit (unsigned long, int);
double map_fftlen_to_timing (unsigned long);
unsigned long pick_fft_length (unsigned long);

void tempFileName (char	*, unsigned long);
int fileExists (char *);
int readFileHeader (char *, int *, short *, unsigned long *);
int writeResults (char	*);

int communicateWithServer ();

/* Routines called by common routines */

int LoadPrimeNet ();
void OutputStr (char *);
void guessCpuType ();
void doMiscTasks ();
int escapeCheck ();
