/*
   in mQ_next - we can compute too many values in the last group
   is sieve tuned properly?  could be even faster.  use assembly?
	do sieve once and store in a file?
   delete multiply by 1.0 in converttobinary
   handle B1 and/or B2 above 2^32
   write and test all the fft code
   add save file support
   make all the assembly routines faster
   write an FFT28???
   Implement an FFT stage 2
*/

/**************************************************************
 *
 *	ecm.c
 *
 *	ECM factoring program
 *
 *	Original author:  Richard Crandall - www.perfsci.com
 *	Adapted to Mersenne numbers and optimized by George Woltman
 *	Further optimizations from Paul Zimmerman's GMP-ECM program
 *	Other important ideas courtesy of Peter Montgomery.
 *
 *	c. 1997 Perfectly Scientific, Inc.
 *	c. 1998 Just For Fun Software, Inc.
 *	All Rights Reserved.
 *
 *************************************************************/

#define GROUP_MODINV	32
#define LUCAS_CHAINS

/* A psuedo declaration for our big numbers.  The actual pointers to */
/* these big numbers are to the data array. */

typedef struct {
/*	char	pad[32];	   Used to track unnormalized add/sub */
				/* and original address */
	double	data[80];	/* The big number broken into chunks */
} *gwnum;

/* global variables */

unsigned long D;		/* Stage 2 size */
unsigned long *N;		/* Binary representation of */
unsigned long Nlen;		/* number to factor */
EXTERNC unsigned long FFTLEN=0;	/* The FFT size we are using */
EXTERNC void *SRCARG = NULL;	/* For assembly language arg passing */
EXTERNC void *SRC2ARG = NULL;	/* For assembly language arg passing */
EXTERNC void *DESTARG = NULL;	/* For assembly language arg passing */
EXTERNC double *NORM_ARRAY=NULL;/* Array of normalization multipliers */
EXTERNC unsigned long *U=NULL;	/* Binary representation of */
EXTERNC unsigned long Ulen=0;	/* larger GCD number */
EXTERNC unsigned long *V=NULL;	/* Binary representation of */
EXTERNC unsigned long Vlen=0;	/* smaller GCD number */
EXTERNC unsigned long EGCD_A=0;	/* Returned by assembly GCD helper */
EXTERNC unsigned long EGCD_B=0;
EXTERNC unsigned long EGCD_C=0;
EXTERNC unsigned long EGCD_D=0;
EXTERNC unsigned long EGCD_ODD=0;
EXTERNC unsigned long CARRYH=0;	/* For multi-precision asm routines */
EXTERNC unsigned long CARRYL=0;
EXTERNC unsigned long RES=0;

gwnum	x = NULL, z = NULL,
	xs = NULL, zs = NULL, xorg = NULL, zorg = NULL,
	t1 = NULL, t2 = NULL, t3 = NULL, t4 = NULL,
	gg = NULL, Ad4 = NULL;
gwnum	xt = NULL, zt = NULL, xA = NULL, zA = NULL, xB = NULL, zB = NULL,
	xC = NULL, zC = NULL;
gwnum	*nQx = NULL;		/* Array of data used in stage 2 */
gwnum	xgrp[GROUP_MODINV], zgrp[GROUP_MODINV], prevgrp[GROUP_MODINV];
gwnum	tmpgrp[GROUP_MODINV];

/* Macros to interface with assembly code */
/* The assembly routines are designed to provide a flexible way of */
/* multiplying two numbers.  If you will use a value in several multiplies */
/* you can perform the forward transform just once.  Furthermore, the */
/* multiply routines are tuned to allow one unnormalized addition prior */
/* to a multiply without introducing too much convolution error.  Thus: */
/* Legal:	gwaddquick (t1, t2); gwmul (t2, x); */
/* Legal:	gwfft (t1, t1); gwfft (t2, t2); */
/*		gwaddquick (t1, t2); gwfftmul (t2, x); */
/* Not Legal:	gwaddquick (t1, t2); gwaddquick (y, x); gwmul (t2, x); */
/* Not Legal:	gwfft (t1, t1); gwfft (t2, t2); */
/*		gwaddquick (t1, t2); gwfftfftmul (t2, t2); */

/* A brief description of each of the "gw" routines: */
/* gwsetup	Initializes the gw code for a specific Mersenne exponent */
/* gwswap	Quickly swaps two gw numbers */
/* gwcopy(s,d)	Copies gwnum s to d */
/* gwadd3quick	Adds two numbers WITHOUT normalizing */
/* gwsub3quick	Subtracts two numbers WITHOUT normalizing */
/* gwadd3	Adds two numbers and normalizes them if necessary */
/* gwsub3	Subtracts two numbers and normalizes them if necessary */
/* gwfft	Perform the forward Fourier transform on a number */
/* gwsquare	Multiplies a number by itself */
/* gwmul(s,d)	Computes d=s*d.  NOTE: s is replaced by its FFT */
/* gwsafemul	Like gwmul but s is not replaced with its FFT */
/* gwfftmul(s,d) Computes d=s*d.  NOTE: s must have been previously FFTed */
/* gwfftfftmul(s1,s2,d) Computes d=s1*s2.  Both s1 and s2 must have */
/*		been previously FFTed */

EXTERNC void gw_setup ();
EXTERNC void gw_copy ();
EXTERNC void gw_addq ();
EXTERNC void gw_subq ();
EXTERNC void gw_add ();
EXTERNC void gw_sub ();
EXTERNC void gw_fft ();
EXTERNC void gw_square ();
EXTERNC void gw_mul ();
EXTERNC void gw_mulf ();

#define gwsetup(p)	{PARG = p; gw_setup ();}
#define gwswap(s,d)	{gwnum t; t = s; s = d; d = t;}
#define gwcopy(s,d)	{SRCARG = s; DESTARG = d; gw_copy ();}
#define gwadd3quick(s1,s2,d){SRCARG = s1; SRC2ARG = s2; DESTARG=d; gw_addq ();}
#define gwsub3quick(s1,s2,d){SRCARG = s2; SRC2ARG = s1; DESTARG=d; gw_subq ();}
#define gwaddquick(s,d)	{SRCARG = s; SRC2ARG = d; DESTARG = d; gw_addq ();}
#define gwsubquick(s,d)	{SRCARG = s; SRC2ARG = d; DESTARG = d; gw_subq ();}
#define gwadd(s,d)	{SRCARG = s; SRC2ARG = d; DESTARG = d; gw_add ();}
#define gwsub(s,d)	{SRCARG = s; SRC2ARG = d; DESTARG = d; gw_sub ();}
#define gwadd3(s1,s2,d)	{SRCARG = s1; SRC2ARG = s2; DESTARG = d; gw_add ();}
#define gwsub3(s1,s2,d)	{SRCARG = s2; SRC2ARG = s1; DESTARG = d; gw_sub ();}
#define gwfft(s,d)	{SRCARG = s; DESTARG = d; gw_fft ();}
#define gwsquare(s)	{DESTARG = s; gw_square ();}
#define gwfftmul(s,d)	{SRCARG = s; DESTARG = d; gw_mul ();}
#define gwfftfftmul(s1,s2,d){SRCARG=s1; SRC2ARG=s2; DESTARG=d; gw_mulf ();}
#define gwmul(s,d)	{gwfft(s,s); gwfftmul (s,d);}
#define gwsafemul(s,d)	{gwfft(s,t4); gwfftmul(t4,d);}

EXTERNC int egcdhlp ();
EXTERNC long egetval ();
EXTERNC void eaddhlp ();
EXTERNC void emuladdhlp ();
EXTERNC void emulsubhlp ();
#define getval(g,i,v)	{SRCARG=(void*)(i); SRC2ARG=g; *(v) = egetval ();}
#define addhlp(a)	{SRCARG=(void*)a; eaddhlp();}
#define muladdhlp(a,b)	{SRCARG=(void*)a; SRC2ARG=(void*)b; emuladdhlp();}
#define mulsubhlp(a,b)	{SRCARG=(void*)a; SRC2ARG=(void*)b; emulsubhlp();}

/* other macros */

#define timer()		((double)(unsigned long)clock()*1000.0/CLOCKS_PER_SEC)
#define bitset(a,i)	{ a[(i) >> 3] |= (1 << ((i) & 7)); }
#define bitclr(a,i)	{ a[(i) >> 3] &= ~(1 << ((i) & 7)); }
#define bittst(a,i)	(a[(i) >> 3] & (1 << ((i) & 7)))

/* function prototypes */

void	ell_dbl (gwnum x1, gwnum z1, gwnum x2, gwnum z2);
int	ell_dbl_norm (gwnum x1, gwnum x2);
void	ell_dbl_norm_delayed (gwnum x1, gwnum x2, gwnum z2);
void	ell_add (gwnum x1, gwnum z1, gwnum x2, gwnum z2);
int	ell_add_norm (gwnum x1, gwnum x2);
void	ell_add_norm_delayed (gwnum x1, gwnum x2, gwnum z2);
void	ell_mul (gwnum xx, gwnum zz, unsigned long n);
int	normalize (gwnum, gwnum);

/* Routine to allocate aligned memory for our big numbers */
/* Memory is allocated on 32-byte boundaries, with an additional */
/* 32 bytes prior to the data for storing useful stuff */

gwnum gwalloc ()
{
	char	*p, *q;
	p = (char *) malloc (32 + 32 + FFTLEN * sizeof (double));
	q = (char *) (((unsigned long) p + 63) & 0xFFFFFFE0);
	* (char **) (q - 32) = p;
	return ((gwnum) q);
}

/* Free one of our special numbers */

void gwfree (
	gwnum	q)
{
	free (* (char **) ((char *) q - 32));
}

/* Convert a double to a gwnum */

void dbltogw (double d, gwnum g)
{
	unsigned long i;
	double	base1, base2;
	int	bits;

	ECM_ADDR = (double *) g;
	bits = (int) (PARG / FFTLEN);
	base1 = (double) (1L << bits);
	base2 = (double) (1L << (bits+1));
	for (i = 0; i < FFTLEN; i++) {
		if (d < base1) {
			ECM_ADDR[i] = d * NORM_ARRAY[3*i+1];
			d = 0.0;
		} else {
			double	base, rem;
			base = is_big_word (PARG, FFTLEN, i) ? base2 : base1;
			rem = fmod (d, base);
			ECM_ADDR[i] = rem * NORM_ARRAY[3*i+1];
			d = (d - rem) / base;
		}
	}
	ECM_ADDR = NULL;
	((long *) g)[-1] = 0;	/* Clear needs-normalize counter */
}


/* Perform setup functions.  This includes decding how big an FFT to */
/* use, allocating memory, calling the FFT setup code, etc. */

void ecm_setup (
	int	p)
{
	unsigned long j;

/* Setup the assembly code */

	gwsetup (p);

/* Allocate memory */

	N = (unsigned long *) malloc ((p / 32 + 1) * sizeof(unsigned long));
	U = (unsigned long *) malloc ((p / 32 + 1) * sizeof(unsigned long));
	V = (unsigned long *) malloc ((p / 32 + 1) * sizeof(unsigned long));
	gg = gwalloc ();
	x = gwalloc ();
	z = gwalloc ();
	xs = gwalloc ();
	zs = gwalloc ();
	xorg = gwalloc ();
	zorg = gwalloc ();
	t1 = gwalloc ();
	t2 = gwalloc ();
	t3 = gwalloc ();
	t4 = gwalloc ();
	Ad4 = gwalloc ();
	xt = gwalloc ();
	zt = gwalloc ();
	xA = gwalloc ();
	zA = gwalloc ();
	xB = gwalloc ();
	zB = gwalloc ();
	xC = gwalloc ();
	zC = gwalloc ();
	nQx = (gwnum *) malloc ((D/2+2) * sizeof (gwnum));
	for (j = 0; j < D/2+2; j++) nQx[j] = gwalloc ();
	for (j = 0; j < GROUP_MODINV; j++) {
		xgrp[j] = gwalloc ();
		zgrp[j] = gwalloc ();
		prevgrp[j] = gwalloc ();
		tmpgrp[j] = gwalloc ();
	}
}

/* Perform cleanup functions. */

void ecm_cleanup ()
{
	unsigned long j;

/* Free memory */

	free (N);
	free (U);
	free (V);
	gwfree (gg);
	gwfree (x);
	gwfree (z);
	gwfree (xs);
	gwfree (zs);
	gwfree (xorg);
	gwfree (zorg);
	gwfree (t1);
	gwfree (t2);
	gwfree (t3);
	gwfree (t4);
	gwfree (Ad4);
	gwfree (xt);
	gwfree (zt);
	gwfree (xA);
	gwfree (zA);
	gwfree (xB);
	gwfree (zB);
	gwfree (xC);
	gwfree (zC);
	for (j = 0; j < D/2+2; j++) gwfree (nQx[j]);
	free (nQx);
	for (j = 0; j < GROUP_MODINV; j++) {
		gwfree (xgrp[j]);
		gwfree (zgrp[j]);
		gwfree (prevgrp[j]);
		gwfree (tmpgrp[j]);
	}
}


/* Use a simple sieve to find prime numbers */

#define MAX_PRIMES	6600
static	unsigned int num_primes = 0;
static	unsigned int *primes = NULL;
static	struct sieve_info {
	unsigned long first_number;
	unsigned int bit_number;
	unsigned int num_bits;
	unsigned long start;
	unsigned long end;
	char	array[4096];
} si;

/* Start sieve by allocate a sieve info structure */

void start_sieve (
	unsigned long start,
	unsigned long end)
{
	unsigned int i, fmax;

	if (start < 2) start = 2;
	si.start = start;
	si.end = end;
	si.first_number = start | 1;
	si.bit_number = 1;
	si.num_bits = 0;

	if (primes == NULL) {
		unsigned int f;
		primes = (unsigned int *) malloc (MAX_PRIMES * 2 * sizeof (unsigned int));
		for (i = 0, f = 3; i < MAX_PRIMES * 2; f += 2)
			if (isPrime (f)) primes[i] = f, i += 2;
	}

/* Determine the first bit to clear */

	fmax = (unsigned int) sqrt ((double) end);
	for (i = 0; i < MAX_PRIMES * 2; i += 2) {
		unsigned long f, r, bit;
		f = primes[i];
		if (f > fmax) break;
		if (si.first_number == 3) {
			bit = (f * f - 3) >> 1;
		} else {
			r = (unsigned long) (si.first_number % f);
			if (r == 0) bit = 0;
			else if (r & 1) bit = (f - r) / 2;
			else bit = (f + f - r) / 2;
			if (f == si.first_number + 2 * bit) bit += f;
		}
		primes[i+1] = bit;
	}
	num_primes = i;
}

/* Return next prime from the sieve */

unsigned long sieve ()
{
	unsigned int i;

/* Return next prime in the sieve */

ret:	while (si.bit_number < si.num_bits) {
		unsigned int bit = si.bit_number++;
		if (bittst (si.array, bit))
			return (si.first_number + 2 * bit);
	}

/* See if we are done */

	si.first_number += 2 * si.num_bits;
	if (si.first_number > si.end) 
		return (0xFFFFFFFF);

/* Fill the sieve with ones, then zero out the composites */
/*static unsigned long xxx, yyy;
  static double time1 = 0.0;
  RDTSC (xxx);*/
	memset (si.array, 0xFF, sizeof (si.array));
	si.num_bits = (unsigned int) (si.end - si.first_number) / 2 + 1;
	if (si.num_bits > sizeof (si.array) * 8)
		si.num_bits = sizeof (si.array) * 8;
	for (i = 0; i < num_primes; i += 2) {
		unsigned int f, bit;
		f = primes[i];
		for (bit = primes[i+1]; bit < si.num_bits; bit += f)
			bitclr (si.array, bit);
		primes[i+1] = bit - si.num_bits;
	}
	si.bit_number = 0;
/*RDTSC (yyy); time1 += yyy - xxx;*/

/* Return 2 as a special case */

	if (si.start == 2 && si.first_number == 3) return (2);
	goto ret;
}

void end_sieve ()
{
}

/**************************************************************
 *
 *	Functions
 *
 **************************************************************/

/* computes 2P=(x2:z2) from P=(x1:z1), uses the global variables Ad4 */

void ell_dbl (
	gwnum	x1,
	gwnum	z1,
	gwnum	x2,
	gwnum	z2)
{				/* 2 FFTs, 5 muls, 4 adds */
	gwadd3 (x1, z1, t1);	/* t1 = (x1 + z1)^2 */
	gwsquare (t1);
	gwsub3 (x1, z1, x2);	/* t2 = (x1 - z1)^2 (store in x2) */
	gwsquare (x2);
	gwsub3 (t1, x2, t3);	/* t3 = t1 - t2 = 4 * x1 * z1 */
	gwfftmul (Ad4, x2);	/* x2 = t2 * Ad4 */
	gwadd3 (x2, t3, z2);	/* z2 = (t2 * Ad4 + t3) * t3 */
	gwmul (t3, z2);
	gwmul (t1, x2);		/* x2 = x2 * t1 */
}

/* This routine is called prior to a series of many */
/* ell_add_fft and ell_dbl_fft calls.  The sequence */
/* ends by calling ell_add_fft_last. */

/* Note: We used to simply just FFT x1 and z1.  However, */
/* convolution error in computing (x1+z1)^2 and the like */
/* was too great.  Instead, we now save the FFTs of (x1+z1) */
/* and (x1-z1).  The multiplication by xdiff and zdiff is */
/* now more complicated, but convolution errors are reduced */
/* since only one argument of any multiply will involve a */
/* value that is the sum of two FFTs rather than computing */
/* a properly normalized sum and then taking the FFT. */

void ell_begin_fft (
	gwnum	x1,
	gwnum	z1,
	gwnum	x2,
	gwnum	z2)
{
	gwadd3 (x1, z1, t1);	/* t1 = (x1 + z1) */
	gwsub3 (x1, z1, t2);	/* t2 = (x1 - z1) */
	gwfft (t1, x2);
	gwfft (t2, z2);
}

/* Like ell_dbl, but the input arguments have already been FFTed */
/* The output arguments are also FFTed */

void ell_dbl_fft (
	gwnum	x1,
	gwnum	z1,
	gwnum	x2,
	gwnum	z2)
{				/* 12 FFTs, 4 adds */
	gwfftfftmul (x1, x1, t1);/* t1 = (x1 + z1)^2 */
	gwfftfftmul (z1, z1, x2);/* t2 = (x1 - z1)^2 (store in x2) */
	gwsub3 (t1, x2, t3);	/* t3 = t1 - t2 = 4 * x1 * z1 */
	gwfftmul (Ad4, x2);	/* x2 = t2 * Ad4 */
	gwadd3quick (x2, t3, z2);/* z2 = (t2 * Ad4 + t3) * t3 */
	gwmul (t3, z2);
	gwmul (t1, x2);		/* x2 = x2 * t1 */
	gwadd3 (x2, z2, t1);	/* t1 = (x2 + z2) */
	gwsub3 (x2, z2, t2);	/* t2 = (x2 - z2) */
	gwfft (t1, x2);
	gwfft (t2, z2);
}

/* Like ell_dbl except that z1 is one, and the x2 result is normalized */
/* so that z2 is one. */

int ell_dbl_norm (
	gwnum	x1,
	gwnum	x2)
{				/* 1 copy, 2 FFTs, 4 muls, 4 adds */
	gwcopy (x1, t1);
	* (double *) t1 += 1.0;
	gwsquare (t1);		/* t1 = (x1 + 1)^2 */
	gwadd3quick (x1, x1, t3);
	gwadd (t3, t3);		/* t3 = 4 * x1 */
	gwsub3 (t1, t3, x2);	/* t2 = t1 - t3 = (x1 - 1)^2 (store in x2) */
	gwfftmul (Ad4, x2);	/* x2 = t2 * Ad4 */
	gwadd3 (x2, t3, t2);	/* z2 = (t2 * Ad4 + t3) * t3 */
	gwmul (t3, t2);
	gwmul (t1, x2);		/* x2 = x2 * t1 */
	return (normalize (x2, t2));
}

/* Like ell_dbl_norm except that the normalize is postponed */
/* until a later date. */

void ell_dbl_norm_delayed (
	gwnum	x1,
	gwnum	x2,
	gwnum	z2)
{				/* 1 copy, 2 FFTs, 4 muls, 4 adds */
	gwcopy (x1, t1);
	* (double *) t1 += 1.0;
	gwsquare (t1);		/* t1 = (x1 + 1)^2 */
	gwadd3 (x1, x1, t3);
	gwadd (t3, t3);		/* t3 = 4 * x1 */
	gwsub3 (t1, t3, x2);	/* t2 = t1 - t3 = (x1 - 1)^2 (store in x2) */
	gwfftmul (Ad4, x2);	/* x2 = t2 * Ad4 */
	gwadd3 (x2, t3, z2);	/* z2 = (t2 * Ad4 + t3) * t3 */
	gwmul (t3, z2);
	gwmul (t1, x2);		/* x2 = x2 * t1 */
}

/* adds Q=(x2:z2) and R=(x1:z1) and puts the result in (x2:z2),
   Assumes that Q-R=P or R-Q=P where P=(xorg:zorg).  xorg and
   zorg must have already been FFTed. */

void ell_add (
	gwnum 	x1,
	gwnum 	z1,
	gwnum 	x2,
	gwnum 	z2)
{				/* 2 FFTs, 6 muls, 6 adds */
	gwadd3 (x1, z1, t1);	/* t1 = (x1 + z1)(x2 - z2) */
	gwsub3 (x2, z2, t2);
	gwmul (t2, t1);
	gwsub3 (x1, z1, t2);	/* t2 = (x1 - z1)(x2 + z2) */
	gwadd (x2, z2);
	gwmul (t2, z2);
	gwadd3 (z2, t1, x2);	/* x2 = (t2 + t1)^2 * zor */
	gwsquare (x2);
	gwfftmul (zorg, x2);
	gwsub (t1, z2);		/* z2 = (t2 - t1)^2 * xor */
	gwsquare (z2);
	gwfftmul (xorg, z2);
}

/* Like ell_add but input arguments have already been FFTed, */
/* difference argument is passed in, and output variable */
/* is also passed in.  Results are FFTed before returning. */

void ell_add_fft (
	gwnum 	x1,
	gwnum 	z1,
	gwnum 	x2,
	gwnum 	z2,
	gwnum 	xdiff,
	gwnum 	zdiff,
	gwnum 	x3,
	gwnum 	z3)
{				/* 12 FFTs, 6 adds */
	gwfftfftmul (x1, z2, t1);/* t1 = (x1 + z1)(x2 - z2) */
	gwfftfftmul (x2, z1, t2);/* t2 = (x1 - z1)(x2 + z2) */
	gwadd3 (t2, t1, t3);	/* x3 = (t2 + t1)^2 * zdiff */
	gwsquare (t3);
	gwsub (t1, t2);		/* z3 = (t2 - t1)^2 * xdiff */
	gwsquare (t2);
	gwsub3quick (xdiff, zdiff, t1);	/* zdiff = (xdiff - zdiff) */
	gwfftmul (t1, t3);
	gwadd3quick (xdiff, zdiff, t1);	/* xdiff = (xdiff + zdiff) */
	gwfftmul (t1, t2);
	gwadd3 (t3, t2, t1);	/* t1 = (x3 + z3) */
	gwsub3 (t3, t2, t2);	/* t2 = (x3 - z3) */
	gwfft (t1, x3);
	gwfft (t2, z3);
}

/* Like ell_add_fft but output arguments are not FFTed. */
/* Output argument cannot be the same as the xdiff/zdiff */
/* input arguments. */

void ell_add_fft_last (
	gwnum 	x1,
	gwnum 	z1,
	gwnum 	x2,
	gwnum 	z2,
	gwnum 	xdiff,
	gwnum 	zdiff,
	gwnum 	x3,
	gwnum 	z3)
{				/* 10 FFTs, 6 adds */
	gwfftfftmul (x1, z2, t1);/* t1 = (x1 + z1)(x2 - z2) */
	gwfftfftmul (x2, z1, t2);/* t2 = (x1 - z1)(x2 + z2) */
	gwadd3 (t2, t1, x3);	/* x3 = (t2 + t1)^2 * zor */
	gwsquare (x3);
	gwsub3quick (xdiff, zdiff, t4);	/* zdiff = (xdiff - zdiff) */
	gwfftmul (t4, x3);
	gwsub3 (t2, t1, z3);	/* z3 = (t2 - t1)^2 * xor */
	gwsquare (z3);
	gwadd3quick (xdiff, zdiff, t4);	/* xdiff = (xdiff + zdiff) */
	gwfftmul (t4, z3);
}

/* Like ell_add, but z1, z2, and zorg are one because of nomalization */
/* Result is normalized before returning */

int ell_add_norm (
	gwnum 	x1,
	gwnum 	x2)
{				/* 1 FFT, 4 muls, 1 adds */
	gwsub3 (x1, x2, t3);	/* z2 = (x1 - x2)^2 * xor */
	gwsquare (t3);
	gwfftmul (xorg, t3);
	gwsafemul (x1, x2);	/* x2 = (x1*x2 - 1)^2 * zor */
	* (double *) x2 -= 1.0;
	gwsquare (x2);
	return (normalize (x2, t3));
}

/* Like ell_add_norm except that the normalize is postponed */
/* until a later date. */

void ell_add_norm_delayed (
	gwnum 	x1,
	gwnum 	x2,
	gwnum	z2)
{				/* 1 FFT, 4 muls, 1 adds */
	gwsub3 (x1, x2, z2);	/* z2 = (x1 - x2)^2 * xor */
	gwsquare (z2);
	gwfftmul (xorg, z2);
	gwsafemul (x1, x2);	/* x2 = (x1*x2 - 1)^2 * zor */
	* (double *) x2 -= 1.0;
	gwsquare (x2);
}

/* Multiplies the point (xx,zz) by n using a combination */
/* of ell_dbl and ell_add calls */

#ifndef LUCAS_CHAINS
void ell_mul (
	gwnum	xx,
	gwnum	zz,
	unsigned long n)
{
	unsigned long c, zeros;

	for (zeros = 0; (n & 1) == 0; zeros++) n >>= 1;

	if (n > 1) {
		ell_begin_fft (xx, zz, xorg, zorg);

		c = (unsigned long)(1<<31);
		while ((c&n) == 0) c >>= 1;
		c >>= 1;

		/* If the second bit is zero, we can save one ell_dbl call */

		if (c&n) {
			gwcopy (xorg, xx); gwcopy (zorg, zz);
			ell_dbl_fft (xx, zz, xs, zs);
		} else {
			ell_dbl_fft (xorg, zorg, xx, zz);
			ell_add_fft (xorg, zorg, xx, zz, xorg, zorg, xs, zs);
			c >>= 1;
		}

		/* Do the rest of the bits */

		do {
			if (c&n) {
				if (c == 1) {
					ell_add_fft_last (xs, zs, xx, zz,
							  xorg, zorg, xx, zz);
				} else {
					ell_add_fft (xs, zs, xx, zz,
						     xorg, zorg, xx, zz);
					ell_dbl_fft (xs, zs, xs, zs);
				}
			} else {
				ell_add_fft (xx, zz, xs, zs,
					     xorg, zorg, xs, zs);
				ell_dbl_fft (xx, zz, xx, zz);
			}
			c >>= 1;
		} while (c);
	}

	while (zeros--) ell_dbl (xx, zz, xx, zz);
}
#endif


/* Perform an elliptic multiply using an algorithm developed by */
/* Peter Montgomery.  Basically, we try to find a near optimal */
/* Lucas chain of additions that generates the number we are */
/* multiplying by.  This minimizes the number of calls to ell_dbl */
/* and ell_add. */

/* The costing function assigns an ell_dbl call a cost of 12 and */
/* an ell_add call a cost of 12.  This cost estimates the number */
/* of forward and inverse transforms performed. */

#ifdef LUCAS_CHAINS
#define swap(a,b)	{t=a;a=b;b=t;}

unsigned long lucas_cost (
	unsigned long n,
	double	v)
{
	unsigned long c, d, e, t, dmod3, emod3;

	c = 0;
	while (n != 1) {
	    d = (unsigned long) (n/v+0.5); e = n - d;
	    d = d - e;

	    c += 12;

	    while (d != e) {
		if (d < e) {
			swap (d,e);
		}
		if (d <= e + (e >> 2)) {
			if ((dmod3 = d%3) == 3 - (emod3 = e%3)) {
				t = d;
				d = (d+d-e)/3;
				e = (e+e-t)/3;
				c += 36;
				continue;
			}
			if (dmod3 == emod3 && (d&1) == (e&1)) {
				d = (d-e) >> 1;
				c += 24;
				continue;
			}
		}
		if (d <= (e << 2)) {
			d = d-e;
			c += 12;
		} else if ((d&1) == (e&1)) {
			d = (d-e) >> 1;
			c += 24;
		} else if ((d&1) == 0) {
			d = d >> 1;
			c += 24;
		} else if ((dmod3 = d%3) == 0) {
			d = d/3-e;
			c += 48;
		} else if (dmod3 == 3 - (emod3 = e%3)) {
			d = (d-e-e)/3;
			c += 48;
		} else if (dmod3 == emod3) {
			d = (d-e)/3;
			c += 48;
		} else {
			e = e >> 1;
			c += 24;
		}
	    }
	    c += 12;
	    n = d;
	}

	return (c);
}

void lucas_mul (
	gwnum	xx,
	gwnum	zz,
	unsigned long n,
	double	v)
{
	unsigned long d, e, t, dmod3, emod3;

	while (n != 1) {
	    ell_begin_fft (xx, zz, xA, zA);			/* A */
	    ell_dbl_fft (xA, zA, xB, zB);			/* B = 2*A */
	    gwcopy (xA, xC); gwcopy (zA, zC);			/* C = A */

	    d = (unsigned long) (n/v+0.5); e = n - d;
	    d = d - e;

	    while (d != e) {
		if (d < e) {
			swap (d,e);
			gwswap (xA, xB); gwswap (zA, zB);
		}
		if (d <= e + (e >> 2)) {
			if ((dmod3 = d%3) == 3 - (emod3 = e%3)) {
				ell_add_fft (xA, zA, xB, zB, xC, zC, xs, zs);/* S = A+B */
				ell_add_fft (xA, zA, xs, zs, xB, zB, xt, zt);/* T = A+S */
				ell_add_fft (xs, zs, xB, zB, xA, zA, xB, zB);/* B = B+S */
				gwswap (xt, xA); gwswap (zt, zA);	/* A = T */
				t = d;
				d = (d+d-e)/3;
				e = (e+e-t)/3;
				continue;
			}
			if (dmod3 == emod3 && (d&1) == (e&1)) {
				ell_add_fft (xA, zA, xB, zB, xC, zC, xB, zB);/* B = A+B */
				ell_dbl_fft (xA, zA, xA, zA);	/* A = 2*A */
				d = (d-e) >> 1;
				continue;
			}
		}
		if (d <= (e << 2)) {
			ell_add_fft (xA, zA, xB, zB, xC, zC, xC, zC);/* B = A+B */
			gwswap (xB, xC); gwswap (zB, zC);	/* C = B */
			d = d-e;
		} else if ((d&1) == (e&1)) {
			ell_add_fft (xA, zA, xB, zB, xC, zC, xB, zB);/* B = A+B */
			ell_dbl_fft (xA, zA, xA, zA);		/* A = 2*A */
			d = (d-e) >> 1;
		} else if ((d&1) == 0) {
			ell_add_fft (xA, zA, xC, zC, xB, zB, xC, zC);/* C = A+C */
			ell_dbl_fft (xA, zA, xA, zA);		/* A = 2*A */
			d = d >> 1;
		} else if ((dmod3 = d%3) == 0) {
			ell_dbl_fft (xA, zA, xs, zs);		/* S = 2*A */
			ell_add_fft (xA, zA, xB, zB, xC, zC, xt, zt);/* T = A+B */
			ell_add_fft (xs, zs, xA, zA, xA, zA, xA, zA);/* A = S+A */
			ell_add_fft (xs, zs, xt, zt, xC, zC, xC, zC);/* B = S+B */
			gwswap (xB, xC); gwswap (zB, zC);	/* C = B */
			d = d/3-e;
		} else if (dmod3 == 3 - (emod3 = e%3)) {
			ell_add_fft (xA, zA, xB, zB, xC, zC, xs, zs);/* S = A+B */
			ell_add_fft (xA, zA, xs, zs, xB, zB, xB, zB);/* B = A+S */
			ell_dbl_fft (xA, zA, xs, zs);		/* S = 2*A */
			ell_add_fft (xs, zs, xA, zA, xA, zA, xA, zA);/* A = S+A */
			d = (d-e-e)/3;
		} else if (dmod3 == emod3) {
			ell_add_fft (xA, zA, xB, zB, xC, zC, xs, zs);/* S = A+B */
			ell_add_fft (xA, zA, xC, zC, xB, zB, xC, zC);/* C = A+C */
			gwswap (xs, xB); gwswap (zs, zB);	/* B = S */
			ell_dbl_fft (xA, zA, xs, zs);		/* S = 2*A */
			ell_add_fft (xs, zs, xA, zA, xA, zA, xA, zA);/* A = S+A */
			d = (d-e)/3;
		} else {
			ell_add_fft (xB, zB, xC, zC, xA, zA, xC, zC);/* C = C-B */
			ell_dbl_fft (xB, zB, xB, zB);		/* B = 2*B */
			e = e >> 1;
		}
	    }

	    ell_add_fft_last (xB, zB, xA, zA, xC, zC, xx, zz);			/* A = A+B */

	    n = d;
	}
}

/* Try a series of Lucas chains to find the cheapest. */
/* First try v = (1+sqrt(5))/2, then (2+v)/(1+v), then (3+2*v)/(2+v), */
/* then (5+3*v)/(3+2*v), etc.  Finally, execute the cheapest. */

void ell_mul (
	gwnum	xx,
	gwnum	zz,
	unsigned long n)
{
	unsigned long zeros;

	for (zeros = 0; (n & 1) == 0; zeros++) n >>= 1;

	if (n > 1) {
		unsigned long c, min;
		double	minv;

		min = lucas_cost (n, minv = 1.61803398875);/*v=(1+sqrt(5))/2*/

		c = lucas_cost (n, 1.38196601125);	/*(2+v)/(1+v)*/
		if (c < min) min = c, minv = 1.38196601125;

		c = lucas_cost (n, 1.72360679775);	/*(3+2*v)/(2+v)*/
		if (c < min) min = c, minv = 1.72360679775;

		c = lucas_cost (n, 1.580178728295);	/*(5+3*v)/(3+2*v)*/
		if (c < min) min = c, minv = 1.580178728295;

		c = lucas_cost (n, 1.632839806089);	/*(8+5*v)/(5+3*v)*/
		if (c < min) min = c, minv = 1.632839806089;

		c = lucas_cost (n, 1.612429949509);	/*(13+8*v)/(8+5*v)*/
		if (c < min) min = c, minv = 1.612429949509;

		c = lucas_cost (n, 1.620181980807);	/*(21+13*v)/(13+8*v)*/
		if (c < min) min = c, minv = 1.620181980807;

		c = lucas_cost (n, 1.617214616534);	/*(34+21*v)/(21+13*v)*/
		if (c < min) min = c, minv = 1.617214616534;

		c = lucas_cost (n, 1.618347119656);	/*(55+34*v)/(34+21*v)*/
		if (c < min) min = c, minv = 1.618347119656;

		c = lucas_cost (n, 1.617914406529);	/*(89+55*v)/(55+34*v)*/
		if (c < min) min = c, minv = 1.617914406529;

		lucas_mul (xx, zz, n, minv);
	}
	while (zeros--) ell_dbl (xx, zz, xx, zz);
}
#endif

/* Subtract a multiple of v from u, used in gcd and division routines. */
/* If z is given, then u/v is added to z */

void divide_helper (
	unsigned long *u,	/* Number to divide (and return modulo in */
	unsigned long *ulen,
	unsigned long *v,	/* Number to divide by */
	unsigned long vlen,
	unsigned long *z,	/* Number to add u/v into */
	unsigned long *zlen)
{
	unsigned long q;
	double u1, v1, q1;
	unsigned long diff_length;
	unsigned long i, carry;

/* Estimate u / v */

	diff_length = *ulen - vlen;
	u1 = (u[*ulen-1] * 4294967296.0 +
	     (*ulen > 1 ? u[*ulen-2] : 0)) * 4294967296.0 +
	     (*ulen > 2 ? u[*ulen-3] : 0);
	v1 = (v[vlen-1] * 4294967296.0 +
	     (vlen > 1 ? v[vlen-2] : 0)) * 4294967296.0 +
	     (vlen > 2 ? v[vlen-3] : 0);
	q1 = u1 / v1;
	if (q1 < 1.0 && u1 != v1) {
		q1 = q1 * 4294967296.0;
		diff_length--;
	}
	
/* Fudge the conversion to an integer for extra safety */

	q = (unsigned long) (q1 - 0.001);
	if (q == 0) {
		if (diff_length == 0) q = 1;
		else diff_length--, q = 4000000000;
	}

/* Compute u = u - q * v */

	CARRYH = CARRYL = RES = 0;
	for (i = diff_length; i < *ulen; i++) {
		RES = CARRYL;
		CARRYL = CARRYH;
		CARRYH = (long) CARRYH >> 31; /* Carry is really signed */
		addhlp (u[i]);
		if (i - diff_length < vlen) mulsubhlp (q, v[i-diff_length]);
		u[i] = RES;
	}

/* Decrease ulen */

	while (*ulen && u[*ulen-1] == 0) (*ulen)--;

/* Add u/v to z */

	if (z != NULL) {
		if (*zlen <= diff_length)
			memset (z, 0, (diff_length + 1) * sizeof (unsigned long));
		carry = q;
		for (i = diff_length; carry; i++) {
			unsigned long hi, lo;
			lo = (z[i] & 0xFFFF) + (carry & 0xFFFF);
			carry = (carry >> 16) + (lo >> 16);
			lo = lo & 0xFFFF;
			hi = (z[i] >> 16) + carry;
			carry = hi >> 16;
			hi = hi & 0xFFFF;
			z[i] = (hi << 16) + lo;
		}
		if (*zlen < i) *zlen = i;
	}
}

/* Compare two binary numbers */

int compare (
	unsigned long *u,
	unsigned long ulen,
	unsigned long *v,
	unsigned long vlen)
{
	int	i;

	if (ulen > vlen) return (1);
	if (ulen < vlen) return (-1);

	for (i = ulen-1; i >= 0; i--) {
		if (u[i] > v[i]) return (1);
		if (u[i] < v[i]) return (-1);
	}
	return (0);
}

/* Divide one binary number by another */

int bindiv (
	unsigned long *n,
	unsigned long *nlen,
	unsigned long *u,
	unsigned long ulen)
{

/* Divide n by u */

	Vlen = 0;
	while (*nlen) {
		if (compare (n, *nlen, u, ulen) < 0) return (FALSE);
		divide_helper (n, nlen, u, ulen, V, &Vlen);
	}

/* Copy n/u to n */

	memcpy (n, V, Vlen * sizeof (unsigned long));
	*nlen = Vlen;
	return (TRUE);
}

/* Set N, the number we are trying to factor */

int setN ()
{
	unsigned long *outptr;
	unsigned long p;
	FILE	*fd;
	char	buf[500];

/* Create the binary representation of the number we are factoring */

	outptr = N;
	for (p = PARG; p >= 32; p -= 32) *outptr++ = 0xFFFFFFFF;
	*outptr++ = (1L << p) - 1;
	Nlen = (outptr - N);

/* Open file of known factors */

	fd = fopen ("lowm.txt", "r");
	if (fd == NULL) return (TRUE);

/* Loop until the entire file is processed */
/* We are looking for lines of the form: "M( 2843 )C: 142151" */

	while (fscanf (fd, "%s", &buf) != EOF) {
		char	*bufp;

		if (buf[0] != 'M') continue;
		fscanf (fd, "%ld", &p);
		if (p > PARG) break;
		if (p < PARG) continue;
		fscanf (fd, "%s", &buf);
		if (buf[1] != 'C') continue;

/* What is the factor?  Store it in U */

		U[0] = 0;
		Ulen = 1;
		fscanf (fd, "%s", buf);
		for (bufp = buf; *bufp; bufp++) {
			unsigned long i, carry;
			carry = *bufp - '0';
			for (i = 0; i < Ulen; i++) {
				unsigned long hi, lo;
				hi = U[i] >> 16;
				lo = U[i] & 0xFFFF;
				lo = lo * 10 + carry;
				carry = lo >> 16;
				lo = lo & 0xFFFF;
				hi = hi * 10 + carry;
				carry = hi >> 16;
				hi = hi & 0xFFFF;
				U[i] = (hi << 16) + lo;
			}
			if (carry) U[Ulen++] = carry;
		}

/* Divide N by factor */

		if (!bindiv (N, &Nlen, U, Ulen)) {
			fclose (fd);
			OutputBoth ("Factor in lowm.txt does not divide 2^P-1\n");
			return (FALSE);
		}
	}

/* Close file and return */

	fclose (fd);
	return (TRUE);
}

/* Convert a value from FFT-ready format to binary */

void convertToBinary (
	gwnum	gg,
	unsigned long *v,
	unsigned long *vlen)
{
	long	val;
	int	j, bits, bits_in_a_word, bitsout, carry;
	unsigned long i, *outptr;

/* The assembly code can leave the value unnormalized. */
/* Multiply by one, just to normalize the value */

	dbltogw (1.0, t4);
	gwmul (t4, gg);

/* Check for a carry out of the most significant word */

	carry = 0;
	for (i = FFTLEN-1; i; i--) {
		getval (gg, i, &val);
		if (val != 0) {
			carry = (val > 0) ? 0 : -1;
			break;
		}
	}

/* Collect bits until we have all of them */

	bits_in_a_word = (int) (PARG / FFTLEN);
	bitsout = 0;
	outptr = v;
	*outptr = 0;
	for (i = 0; i < FFTLEN; i++) {
		getval (gg, i, &val);
		bits = bits_in_a_word;
		if (is_big_word (PARG, FFTLEN, i)) bits++;
		val += carry;
		for (j = 0; j < bits; j++) {
			*outptr >>= 1;
			if (val & 1) *outptr += 0x80000000;
			val >>= 1;
			bitsout++;
			if (bitsout == 32) {
				outptr++;
				*outptr = 0;
				bitsout = 0;
			}
		}
		carry = val;
	}
	*outptr >>= (32 - bitsout);
	outptr++;

/* Set the length */

	*vlen = (outptr - v);
	while (*vlen && v[*vlen-1] == 0) (*vlen)--;
}

/* Perform multi-precision binary arithmetic for the */
/* extended GCD code.  Computes e1 = e1 + e2 * e3 */

void extmuladd (
	unsigned long *e2,
	unsigned long e2len,
	unsigned long *e3,
	unsigned long e3len,
	unsigned long *e1,
	unsigned long *e1len)
{
	unsigned long i, j, k;

	CARRYH = CARRYL = 0;
	for (i = 0; i < e2len + e3len - 1; i++) { /* loop through output */
		if (i < e2len) {
			k = i;
			j = 0;
		} else {
			k = e2len - 1;
			j = i - e2len + 1;
		}
		RES = CARRYL;
		CARRYL = CARRYH;
		CARRYH = 0;
		if (i < *e1len) addhlp (e1[i]);
		for ( ; ; j++, k--) {
			muladdhlp (e2[k], e3[j]);
			if (j == e3len-1 || k == 0) break;
		}
		e1[i] = RES;
	}
	if (CARRYL) e1[i++] = CARRYL;
	if (CARRYH) e1[i++] = CARRYH;
	*e1len = i;
}

/* Convert a binary value back to the FFT format */

void exttogw (
	unsigned long *e1,
	unsigned long e1len,	/* Must be at least one */
	int	neg,
	gwnum	g)
{
	unsigned long i, mask1, mask2;
	int	bits1, bits2, bits_in_next_binval;
	unsigned long binval, carry;

	ECM_ADDR = (double *) g;
	bits1 = (int) (PARG / FFTLEN);
	bits2 = bits1 + 1;
	mask1 = (1L << bits1) - 1;
	mask2 = (1L << bits2) - 1;
	binval = *e1++; e1len--; bits_in_next_binval = 32;
	carry = 0;
	for (i = 0; i < FFTLEN; i++) {
		int	big_word, bits;
		long	value, mask;
		big_word = is_big_word (PARG, FFTLEN, i);
		bits = big_word ? bits2 : bits1;
		mask = big_word ? mask2 : mask1;
		value = binval & mask;
		if (neg) value = value ^ mask;
		value = value + carry;
		if (value > (mask >> 1) && i != FFTLEN - 1) {
			value = value - (mask + 1);
			carry = 1;
		} else {
			carry = 0;
		}
		ECM_ADDR[i] = value * NORM_ARRAY[3*i+1];

		binval >>= bits;
		if (e1len == 0) continue;
		if (bits_in_next_binval < bits) {
			if (bits_in_next_binval)
				binval |= (*e1 >> (32 - bits_in_next_binval)) << (32 - bits);
			bits -= bits_in_next_binval;
			e1++; e1len--; bits_in_next_binval = 32;
			if (e1len == 0) continue;
		}
		binval |= (*e1 >> (32 - bits_in_next_binval)) << (32 - bits);
		bits_in_next_binval -= bits;
	}
	ECM_ADDR = NULL;
	((long *) g)[-1] = 0;	/* Clear needs-normalize counter */
}

/* Perform multi-precision binary arithmetic for the */
/* extended GCD code.  Computes A * e1 + B * e2 and */
/* C * e1 + D * e2.  Values are then stored in e1 and e2. */

void extmmadd (
	unsigned long *e1,
	unsigned long *e1len,
	unsigned long *e2,
	unsigned long *e2len,
	unsigned long A,
	unsigned long B,
	unsigned long C,
	unsigned long D)
{
	unsigned long i, len;
	unsigned long carry1h, carry1l, carry2h, carry2l;

	len = (*e1len > *e2len) ? *e1len : *e2len;
	carry1h = carry1l = carry2h = carry2l = 0;
	for (i = 0; i < len; i++) {
		unsigned long x, y;
		x = (i < *e1len) ? e1[i] : 0;
		y = (i < *e2len) ? e2[i] : 0;

		CARRYH = 0;
		CARRYL = carry1h;
		RES = carry1l;
		muladdhlp (A, x);
		muladdhlp (B, y);
		e1[i] = RES;
		carry1l = CARRYL;
		carry1h = CARRYH;

		CARRYH = 0;
		CARRYL = carry2h;
		RES = carry2l;
		muladdhlp (C, x);
		muladdhlp (D, y);
		e2[i] = RES;
		carry2l = CARRYL;
		carry2h = CARRYH;
	}
	if (carry1l) e1[i++] = carry1l;
	if (carry1h) e1[i++] = carry1h;
	*e1len = i;
	if (carry2l) e2[len++] = carry2l;
	if (carry2h) e2[len++] = carry2h;
	*e2len = len;
}

/* Perform multi-precision binary arithmetic for the */
/* extended GCD code.  Computes A * e1 - B * e2 and */
/* D * e2 - C * e1 .  Values are then stored in e1 and e2 */

void extmmsub (
	unsigned long *e1,
	unsigned long *e1len,
	unsigned long *e2,
	unsigned long *e2len,
	unsigned long A,
	unsigned long B,
	unsigned long C,
	unsigned long D)
{
	unsigned long i, len;
	unsigned long carry1h, carry1l, carry2h, carry2l;

	len = (*e1len > *e2len) ? *e1len : *e2len;
	carry1h = carry1l = carry2h = carry2l = 0;
	for (i = 0; i < len; i++) {
		unsigned long x, y;
		x = (i < *e1len) ? e1[i] : 0;
		y = (i < *e2len) ? e2[i] : 0;

		CARRYH = (long) carry1h >> 31;	/* Carry is really signed */
		CARRYL = carry1h;
		RES = carry1l;
		muladdhlp (A, x);
		mulsubhlp (B, y);
		e1[i] = RES;
		carry1l = CARRYL;
		carry1h = CARRYH;

		CARRYH = (long) carry2h >> 31;	/* Carry is really signed */
		CARRYL = carry2h;
		RES = carry2l;
		muladdhlp (D, y);
		mulsubhlp (C, x);
		e2[i] = RES;
		carry2l = CARRYL;
		carry2h = CARRYH;
	}
#ifdef DEBUG
	if (carry1l || carry1h || carry2l || carry2h)
		OutputBoth ("Carry in mmsub!\n");
#endif
	while (*e1len && e1[*e1len-1] == 0) (*e1len)--;
	while (*e2len && e2[*e2len-1] == 0) (*e2len)--;
}

/* Computes the modular inverse of a number */
/* This is done using the extended GCD algorithm */
/* Return TRUE if we accidentally find a factor */

int modinv (
	gwnum b)
{
	unsigned long *t, *e1, *e2, *e3;
	unsigned long tlen, e1len, e2len, e3len;
	int	neg;
/*loop:gwcopy (b,zorg);*/

/* Convert input number to binary */
  
	convertToBinary (b, V, &Vlen);

/* Copy the number we are factoring */

	memcpy (U, N, Nlen * sizeof (unsigned long));
	Ulen = Nlen;

/* Three auxillary binary numbers manage the extended GCD calculations */
/* To save memory, we reuse the space of three temporaries */

	e1 = (unsigned long *) t1;
	e2 = (unsigned long *) t2;
	e1len = 0;
	e2[0] = 1; e2len = 1;
	neg = TRUE;

/* Trim V until it is smaller than U */

	while (compare (U, Ulen, V, Vlen) < 0)
		divide_helper (V, &Vlen, U, Ulen, NULL, NULL);

/* Do the extended GCD */

	while (Vlen > 0) {
		unsigned int diff_length;

/* If the quotient will fit in one word, use the GCD helper */
/* function which can do several GCD steps in single precision, */
/* postponing multi-precision operations as long as possible. */

/* The assembly language egcdhlp routine returns A,B,C,D */
/* as defined in Knuth vol. 2's description of extended GCD for */
/* large numbers. */

/* There is a remote chance that the GCD helper routine can not */
/* correctly determine a single quotient.  In that case egcdhlp */
/* returns FALSE and we use the slower code that can accumulate */
/* the proper quotient. */

		diff_length = Ulen - Vlen;
		if ((diff_length == 0 ||
		     (diff_length == 1 && U[Ulen-1] < V[Vlen-1])) &&
		     egcdhlp ()) {
			extmmadd (e1, &e1len, e2, &e2len,
				  EGCD_A, EGCD_B, EGCD_C, EGCD_D);
			if (EGCD_ODD) {
				extmmsub (V, &Vlen, U, &Ulen,
					  EGCD_B, EGCD_A, EGCD_D, EGCD_C);
				neg = !neg;
				t = U; tlen = Ulen;
				U = V; Ulen = Vlen;
				V = t; Vlen = tlen;
			} else {
				extmmsub (U, &Ulen, V, &Vlen,
					  EGCD_A, EGCD_B, EGCD_C, EGCD_D);
			}
#ifdef DEBUG
			if (compare (U, Ulen, V, Vlen) < 0)
				OutputBoth ("U < V in modular inverse!\n");
#endif
			continue;
		}

/* Build the quotient in e3 one word at a time (until V > U) */

		e3 = (unsigned long *) t3;
		e3len = 0;
		do
			divide_helper (U, &Ulen, V, Vlen, e3, &e3len);
		while (compare (U, Ulen, V, Vlen) >= 0);

/* Make U the larger number and maintain extended GCD information */

		extmuladd (e2, e2len, e3, e3len, e1, &e1len);
		t = e1; tlen = e1len;
		e1 = e2; e1len = e2len;
		e2 = t; e2len = tlen;
		neg = !neg;
		t = U; tlen = Ulen;
		U = V; Ulen = Vlen;
		V = t; Vlen = tlen;
	}

/* Convert from binary to FFT-ready representation */

	exttogw (e1, e1len, neg, b);
/* sanity check - multiply and see if value is one */
/*  gwcopy (zorg, t2);
    gwsafemul (b, zorg);
    dbltogw (1.0, t1); gwmul (t1, zorg);
    convertToBinary (zorg, V, &Vlen);
    while (compare (N, Nlen, V, Vlen) < 0)
	divide_helper (V, &Vlen, N, Nlen, NULL, NULL);
    if (V[0] != 1 || Vlen != 1){
    OutputStr ("Modinv disaster\n"); gwcopy (t2, b); goto loop;} */

/* Return TRUE if we unexpectedly found a factor */

	return (Ulen > 1 || U[0] > 1);
}

/* Computes the modular inverse of an array of numbers */
/* Uses extra multiplications to make only one real modinv call */
/* Uses the simple formula 1/a = b * 1/ab, 1/b = a * 1/ab */
/* Return TRUE if we accidentally find a factor */


int grouped_modinv (
	gwnum	*b,
	unsigned int size,
	gwnum	*tmp)
{
	unsigned int i;
	gwnum	*orig_tmp;

/* Handle group of 1 as a special case */

	if (size == 1) return (modinv (*b));

/* Handle an odd size */

	orig_tmp = tmp;
	if (size & 1) {
		gwswap (b[0], *tmp);
		tmp++;
	}

/* Multiply each pair of numbers */

	for (i = (size & 1); i < size; i += 2) {
		gwfft (b[i], b[i]);
		gwfft (b[i+1], b[i+1]);
		gwfftfftmul (b[i], b[i+1], *tmp);
		tmp++;
	}

/* Recurse */

	if (grouped_modinv (orig_tmp, (size+1) / 2, tmp)) return (TRUE);

/* Handle an odd size */

	if (size & 1) {
		gwswap (b[0], *orig_tmp);
		orig_tmp++;
	}

/* Now perform multiplications on each pair to get the modular inverse */

	for (i = (size & 1); i < size; i += 2) {
		gwfft (*orig_tmp, *orig_tmp);
		gwfftfftmul (*orig_tmp, b[i], b[i]);
		gwfftfftmul (*orig_tmp, b[i+1], b[i+1]);
		gwswap (b[i], b[i+1]);
		orig_tmp++;
	}
	return (FALSE);
}

/* Takes a an array of points (a,b) and multiplies each by */
/* a value such that b will be one */
/* Return TRUE if we accidentally find a factor */

int grouped_normalize (
	gwnum	*a,
	gwnum	*b,
	unsigned int size)
{
	unsigned int i;

/* Compute the modular inverses and scale up the first input values */

	if (grouped_modinv (b, size, tmpgrp)) return (TRUE);
	for (i = 0; i < size; i++) gwmul (b[i], a[i]);
	return (FALSE);
}

/* Takes a point (a,b) and multiplies it by a value such that b will be one */
/* Return TRUE if we accidentally find a factor */

int normalize (
	gwnum	a,
	gwnum	b)
{

/* Compute the modular inverse and scale up the first input value */

	if (modinv (b)) return (TRUE);
	gwmul (b, a);
	return (FALSE);
}

/* Do a GCD to see if a factor was found */
/* Destroys the input value, the GCD is returned in U */

int gcd (
	gwnum	gg)
{
	return (modinv (gg));
}


/* From R. P. Brent, priv. comm. 1996:
Let s > 5 be a pseudo-random seed (called $\sigma$ in the Tech. Report),

	u/v = (s^2 - 5)/(4s)

Then starting point is (x_1, y_1) where

	x_1 = (u/v)^3
and
	a = (v-u)^3(3u+v)/(4u^3 v) - 2
*/

void choose12 (
	gwnum 	x,
	gwnum 	z,
	double 	curve)
{
	dbltogw (curve, zs);
	gwcopy (zs, xs);
	gwsquare (xs);			/* s^2 */
	dbltogw (5.0, t2);
	gwsub (t2, xs);			/* u = s^2 - 5 */
	dbltogw (4.0, t4);
	gwmul (t4, zs);			/* v = 4*s */
	gwcopy (xs, x);
	gwsquare (x);
	gwsafemul (xs, x);		/* x = u^3 */
	gwcopy (zs, z);
	gwsquare (z);
	gwsafemul (zs, z);		/* z = v^3 */

	/* Now for A. */
	gwcopy (zs, t2);
	gwsub (xs, t2);
	gwcopy (t2, t3);
	gwsquare (t2);
	gwmul (t3, t2);		/* (v-u)^3 */
	gwcopy (xs, t3);
	gwadd (t3, t3);
	gwadd (xs, t3);
	gwadd (zs, t3);
	gwmul (t3, t2);		/* An = (v-u)^3 (3u+v) */
	gwcopy (zs, t3);
	gwsafemul (xs, t3);
	gwsquare (xs);
	gwsafemul (xs, t3);
	dbltogw (4.0, Ad4);
	gwmul (t3, Ad4);	/* An/Ad is now A + 2 */
	normalize (Ad4, t2);	/* Normalize so that An is one */
	dbltogw (4.0, t1);	/* For extra speed, precompute Ad * 4 */
	gwmul (t1, Ad4);
	gwfft (Ad4, Ad4);	/* Even more speed, save FFT of Ad4 */
}

/* Format cpu time as a string */

char * tottime (
	double	temp)
{
static	char	tbuf[15];
	sprintf (tbuf, "%.3f", (timer () - temp) / 1000.0);
	return (tbuf);
}

/* Clear the exponent from the work-to-do-file */

void updateIniFile (
	unsigned long p,
	unsigned long B1,
	unsigned long curves_to_do,
	unsigned long curves_done)
{
	unsigned int i, lines;
	char	buf[80];

	IniFileOpen (WORKTODO_FILE, 0);
	lines = IniGetNumLines (WORKTODO_FILE);
	for (i = 1; i <= lines; i++) {
		struct work_unit w;
		parseWorkToDoLine (i, &w);
		if (p == w.p &&
		    B1 == w.B1 &&
		    curves_to_do == w.curves_to_do) { 
			IniDeleteLine (WORKTODO_FILE, i);
			if (curves_done < curves_to_do) {
				sprintf (buf, "%lu,%lu,%lu,%lu,%lu,%.0f",
					 w.p, w.B1, w.B2, w.curves_to_do,
					 curves_done, w.curve);
				IniInsertLineAsString (WORKTODO_FILE, i, "ECM", buf);
			}
			break;
		}
	}
	IniFileClose (WORKTODO_FILE);
}

/* If the sum of the output values is an error (such as infinity or NaN) */
/* then return TRUE. */

int test_sumout ()
{
	short fp_status;
#if defined (__linux__) || defined (__EMX__)
	__asm__ __volatile__ (
		"fldl %1\n fxam\n fnstsw %0\n fstpl %1"
		: "=m" (fp_status)
		: "m " (SUMOUT)
	);
#else
	__asm fld SUMOUT
	__asm fxam
	__asm fnstsw fp_status
	__asm fstp SUMOUT
#endif
	return (fp_status & 0x0100);
}

/* These routines manage the computing of mQ in normalized stage 2 */

unsigned int mQ_array_counter = 0;
unsigned int mQ_group_counter = 0;
void mQ_init ()
{
	mQ_array_counter = GROUP_MODINV;
	mQ_group_counter = 0;
}
int mQ_next (gwnum *ret)
{
	unsigned int i;

	/* Return one of the precomputed values in the array */
	if (mQ_array_counter < GROUP_MODINV) {
		*ret = xgrp[mQ_array_counter++];
		return (FALSE);
	}
	/* Compute another group of mQ values */
	mQ_group_counter++;
	switch (mQ_group_counter) {
	/* On first call, return Q (in x) */
	/* and compute next GROUP_MODINV-1 values */
	case 1:
		/* Copy Q to prevgrp[0] */
		gwcopy (x, xgrp[0]);
		gwcopy (z, zgrp[0]);

		/* Compute (2D+1)Q */
		ell_mul (x, z, 2*D+1);
		gwswap (x, xgrp[1]);
		gwswap (z, zgrp[1]);

		/* Compute next GROUP_MODINV-2 values */
		for (i = 2; i < GROUP_MODINV; i++) {
			gwfft (xgrp[i-2], xorg);
			gwfft (zgrp[i-2], zorg);
			gwcopy (xgrp[i-1], xgrp[i]);
			gwcopy (zgrp[i-1], zgrp[i]);
			ell_add (nQx[(D>>1)+1], zgrp[0], xgrp[i], zgrp[i]);
		}
		break;
	/* Compute the second group of values */
	case 2:
		for (i = 0; i < GROUP_MODINV; i++)
			gwswap (xgrp[i], prevgrp[i]);
		dbltogw (1.0, z);
		gwfft (prevgrp[GROUP_MODINV-2], xorg);
		gwcopy (prevgrp[GROUP_MODINV-1], xgrp[0]);
		ell_add_norm_delayed (nQx[(D>>1)+1], xgrp[0], zgrp[0]);
		gwfft (prevgrp[GROUP_MODINV-1], xorg);
		gwfft (z, zorg);
		gwcopy (xgrp[0], xgrp[1]);
		gwcopy (zgrp[0], zgrp[1]);
		ell_add (nQx[(D>>1)+1], z, xgrp[1], zgrp[1]);
		for (i = 2; i < GROUP_MODINV; i++) {
			gwfft (xgrp[i-2], xorg);
			gwfft (zgrp[i-2], zorg);
			gwcopy (xgrp[i-1], xgrp[i]);
			gwcopy (zgrp[i-1], zgrp[i]);
			ell_add (nQx[(D>>1)+1], z, xgrp[i], zgrp[i]);
		}
		for (i = 1; i < GROUP_MODINV; i *= 2)
			ell_dbl (nQx[(D>>1)+1], z, nQx[(D>>1)+1], z);
		if (normalize (nQx[(D>>1)+1], z)) return (TRUE);
		break;
	/* Compute another group of values using two previous */
	/* sets of values. */
	default:
		for (i = 0; i < GROUP_MODINV; i++) {
			gwfft (prevgrp[i], xorg);
			gwcopy (xgrp[i], prevgrp[i]);
			ell_add_norm_delayed (nQx[(D>>1)+1], xgrp[i], zgrp[i]);
		}
	}

/* Normalize the group of values */

	if (grouped_normalize (xgrp, zgrp, GROUP_MODINV)) return (TRUE);

/* Return the first value in the array */

	mQ_array_counter = 1;
	*ret = xgrp[0];
	return (FALSE);
}

/* Test if N is a probable prime */
/* Compute i^(N-1) mod N for i = 3,5,7 */

int isProbablePrime ()
{
	unsigned long i, j, c;
	for (i = 3; i <= 7; i += 2) {
		dbltogw ((double) 1.0, t1);
		dbltogw ((double) i, t2);
		gwfft (t2, t2);
		for (j = 1; j <= Nlen; j++) {
			for (c = 0x80000000; c; c >>= 1) {
				gwsquare (t1);
				if (N[Nlen-j] & c) gwfftmul (t2, t1);
			}
		}
		convertToBinary (t1, V, &Vlen);
		while (compare (N, Nlen, V, Vlen) < 0)
			divide_helper (V, &Vlen, N, Nlen, NULL, NULL);
		if (V[0] != i) return (FALSE);
	}
	return (TRUE);
}

/*#define CHECK_LOWM*/
#ifdef CHECK_LOWM
void prime_check ()
{
	unsigned long p, last_p;
	FILE	*fd;
	char	buf[500];

/* Open file of known factors */

	fd = fopen ("lowm.txt", "r");
	if (fd == NULL) return;

/* Loop until the entire file is processed */
/* We are looking for lines of the form: "M( 2843 )C: 142151" */

	last_p = 0;
	while (fscanf (fd, "%s", &buf) != EOF) {
		if (buf[0] != 'M') continue;
		fscanf (fd, "%ld", &p);
		if (p == last_p) continue;

		ecm_setup (p);
		if (!setN ()) return;
		if (isProbablePrime ()) {
			sprintf (buf, "M%ld is probable prime\n", p);
			OutputBoth (buf);
		} else {
			sprintf (buf, "M%ld OK\n", p);
			OutputStr (buf);
		}
		ecm_cleanup ();
		last_p = p;
	}

/* Close file and return */

	fclose (fd);
}
#endif

/**************************************************************
 *
 *	Main Function
 *
 **************************************************************/

int ecm (
	unsigned long p,		/* Mersenne exponent */
	unsigned long B,		/* Pass 1 bound */
	unsigned long C,		/* Pass 2 bound */
	unsigned long curves_to_do,	/* Number of curves to test */
	unsigned long curves_completed,	/* Number of curves already done */
	double	specific_curve)		/* Debug only - curve to test */
{
	unsigned long SQRT_B;
	double	curve;
	unsigned long i, j, m, pass, prime;
	char	buf[100];
	int	retval, stage;
	time_t	start_time, current_time;

/* Get the current time */

	time (&start_time);

/* Choose a default value for the second bound if none was specified */

	if (C == 0) C = (B >= 42900000) ? 4290000000 : B * 100;
	if (C < B) C = B;

/* Determine stage 2's size */

	D = (unsigned long) sqrt (C / 2.0);
	if (D % 2) D++; /* ensures m and m-D have same parity in step 2 */

/* Set other constants */

	SQRT_B = (unsigned long) sqrt (B);

/* Perform setup functions */

	if (! loadfac ()) return (FALSE);
	ecm_setup (p);
	SUMOUT = 0;
	if (!setN ()) return (FALSE);

#ifdef CHECK_LOWM
prime_check ();
#endif

/*#define TIMING*/
#ifdef TIMING
{
unsigned long i, xxx, yyy;
double	time1, time2, time3, time4, time5, time6, time7;
dbltogw (283457283657.0, nQx[2]);
for (i = 1; i <= 50; i++) gwsquare (nQx[2]); /* gen big random number */
gwcopy (nQx[2], nQx[3]);
gwcopy (nQx[2], nQx[1]);
time1 = time2 = time3 = time4 = time5 = time6 = time7 = 0.0;
for (i = 1; i <= 100; i++) {
	RDTSC (xxx);
	gwsquare (nQx[2]);
	RDTSC (yyy); time1 += yyy - xxx;
	RDTSC (xxx);
	gwmul (nQx[2], nQx[3]);
	RDTSC (yyy); time2 += yyy - xxx;
	RDTSC (xxx);
	gwfftmul (nQx[2], nQx[3]);
	RDTSC (yyy); time3 += yyy - xxx;
	RDTSC (xxx);
	normalize (nQx[1], nQx[3]);
	RDTSC (yyy); time4 += yyy - xxx;
	RDTSC (xxx);
	gwfftfftmul (nQx[2], nQx[2], nQx[2]);
	RDTSC (yyy); time6 += yyy - xxx;
	RDTSC (xxx);
	gwfft (nQx[2], nQx[2]);
	RDTSC (yyy); time5 += yyy - xxx;
	RDTSC (xxx);
	gwadd (nQx[2], nQx[2]);
	RDTSC (yyy); time7 += yyy - xxx;
	gwcopy (nQx[1], nQx[2]);
}
OutputStr ("100 squares: ");  OutputClocks (time1);
OutputStr ("100 muls: ");  OutputClocks (time2);
OutputStr ("100 fftmuls: ");  OutputClocks (time3);
OutputStr ("100 ffts: ");  OutputClocks (time5);
OutputStr ("100 fftfftmuls: ");  OutputClocks (time6);
OutputStr ("100 normalizes: ");  OutputClocks (time4);
OutputStr ("100 adds: ");  OutputClocks (time7);
	RDTSC (xxx);
	start_sieve (2, C+D+D);
	i = 0;
	while (sieve () < 0xFFFFFFFF)
		if ((++i & 0xFFF) == 0) {
			RDTSC (yyy); time1 += yyy - xxx;
			RDTSC (xxx);
		}
	RDTSC (yyy); time1 += yyy - xxx; if (yyy < xxx) time1 += 4294967296.0;
	end_sieve ();
sprintf (buf, "Sieve: %ld primes found.  ", i);
OutputStr (buf);  OutputClocks (time1);
}		
#endif

/*#define TESTING*/
#ifdef TESTING	
{ unsigned long i;
ECM_ADDR = (double *) t1;
for (i = 0; i < FFTLEN; i++)
	ECM_ADDR[i] = (i == 2) ? 1.0 : 0.0;
ECM_ADDR = (double *) t2;
for (i = 0; i < FFTLEN; i++)
	ECM_ADDR[i] = (i < 16) ? 1.0 : 0.0;
gwcopy (t1,t3);
gwmul (t1, t2);
gwfftmul (t1, t2);
gwadd (t2, t2);
gwcopy (t1, t2);
gwfftfftmul (t1, t2, t2);
gwsquare (t3);
}
#endif

/*#define echk*/
#ifdef echk
for ( ; ; p++) {
ecm_setup (p);
dbltogw (1.0, t1);
dbltogw (283457283654.0, t2);
gwmul (t1, t2);
dbltogw (23876287342234.0, t3);
gwfftmul (t1, t3);
ERRCHK = 1; MAXERR = 0.0;
for (int i = 1; i <= 1000; i++) {
gwsquare (t3);
gwsafemul (t2, t3);
gwsquare (t2);
}
p++;
}
#endif

	srand ((unsigned) time (NULL));
	for (pass = curves_completed + 1; pass <= curves_to_do; pass++) {
		double	temp;
		unsigned long iters;

restart1:	sprintf (buf, "curve #%ld", pass);
		title (buf);

		MAXERR = 0.0;

		/* Choose curve with order divisible by 16 and choose
		 *	a point (x/z) on said curve.
		 */
		do {
			curve = (rand () & 0x1F) * 65536.0 * 65536.0 * 65536.0;
			curve += (rand () & 0xFFFF) * 65536.0 * 65536.0;
			if (CPU_TYPE >= 5) RDTSC (j);
			curve += j ^ ((unsigned long) rand () << 16);
		} while (curve <= 5.0);
		if (specific_curve > 5.0) curve = specific_curve;

		choose12 (x, z, curve);
		sprintf (buf, "ECM on M%ld: curve #%ld with s = %.0f, B1 = %lu, B2 = %lu:\n",
			 PARG, pass, curve, B, C);
		OutputStr (buf);

		stage = 1;
		temp = timer (); iters = 0;
		start_sieve (2, C+D+D);
		for ( ; ; ) {
			prime = sieve ();
			if (prime > B) break;

/* Apply as many powers of prime as long as prime^n <= B */

			ell_mul (x, z, prime);
			if (prime <= SQRT_B) {
				unsigned long mult, max;
				mult = prime;
				max = B / prime;
				for ( ; ; ) {
					ell_mul (x, z, prime);
					mult *= prime;
					if (mult > max) break;
				}
			}
			if (++iters >= ITER_OUTPUT) {
				sprintf (buf, "At prime %lu.  Time thusfar %s sec.\n",
					 prime, tottime (temp));
				OutputStr (buf);
				iters = 0;
			}
			if (escapeCheck ()) {
				retval = FALSE;
				sprintf (buf, "Break at prime %lu.\n", prime);
				OutputStr (buf);
				if (gcd (z)) goto bingo;
				goto update_and_exit;
			}
			if (test_sumout ()) goto error;
		}

		sprintf (buf, "Stage 1 complete.  Time: %s sec.\n", tottime (temp));
		OutputStr (buf);

/* Check to see if we found a factor, as a by product normalize x */

		if (normalize (x, z)) {
			retval = TRUE;
			goto bingo;
		}
restart2:	dbltogw (1.0, z);

/* For debugging, output the normalized value of x */

/*#define XPRINT*/
#ifdef XPRINT
{
	char	buf[10000], *bufp;

	sprintf (buf, "Value of x at end of pass 1:\n");
	OutputBoth (buf);

	convertToBinary (x, U, &Ulen);
	while (compare (N, Nlen, U, Ulen) < 0)
		divide_helper (U, &Ulen, N, Nlen, NULL, NULL);

	bufp = &buf[9999];
	*bufp = 0;
	*--bufp = '\n';
	while (Ulen) {
		unsigned long i, carry = 0;
		for (i = Ulen-1; ; i--) {
			double x;
			x = carry * 4294967296.0 + U[i];
			U[i] = (unsigned long) (x / 10.0);
			carry = (unsigned long) (x - (double) U[i] * 10.0);
			if (i == 0) break;
		}
		*--bufp = (char) ('0' + carry);
		while (Ulen && U[Ulen-1] == 0) Ulen--;
	}
	OutputBoth (bufp);
}
#endif

/* Print out round off error */

		if (ERRCHK) {
			sprintf (buf, "Round off: %.10g\n", MAXERR);
			OutputStr (buf);
		}

/* Skip second pass if so requested */

		if (C <= B) continue;

/*
   Stage 2:  If NORMALIZED_STEP2 is defined, use an improved standard
   continuation from Paul Zimmerman, otherwise use the continuation
   from Crandall's code.
   x, z: coordinates of Q at the beginning of stage 2
*/

/* Initialize variables for second pass */

		sprintf (buf, "Initializing second stage, curve %d\n", pass);
		OutputStr (buf);
		temp = timer ();

/* Zimmermann's stage 2 init - enhanced by reducing the number */
/* of modular inverse calls. */

		/* Copy Q */
		gwcopy (x, nQx[0]);

		/* Compute 2*Q */
		ell_dbl_norm_delayed (x, nQx[1], zgrp[0]);
		gwfft (nQx[1], xorg); gwfft (zgrp[0], zorg);

		j = 1;		/* Count of grouped values */
		for (i = 2; i <= D/2; i++) {	/* compute 2*i*Q */
			if ((i & 1) == 0) {	/* 2*i*Q = 2*(i*Q) */
				if (i <= GROUP_MODINV)
					ell_dbl (nQx[(i>>1)], zgrp[(i>>1)-1],
						 nQx[i], zgrp[j]);
				else
					ell_dbl_norm_delayed (nQx[i>>1],
							      nQx[i], zgrp[j]);
			} else {		/* 2*i*Q = (i+1)*Q+(i-1)*Q */
				gwcopy (nQx[(i+1)>>1], nQx[i]);
				if (i <= GROUP_MODINV) {
					gwcopy (zgrp[((i+1)>>1)-1], zgrp[j]);
					ell_add (nQx[(i-1)>>1],
						 zgrp[((i-1)>>1)-1],
						 nQx[i], zgrp[j]);
				} else
					ell_add_norm_delayed (nQx[(i-1)>>1],
							      nQx[i], zgrp[j]);
			}

			if (escapeCheck ()) {
				retval = FALSE;
				goto update_and_exit;
			}
			if (test_sumout ()) goto error;

			if (++j < GROUP_MODINV) continue;
			if (grouped_normalize (&nQx[i+1-GROUP_MODINV],
					       zgrp, GROUP_MODINV)) goto bingo;
			if (i == GROUP_MODINV) gwfft (nQx[1], xorg);
			j = 0;
		}

		/* Compute 2*D*Q */
		ell_dbl (nQx[i-1], zgrp[j-1], nQx[i], zgrp[j]);
		j++;
		if (grouped_normalize (&nQx[i+1-j], zgrp, j)) goto bingo;

		/* Precompute the transforms of nQx */
		for (i = 0; i <= D/2; i++) gwfft(nQx[i], nQx[i]);

		/* Now that nQx[0] has been transformed, set stage variable */
		stage = 2;

		/* Init code that computes m*Q */
		mQ_init ();

/* Initialization of stage 2 complete */

		sprintf (buf, "Stage 2 init complete.  Time: %s sec.\n",
			 tottime (temp));
		OutputStr (buf);

/* Now do stage 2 */

		sprintf (buf, "Commencing second stage, curve %d...\n", pass);
		OutputStr (buf);
		temp = timer (); iters = 0;

/* Zimmermann's stage 2 */

		dbltogw (1.0, gg);
		iters = 0;
		for (m = 1; m < C + D; m += D + D) {
			gwnum	mQ;
			if (mQ_next (&mQ)) goto bingo;
			gwfft(mQ, t2);
			if (m + D > B) {
				char	pr[20000];
				memset (pr, 0, (D >> 3) + 1);
				for ( ; ; prime = sieve ()) {
					i = prime - (m - D);
					if (i >= D) break;
					gwsub3quick (t2, nQx[(D-i)>>1], t1);
					gwfftmul (t1, gg);
					bitset (pr, i);
				}
				if (prime == m) prime = sieve ();
				for ( ; ; prime = sieve ()) {
					i = prime - (m - D);
					if (i > D + D) break;
					if (bittst (pr, 2*D-i)) continue;
					gwsub3quick (t2, nQx[(i-D)>>1], t1);
					gwfftmul (t1, gg);
				}
			}

			if (++iters >= ITER_OUTPUT) {
				sprintf (buf, "At prime %lu.  Time thusfar %s sec.\n",
					 prime, tottime (temp));
				OutputStr (buf);
				iters = 0;
			}
			if (escapeCheck ()) {
				retval = FALSE;
				sprintf (buf, "Break at prime %lu.\n", prime);
				OutputStr (buf);
				if (gcd (gg)) goto bingo;
				goto update_and_exit;
			}
			if (test_sumout ()) goto error;
		}

/* Stage 2 is complete */

		sprintf (buf, "Stage 2 complete.  Time: %s sec.\n",
			 tottime (temp));
		OutputStr (buf);

/* See if we got lucky! */

		if (gcd (gg)) {
			retval = TRUE;
			goto bingo;
		}

/* Print out round off error */

		if (ERRCHK) {
			sprintf (buf, "Round off: %.10g\n", MAXERR);
			OutputStr (buf);
		}

/* Do not loop if we are testing a specific curve */

more_passes:	if (specific_curve > 5.0) {
			pass = curves_to_do;
			continue;
		}

/* Update curves_completed in INI file every N minutes */

		if (pass != curves_to_do) {
			time (&current_time);
			if ((unsigned long) (current_time - start_time) > DISK_WRITE_TIME * 60) {
				updateIniFile (p, B, curves_to_do, pass);
				start_time = current_time;
			}
		}
	}

/* Output line to results file indicating the number of curves run */

	sprintf (buf, "M%ld completed %ld ECM curves, B1 = %lu, B2 = %lu\n",
		 PARG, curves_to_do, B, C);
	writeResults (buf);
	spoolMessage (PRIMENET_RESULT_MESSAGE, buf);

/* Free memory and return */

	retval = TRUE;

/* Update the exponent in the work-to-do-file */

update_and_exit:
	updateIniFile (p, B, curves_to_do, pass-1);

/* Free memory and return */

exit:	ecm_cleanup ();
	end_sieve ();
	return (retval);

/* Print a message if we found a factor! */

bingo:	{
	char	buf[10000], *bufp;

	sprintf (buf, "ECM found a factor in curve #%ld, stage #%d\n", pass, stage);
	writeResults (buf);
	spoolMessage (PRIMENET_RESULT_MESSAGE, buf);
	sprintf (buf, "Sigma = %.0f, B1 = %lu, B2 = %lu.\n", curve, B, C);
	writeResults (buf);
	spoolMessage (PRIMENET_RESULT_MESSAGE, buf);

	bindiv (N, &Nlen, U, Ulen);

	bufp = &buf[9999];
	*bufp = 0;
	while (Ulen) {
		unsigned long i, carry = 0;
		for (i = Ulen-1; ; i--) {
			double x;
			x = carry * 4294967296.0 + U[i];
			U[i] = (unsigned long) (x / 10.0);
			carry = (unsigned long) (x - (double) U[i] * 10.0);
			if (i == 0) break;
		}
		*--bufp = (char) ('0' + carry);
		while (Ulen && U[Ulen-1] == 0) Ulen--;
	}
	sprintf (buf, "M%ld has a factor: %s\n", PARG, bufp);
	OutputBoth (buf);
	spoolMessage (PRIMENET_RESULT_MESSAGE, buf);

	if (isProbablePrime ()) {
		OutputBoth ("Cofactor is a probable prime!\n");
		clearExponents (PARG, PARG, 0);
		goto exit;
	}

	if (PARG > 5825) goto exit;
	goto more_passes;
	}

/* Output a message saying we are restarting */

error:	OutputBoth ("SUMOUT error occurred.\n");

/* Sleep five minutes before restarting */

	OutputStr ("Waiting 5 minutes before restarting\n");
	Sleep (300000);

/* Restart at appropriate point. */

	SUMOUT = 0;
	if (stage == 1) goto restart1;
	dbltogw (1.0, x);
	gwfft (x, x);
	gwfftfftmul (nQx[0], x, x);
	goto restart2;
}
