/*
 * IPShash.c   Prime95 side of PrimeNet packet hashing
 *             Do not distribute in public domain
 */

#define IPSHASH

#define CRC_INIT_WOLTMAN	0
#define CRC_POLY_WOLTMAN	0x6007

unsigned short* crctable = NULL;

void make_crctable()
{
	unsigned short i, j, r;

	if (crctable != NULL) return;

#ifdef __cplusplus
	crctable = new unsigned short[256];
#else
	crctable = (unsigned short*)malloc(256*sizeof(short));
#endif
	srand((unsigned)time(NULL));	/* seed random generator */

	for (i = 0; i < 256; i++)
	{
		r = i;
		for (j = 0; j < 8; j++)
			if (r & 1)
				r = (r >> 1) ^ CRC_POLY_WOLTMAN;
			else
				r >>= 1;
		crctable[i] = r;
	}
}

unsigned short make_hash_WOLTMAN( char* buf, int bufsize )
{
	unsigned short hash = CRC_INIT_WOLTMAN;

	while ( bufsize-- )
		hash = crctable[(hash ^ (*buf++)) & 0xFF] ^ (hash >> 8);

	return hash;
}

void hash_packet( short operation, void* pkt )
{
	/* compute and load 16-bit hash of packet */
	struct primenetUserInfo *z = (struct primenetUserInfo*)pkt;
	char* p = (char*)&z->salt;

	if (operation == PRIMENET_PING_SERVER_INFO) return;

	make_crctable();
	z->salt = rand();
	z->hash = make_hash_WOLTMAN( p, z->structSize - 4*sizeof(short) );
}
