
#include <windows.h> 
#include "service.h"
#include "main.h"
#include "prime95.h"
#include <direct.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>


// Globals
int THREAD_STOP = FALSE;
int DEBUGGING = FALSE;

//
//  FUNCTION: ServiceStart
//
//  PURPOSE: Actual code of the service
//           that does the work.
//
//  PARAMETERS:
//    dwArgc   - number of command line arguments
//    lpszArgv - array of command line arguments
//
//  RETURN VALUE:
//    none
//
//  COMMENTS:
//    The default behavior is to open a
//    named pipe, \\.\pipe\simple, and read
//    from it.  It the modifies the data and
//    writes it back to the pipe.  The service
//    stops when hServerStopEvent is signalled
//
VOID ServiceStart (DWORD dwArgc, LPTSTR *lpszArgv)
{

    ///////////////////////////////////////////////////
    //
    // Service initialization
    //

    // report the status to the service control manager.
    //
    if (!ReportStatusToSCMgr(
        SERVICE_START_PENDING, // service state
        NO_ERROR,              // exit code
        3000))                 // wait hint
        goto cleanup;

    // report the status to the service control manager.
    //
    if (!ReportStatusToSCMgr(
        SERVICE_RUNNING,       // service state
        NO_ERROR,              // exit code
        0))                    // wait hint
        goto cleanup;

    //
    // End of initialization
    //
    ////////////////////////////////////////////////////////

    ////////////////////////////////////////////////////////
    //
    // Service is now running, perform work until shutdown
    //

	prime_main ();


// free up resources

cleanup:
	freeMem ();
	if (HLIB) FreeLibrary (HLIB);

}

//
//  FUNCTION: ServiceStop
//
//  PURPOSE: Stops the service
//
//  PARAMETERS:
//    none
//
//  RETURN VALUE:
//    none
//
//  COMMENTS:
//    If a ServiceStop procedure is going to
//    take longer than 3 seconds to execute,
//    it should spawn a thread to execute the
//    stop code, and return.  Otherwise, the
//    ServiceControlManager will believe that
//    the service has stopped responding.
//    
VOID ServiceStop()
{
	SetThreadPriority (GetCurrentThread (), THREAD_PRIORITY_NORMAL);
	THREAD_STOP = TRUE;
}


void prime_main ()
{

// Set the priority correctly

	SetPriorityClass (GetCurrentProcess (),
		(PRIORITY < 2 || PRIORITY > 6) ?
			NORMAL_PRIORITY_CLASS :
			IDLE_PRIORITY_CLASS);
	SetThreadPriority (GetCurrentThread (),
		(PRIORITY == 1) ? THREAD_PRIORITY_IDLE :
		(PRIORITY == 2 || PRIORITY == 7) ? THREAD_PRIORITY_LOWEST :
		(PRIORITY == 3 || PRIORITY == 8) ? THREAD_PRIORITY_BELOW_NORMAL :
		(PRIORITY == 4 || PRIORITY == 9) ? THREAD_PRIORITY_NORMAL :
		(PRIORITY == 5 || PRIORITY == 10) ? THREAD_PRIORITY_ABOVE_NORMAL :
		THREAD_PRIORITY_HIGHEST);
	if (CPU_AFFINITY != 99)
		SetThreadAffinityMask (GetCurrentThread (), 1 << CPU_AFFINITY);

// Do some work

	primeContinue ();
}

/* GetIniSettings -- Get initial bounce settings from WIN.INI
 *
 * Return:  None
 */
void GetIniSettings()
{
	nameIniFiles (-1);
	readIniFiles ();

	IniGetString (INI_FILE, "ServiceName", SZSERVICENAME,
		      sizeof (SZSERVICENAME), "NTPrimeService");
	IniGetString (INI_FILE, "DisplayName", SZSERVICEDISPLAYNAME,
		      sizeof (SZSERVICEDISPLAYNAME), "Prime Service");
}
